/*
 * inst.cpp
 *
 * inst(T,U) class
 *
 * Ulrik Petersen
 * Created: 2/26-2001
 * Last update: 3/17-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <inst.h>
#include <assert.h>
#include <logging.h>



////////////////////////////////////////////////////////////
//
// IDDList
//
////////////////////////////////////////////////////////////
IDDList::IDDList() : m_list()
{
}

IDDList::~IDDList()
{
}

IDDList::const_iterator IDDList::begin() const
{
  return m_list.begin();
}

IDDList::const_iterator IDDList::end() const
{
  return m_list.end();
}

long IDDList::length() const
{
  return m_list.size();
}

void IDDList::make_empty()
{
  m_list.erase(m_list.begin(), m_list.end());
}


////////////////////////////////////////////////////////////
//
// CInstContents
//
////////////////////////////////////////////////////////////
CInstContents::CInstContents(const std::vector<std::string>& feature_names)
  : m_feature_names(feature_names), 
    m_is_aggregate(false), 
    m_min_monad(MAX_MONAD),
    m_max_monad(0)
{
}

CInstContents::~CInstContents()
{
  // Delete all CInstObject objects
  Monad_m2InstObjectMap::iterator i(m_map_by_monad_m.begin());
  Monad_m2InstObjectMap::iterator iend(m_map_by_monad_m.end());
  while (i != iend) {
    delete i->second;
    ++i;
  }
  m_map_by_monad_m.erase(m_map_by_monad_m.begin(), iend);
}

// Assume it is not there
void CInstContents::add_object(CInstObject *pObj)
{
  const CSet_of_monad_ms& rsom = pObj->get_monads();
  // Get first monad
  monad_m first_monad = rsom.first();
      
  // Get last monad
  monad_m last_monad = rsom.last();
      
  // Set min_monad
  if (m_min_monad > first_monad)
    m_min_monad = first_monad;
      
  // Set max_monad
  if (m_max_monad < last_monad) 
    m_max_monad = last_monad;

  // Add to m_map_by_monad_m
  Monad_m2InstObjectMap::value_type temp = Monad_m2InstObjectMap::value_type(first_monad, pObj);
  m_map_by_monad_m.insert(temp);
}

//
// Make in result a single mse with min..max
// where min is the minimum monad in ourselves
// and max is the maximum
// The reason it is called big_union is that it used to
// be a set-theoretical union of all of our objects'
// monad-sets.  This took too long, however.  q26.txt
// in particular went from 11.5 minutes to 1.5 minutes
// just by implementing this with just min..max.
void CInstContents::big_union(CSet_of_monad_ms& result)
{
  // Write to log
  LOG_WRITE_TIME("CInstContents::big_union", "Starting big-union.");

  // Set to the empty set
  result.clear();
  
  /*
  // Iterate through map
  Monad_m2InstObjectMap::const_iterator ci(m_map_by_id_d.begin());
  Monad_m2InstObjectMap::const_iterator cend(m_map_by_id_d.end());
  monad_m min_monad, max_monad;
  min_monad = MAX_MONAD;
  max_monad = 0;
  while (ci != cend) {
    // Get first and last monad of object
    const CSet_of_monad_ms& rsom = ci->second->get_monads();
    monad_m first_m = rsom.first();
    monad_m last_m = rsom.last();
    
    // Set min_monad
    min_monad = (min_monad > first_m) ? first_m : min_monad;

    // Set max_monad
    max_monad = (max_monad < last_m) ? last_m : max_monad;

    // Advance iterator
    ++ci;
  }
  */


  // Add m_min_monad .. m_max_monad, but only if there were
  // some members in the m_map_by_id_d.
  // If there were no members, we never got the chance to 
  // set m_min_monad and m_max_monad, and we would get
  // a BadMonadsException by adding the mse (MAX_MONAD, 0).
  if (!(m_min_monad == MAX_MONAD && m_max_monad == 0)) {
    result.add(m_min_monad, m_max_monad);
  }

  // Write to log
  LOG_WRITE_TIME("CInstContents::big_union", "Big-union finished.");
}


//
// Returns in ci1 and ci2 the beginning and end of a series of objects
// that have the starting monad Sm.
// If the list is nonempty, ci2 points to the element just after the 
// last element.
// If there are no objects starting at Sm, both ci1 and ci2 are end()
//
void CInstContents::objects_at_Sm(monad_m Sm, CInst::const_iterator& ci1, CInst::const_iterator& ci2)
{
  // Use equal_range to do the job.  
  // Note that we should have had const_iterators here,
  // but Sun's STL can't return const_iterators for equal_range.
  // So we cast to const_iterator afterwards instead.
  std::pair<Monad_m2InstObjectMap::iterator, Monad_m2InstObjectMap::iterator> p;
  p = m_map_by_monad_m.equal_range(Sm);
  ci1 = p.first;
  ci2 = p.second;

  // If we did not find Sm or any value greater than Sm, return with
  // ci1 == ci2 == end().
  if (ci1 == m_map_by_monad_m.end()) {
    return;
  } else {
    // If we did not find Sm, return end()
    if (ci1->second->get_monads().first() != Sm) {
      ci1 = ci2 = m_map_by_monad_m.end();
    }
  }
  return;
}

//
// This member function moves an object around in the map by monad m,
// placing it at new_first_monad rather than where it is at call-time.
void CInstContents::reassign_object(CInstObject *pObj, monad_m previous_first_monad)
{
  // We shouldn't reach this point;
  // After we changed CInstContents so that the map by monad_m was
  // only calculated after all querying had been done,
  // we no longer need this function, as it was only called
  // from EMdFDB::GetInst().
  assert(false);    

  // Get equal range
  typedef Monad_m2InstObjectMap::iterator MI;
  std::pair<MI,MI> g = m_map_by_monad_m.equal_range(previous_first_monad);
  for (MI p = g.first; p != g.second; ++p) {
    if (pObj == p->second) {
      // So we have found it.
      
      // Delete it
      m_map_by_monad_m.erase(p);
      
      // Insert it anew
      m_map_by_monad_m.insert(Monad_m2InstObjectMap::value_type(pObj->get_monads().first(), pObj));
      
      // Return
      return;
    }
  }

  // We should never be able to reach this.
  // This is because pObj should be found within the equal range
  assert(false);
}


void CInstContents::set_is_aggregate(bool is_aggregate) 
{ 
  m_is_aggregate = is_aggregate; 
}

bool CInstContents::is_aggregate(void) const
{ 
  return m_is_aggregate; 
}


/////////////////////////////////////////////////////////////////
//
// CInst
//
/////////////////////////////////////////////////////////////////

CInst::CInst(CInst* pOther)
{
  m_pContents = pOther->m_pContents;
  m_bDoDelete = false;
}

CInst::CInst(const std::vector<std::string>& feature_names)
{
  m_pContents = new CInstContents(feature_names);
  m_bDoDelete = true;
}

CInst::~CInst()
{
  if (m_bDoDelete)
    delete m_pContents;
}

CInst::const_iterator CInst::begin() const // relative to the current universe
{
  // Find end of map
  CInst::const_iterator cend(m_pContents->m_map_by_monad_m.end());
  
  // Find beginning of universe
  CInst::const_iterator ci;
  ci = m_pContents->m_map_by_monad_m.lower_bound(m_first_monad);
  while (true) {
    // If we hit the end, return with end()
    if (ci == cend) {
      break;
    }

    // Is this set of monads within the range?
    if (ci->second->get_monads().last() <= m_last_monad) {
      // Yes, break out of loop
      break; 
    } else {
      // No

      // Advance iterator
      ++ci;

      // Is this not the end?
      if (ci != cend) {
	// Yes, this is not the end

	// Is there no hope because this iterator points to an object
	// whose first monad is greater than m_last_monad?
	if (ci->second->get_monads().first() > m_last_monad) {
	  // Yes, set ci to the end of the map
	  ci = cend;
	}
      }
    }
  }
  return ci;
}

CInst::const_iterator CInst::next(CInst::const_iterator& ci) // relative to the current universe
{
  // Get end
  CInst::const_iterator cend = m_pContents->m_map_by_monad_m.end();

  // If we have already hit the end, return ci
  if (ci == cend)
    return ci;

  // Check in a loop
  while (true) {
    // Advance ci
    ++ci;

    // If we have hit the end, break out of loop
    if (ci == cend)
      break;

    // Check that ci is within the bounds
    const CSet_of_monad_ms& monads = ci->second->get_monads();
    if (monads.first() >= m_first_monad
	&& monads.last() <= m_last_monad) {
      // It was, break out of loop and return ci
      break;
    }

    // Is there no hope?
    if (monads.first() > m_last_monad) {
      // Yes, there is no hope.  Set ci to the end and break out of loop
      ci = cend;
      break;
    }
  }
  return ci;
}
