/*
 * pgconn.cpp
 *
 * PostgreSQL Database connection 
 *
 * Ulrik Petersen
 * Created: 1/27-2001
 * Last update: 4/28-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <pgconn.h>
#include <logging.h>
#include <iostream>
#include <cstdlib>
#include <sstream>


PgEMdFConnection::PgEMdFConnection(const std::string& host, const std::string& user, const std::string& passwd, const std::string& database_name)
  : m_pConn(NULL), m_pResult(NULL), m_bTransactionInProgress(false)
{
  std::ostringstream my_connection_str;

  // Use database emdf if database name not given
  std::string db;
  if (database_name.length() == 0) {
    db = "emdf";
  } else {
    db = database_name;
  }

  // Make connection string
  my_connection_str << "host='" << host << "' user='" << user << "' password='" << passwd << "' dbname='" << db << "'";

  // Connect
  m_pConn = PQconnectdb(my_connection_str.str().c_str());

#ifdef EMDF_DEBUG
  // Check that connection OK
  if (!ConnectionOk()) {
    cerr << "pgconn.cpp: Connection not OK connecting to database '"
	 << my_database << "'.\n";
    std::string error_message;
    bool res_err;
    res_err = ErrorMessage(error_message);
    std::cout << "Error message = '" << error_message << "'\n";
    std::cout << "res_err = " << res_err << "\n";
  }
#endif // EMDF_DEBUG
}

PgEMdFConnection::~PgEMdFConnection()
{
  // Commit transaction if in progress
  if (m_bTransactionInProgress)
    CommitTransaction();

  // Remove result
  ReleaseResults();

  // Close down connection
  if (m_pConn != 0) 
    PQfinish((PGconn*) m_pConn);
  m_pConn=NULL;
}


// Status of connection
// Returns true on connection good, false on connection bad
// Should be called right after constructor to see if we
// succeeded in setting up the connection.
bool PgEMdFConnection::ConnectionOk(void)
{
  if (m_pConn == 0)
    return false;
  else {
    return PQstatus((PGconn*) m_pConn) != CONNECTION_BAD;
  }
}


// Executing
bool PgEMdFConnection::Exec(const std::string& query)
{
  if (m_pConn == 0)
    return false;
  else {
    // Check that connection is OK
    if (!ConnectionOk())
      return false;

    // Write to log
#if EMDROS_LOGGING
    std::string log = "Starting query.\n" + query;
    LOG_WRITE_TIME("PgEMdFConnection::Exec", log);
#endif

    // Clear the Result Stucture if needed
    ReleaseResults();

    // Execute
    ExecStatusType status;
    m_pResult = PQexec((PGconn*) m_pConn, query.c_str());
  
    // Return the status
    if (m_pResult != NULL)
      status = PQresultStatus((PGresult*) m_pResult);
    else 
      status = PGRES_FATAL_ERROR;

    // Write to log
    LOG_WRITE_TIME("PgEMdFConnection::Exec", "Query finished.");

    // Return true or false based on status
    return (status == PGRES_COMMAND_OK) 
      || (status == PGRES_TUPLES_OK);
  }
}


// Get last error message
// Returns true on error message available 
// (even if it is "OK"), false on error 
// message not available.
bool PgEMdFConnection::ErrorMessage(std::string& out)
{
  if (m_pConn == 0)
    return false;
  else {
    out = (const char*) PQerrorMessage((PGconn*) m_pConn); 

    if (out.length() > 0)
      return true;
    else
      return false;
  }
}


// Transactions
bool PgEMdFConnection::BeginTransaction(void)
{
  if (m_pConn == 0)
    return false;
  else {
    if (m_bTransactionInProgress) {
      return false;
    } else {
      m_bTransactionInProgress = true;
      return Exec("BEGIN");
    }
  }
}

bool PgEMdFConnection::CommitTransaction(void)
{
  if (m_pConn == 0)
    return false;
  else {
    if (m_bTransactionInProgress) {
      m_bTransactionInProgress = false;
      return Exec("END");
    } else {
      std::cerr << "Warning: PgEMdFConnection::CommitTransaction: No transaction in progress." << std::endl << "... Something is wrong in the logic! ..." << std::endl;
      return false;
    }
  }
}

bool PgEMdFConnection::AbortTransaction(void)
{
  if (m_pConn == 0)
    return false;
  else {
    if (m_bTransactionInProgress) {
      m_bTransactionInProgress = false;
      return Exec("ROLLBACK");
    } else {
      std::cerr << "Warning: PgEMdFConnection::AbortTransaction: No transaction in progress." << std::endl << "... Something is wrong in the logic! ..." << std::endl;
      return false;
    }
  }
}


// Tuple-counts
// Both return -1 on not appliccable or error
long PgEMdFConnection::NoOfResultTuples(void)  // SELECT
{
  if (m_pConn == 0)
    return -1;
  else {
    return PQntuples((PGresult*) m_pResult); 
  }
}

long PgEMdFConnection::NoOfCommandTuples(void) // INSERT, UPDATE, DELETE
{
  if (m_pConn == 0)
    return -1;
  else {
    const char *szCmdTuples = PQcmdTuples((PGresult*) m_pResult);
    return szCmdTuples[0] ? atoi(szCmdTuples) : -1;
  }
}


// Field count
// Returns -1 on no fields.
int PgEMdFConnection::NoOfFields(void)
{
  if (m_pConn == 0)
    return -1;
  else {
    return PQnfields((PGresult*) m_pResult); 
  }
}



// Tuple-access by field number
// First tuple is 0
// First field is 0.


bool PgEMdFConnection::Vacuum(bool bAnalyze)
{
  if (m_pConn == 0)
    return false;
  else {
    std::string Query = "VACUUM";
    if (bAnalyze)
      Query += " ANALYZE";
    return Exec(Query);
  }
}

// Release results so they don't take up memory
void PgEMdFConnection::ReleaseResults(void)
{
  // Remove result
  if (m_pResult != 0) 
    PQclear((PGresult*) m_pResult);
  m_pResult=NULL;
}
