/*
 * mql_execute.cpp
 *
 * Functions for executing MQL queries
 * Created: 3/23-2001
 * Last update: 4/10-2004
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2004  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_execute.h>
#include <mql_error.h>
#include <mql_result.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <mql_extern.h>
#include <mql_execution_environment.h>
#include <string_func.h>
#include <logging.h>
#include <mql_sheaf.h>

#undef yyFlexLexer
#define yyFlexLexer xxFlexLexer
#include <MQLFlexLexer.h>

#undef yyFlexLexer
#define yyFlexLexer zzFlexLexer
#include <MQLFlexLexer.h>




bool execute_statement(CMQL_execution_environment *pEE, bool& bResult)
{
  // Declare local variables
  bool return_value = true;

  // Initialize bResult
  bResult = true;

  // Weed
  if (return_value && bResult) {
    pEE->nCompilerStage = COMPILER_STAGE_WEED;
    pEE->pStatement->Weed(bResult);
    if (!bResult) {
      pEE->pError->mql_prepend_error("Weeding failed with compiler error\n");
      return_value =  true;
    } 
  }

  // Create symbol table
  pEE->pSymbolTable = new CMQLSymbolTable();

  // Symbol
  if (return_value && bResult) {
    pEE->nCompilerStage = COMPILER_STAGE_SYMBOL;
    if (!(pEE->pStatement->Symbol(bResult))) {
      pEE->pError->mql_prepend_error("Symbolchecking failed with database error\n");
      return_value =  false;
    }
    if (return_value && !bResult) {
      pEE->pError->mql_prepend_error("Symbolchecking failed with compiler error\n");
      return_value =  true;
    }
  }

  // Type
  if (return_value && bResult) {
    pEE->nCompilerStage = COMPILER_STAGE_TYPE;
    if (!(pEE->pStatement->Type(bResult))) {
      pEE->pError->mql_prepend_error("Typechecking failed with database error\n");
      return_value =  false;
    }
    if (return_value && !bResult) {
      pEE->pError->mql_prepend_error("Typechecking failed with compiler error\n");
      return_value =  true;
    }
  }

  // Monads
  if (return_value && bResult) {
    pEE->nCompilerStage = COMPILER_STAGE_MONADS;
    if (!(pEE->pStatement->Monads(bResult))) {
      pEE->pError->mql_prepend_error("Monads-checking failed with database error\n");
      return_value =  false;
    }
    if (return_value && !bResult) {
      pEE->pError->mql_prepend_error("Monads-checking failed with compiler error\n");
      return_value =  true;
    }
  }

  // Exec
  if (return_value && bResult) {
    pEE->nCompilerStage = COMPILER_STAGE_EXEC;
    if (pEE->pOut->get_output_kind() == kOKXML) {
      pEE->pOut->start_tag("mql_progress", true);
      pEE->pOut->newline();
      pEE->pOut->flush();
    }
    if (!(pEE->pStatement->Exec())) {
      pEE->pError->mql_prepend_error("Executing failed with database error\n");
      return_value =  false;
    } else
      return_value = true;
    if (pEE->pOut->get_output_kind() == kOKXML) {
      pEE->pOut->end_tag("mql_progress", true);
      pEE->pOut->flush();
    }
  }

  // Return
  return return_value;
}

bool execute(CMQL_execution_environment* pEE, bool& bResult)
{
  // Clear errors
  pEE->pError->mql_clear_error();
  pEE->pDB->clear_local_error();

  // Declare return_value
  bool return_value = true;
  
  // Initialize bResult
  bResult = true;

  // Initialize pEE->nCompilerStage
  pEE->nCompilerStage = COMPILER_STAGE_PARSE;

  // Parser
  int iResult = yyparse((void *) pEE);

  // Did parsing fail?
  if (iResult == 1) {
    bResult = false;
    pEE->pError->mql_prepend_error("Parsing failed\n");
    return_value = true;
  } 

  // Was the parsing a success?
  if (bResult) {
    // If so, execute statements
    return_value = execute_statement(pEE, bResult);
  }

  // Return
  return return_value;
}

void execute_query(CMQL_execution_environment *pEE, std::string query, bool& bDBResult, bool& bCompileResult, bool bPrintResult, bool bReportError)
{
  // Clean from previous round (if any)
  pEE->clean();

  // Initialize 
  std::istringstream istrstr(query);
    
  // Lexer
  pEE->pLexer = new xxFlexLexer(&istrstr);

  // Write log
  LOG_WRITE_TIME("execute_query()", "Executing query:");
  LOG_WRITE("execute_query()", query.c_str());

  // Execute
  bDBResult = execute(pEE, bCompileResult);

  // Write log
  LOG_WRITE_TIME("execute_query()", "Execution finished.");


  // Report error if necessary
  std::string strError;
  if (!bDBResult 
      && bReportError 
      && pEE->pOut->get_output_kind() == kOKConsole) {
    std::cerr << "ERROR: Database error while executing query:" << std::endl;
    std::cerr << "--------------------------------------------" << std::endl;
    std::cerr << query << std::endl;
    std::cerr << "--------------------------------------------" << std::endl;
    std::cerr << pEE->pDB->ErrorMessage() << std::endl; 
    std::cerr << pEE->pError->mql_get_error() << std::endl;
  } else {
    if (!bCompileResult 
	&& bReportError
	&& pEE->pOut->get_output_kind() == kOKConsole) {
      std::cerr << "ERROR: Compile error while executing query:" << std::endl;
      std::cerr << "--------------------------------------------" << std::endl;
      std::cerr << query << std::endl;
      std::cerr << "--------------------------------------------" << std::endl;
      std::cerr << pEE->pError->mql_get_error() << std::endl;
    } 
  }
}

bool mql_execute_stream(CMQL_execution_environment *pEE, std::istream& strin, bool& bResult, bool bPrintResult, bool bReportError)
{
  // Take care of XML header
  if (pEE->pOut->get_output_kind() == kOKXML) {
    pEE->pOut->print_XMLDecl();
    pEE->pOut->print_DTD_start("mql_results");
    CMQLResult::print_DTD(pEE->pOut);
    CTable::print_DTD(pEE->pOut);
    CSheaf::print_DTD(pEE->pOut);
    pEE->pOut->print_DTD_end();
    pEE->pOut->start_tag("mql_results", true);
    pEE->pOut->flush();
  }

  // Initialize bExecResult
  bool bExecDBResult = true;
  bool bExecCompileResult = true;

  // Initialize Query-splitting lexer
  zzFlexLexer *pQSL = 
    new zzFlexLexer(&strin);

  // Split into queries
  std::string query;
  pQSL->yylval = new YYSTYPE;
  pQSL->yylval->pString = &query;
  int QSLResult;
  do {
    // Get query out of stream
    QSLResult = pQSL->yylex();

    bool bRunQuery = true;
    if (QSLResult == 0) {
      if (is_other_than_whitespace(query)) {
	query += " GO";
	bRunQuery = true;
      } else {
	bRunQuery = false;
      }
    }
    
    // Run query if necessary
    if (bRunQuery) {
      bool bCompileResult = true;
      bool bDBResult = true;

      // Print pre-result material
      if (bPrintResult && pEE->pOut->get_output_kind() == kOKXML) {
	// Start mql_result
	pEE->pOut->start_tag("mql_result", true);
	pEE->pOut->flush();
      }

      // Execute
      execute_query(pEE, query, bDBResult, bCompileResult, bPrintResult, bReportError);
      
      // Make sure we report the right thing back
      bExecDBResult = bExecDBResult && bDBResult;
      bExecCompileResult = bExecCompileResult && bCompileResult;

      // Make success-result
      bool bSuccess = bDBResult && bCompileResult;

      // Print pre-result material
      if (bPrintResult && pEE->pOut->get_output_kind() == kOKXML) {
	// Output "success" node
	AttributePairList status_attributes;
	status_attributes.push_back(AttributePair("success", bool_alpha2string(bSuccess)));
	pEE->pOut->start_single_tag("status", status_attributes, true);

	// Output error node?
	if (!bSuccess) {
	  //
	  // Start error node
	  //
	  pEE->pOut->start_tag("error", true);

	  //
	  // error_source node
	  // 

	  // Figure out source
	  bool bWasDB;
	  bool bWasCompiler;
	  if (!bDBResult) {
	    bWasDB = true;
	    bWasCompiler = false;
	  } else if (!bCompileResult) {
	    bWasDB = false;
	    bWasCompiler = true;
	  } else {
	    assert(false);
	  }

	  // Make source string
	  std::string error_source_string;
	  error_source_string = (bWasDB) ? "db" : "compiler";
	 
	  // Make error source attributes
	  AttributePairList error_source_attributes;
	  error_source_attributes.push_back(AttributePair("source", error_source_string));

	  // Emit "error_source" node
	  pEE->pOut->start_single_tag("error_source", error_source_attributes, true);

	  //
	  // error_stage node
	  //

	  // Get stage string
	  std::string stage_string;
	  switch (pEE->nCompilerStage) {
	  case COMPILER_STAGE_NONE:	
	    stage_string = "none";
	    break;
	  case COMPILER_STAGE_PARSE:	
	    stage_string = "parse";
	    break;
	  case COMPILER_STAGE_WEED:
	    stage_string = "weed";
	    break;
	  case COMPILER_STAGE_SYMBOL:
	    stage_string = "symbol";
	    break;
	  case COMPILER_STAGE_TYPE:
	    stage_string = "type";
	    break;
	  case COMPILER_STAGE_MONADS:
	    stage_string = "monads";
	    break;
	  case COMPILER_STAGE_EXEC:
	    stage_string = "exec";
	    break;
	  default:
	    assert(false);
	    break;
	  }
	  
	  // Make stage attributes
	  AttributePairList error_stage_attributes;
	  error_stage_attributes.push_back(AttributePair("stage", stage_string));
	  
	  // Emit error_stage node
	  pEE->pOut->start_single_tag("error_stage", error_stage_attributes, true);

	  //
	  // error_message_db node
	  //
	  pEE->pOut->start_tag("error_message_db", true);
	  std::string strError = pEE->pDB->ErrorMessage();
	  pEE->pOut->out_char_data(strError);
	  pEE->pOut->end_tag("error_message_db", true);
	  
	  //
	  // error_message_compiler node
	  //
	  pEE->pOut->start_tag("error_message_compiler", true);
	  strError = pEE->pError->mql_get_error();
	  pEE->pOut->out_char_data(strError);
	  pEE->pOut->end_tag("error_message_compiler", true);

	  // End "error" node
	  pEE->pOut->end_tag("error", true);
	}
	pEE->pOut->flush();
      }

      // Write to log
      LOG_WRITE_TIME("mql_execute_stream()", "Printing results.");

      // Print result
      if (bSuccess && bPrintResult && (pEE->pStatement != 0)) {
	CMQLResult* pMQLResult = pEE->pStatement->get_result();
	if (pMQLResult != 0)
	  pMQLResult->Print(pEE->pOut, pEE);
	pEE->pOut->flush();
      }

      // Write to log
      LOG_WRITE_TIME("mql_execute_stream()", "Printing finished.");

      // Print post-result material
      if (bPrintResult && pEE->pOut->get_output_kind() == kOKXML) {
	pEE->pOut->end_tag("mql_result", true);
	pEE->pOut->flush();
      }
    }
    
    // Clean up
    query = "";
  } while (QSLResult != 0 && !(pEE->bQuit));
  
  // Clean up
  delete pQSL->yylval;
  delete pQSL;

  // Take care of XML footer
  if (pEE->pOut->get_output_kind() == kOKXML) {
    pEE->pOut->end_tag("mql_results", true);
    pEE->pOut->newline();
  }

  // Return execution result
  bResult = bExecCompileResult;
  return bExecDBResult;
}


bool mql_execute_string(CMQL_execution_environment *pEE, const std::string& input, bool& bResult, bool bPrintResult, bool bReportError)
{
  // Lexer
  std::istringstream strstream(input);

  // Execute
  bool bExecResult = mql_execute_stream(pEE, strstream, bResult, bPrintResult, bReportError);
  
  // Return execution result
  return bExecResult;
}


bool mql_execute_file(CMQL_execution_environment *pEE, std::string filename, bool& bResult, bool bPrintResult, bool bReportError)
{
  // Open file
  std::ifstream fin;
  fin.open(filename.c_str());
  if (!fin) {
    pEE->pError->mql_append_error("Could not open " + filename + " for reading.");
    return false;
  }

  // Call other function
  bool bExecResult = mql_execute_stream(pEE, fin, bResult, bPrintResult, bReportError);

  // Return result
  return bExecResult;
}


