/*
 * mql_features_statements.cpp
 *
 * MQL FEATURES statements
 *
 * Ulrik Petersen
 * Created: 3/6-2001 (March 6, 2001)
 * Last update: 3/2-2004
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2004  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_features_statements.h>
#include <string_func.h>


/////////////////////////////////////////////////////////////
//
// GET FEATURES
//
// CGetFeaturesStatement
//
////////////////////////////////////////////////////////////
CGetFeaturesStatement::CGetFeaturesStatement(CMQL_execution_environment *pEE, 
					     CFeature* feature_list,
					     CID_D* id_ds,
					     std::string* object_type_name)
  : CStatement(pEE),
    m_id_ds(id_ds),
    m_features(feature_list),
    m_object_type_name(object_type_name),
    m_object_type_id(NIL)
{
}

CGetFeaturesStatement::~CGetFeaturesStatement()
{
  delete m_features;
  delete m_object_type_name;
  delete m_id_ds;
}

//
// - Check that the object type exists. In doing so, store its object
//   type_id in the AST.
//
// - Check that the objects with the given id_ds exists.
//
// - Check that the objects all belong to the same type, 
//   namely the one given.
//
// - Check that the features exist for the given object type.
//
bool CGetFeaturesStatement::Symbol(bool& bResult)
{
  //
  // - Check that the object type exists. In doing so, store its object
  //   type_id in the AST.
  //
  if (!m_pEE->pDB->ObjectTypeExists(*m_object_type_name, bResult, m_object_type_id))
    return false;
  if (!bResult)
    return true;

  //
  // - Check that all the id_ds refer to objects that exist in the given type.
  //
  // UP 2004-03-02: Commented out because it was prohibitively expensive
  /*
  if (!m_id_ds->SymbolObjectsExistInType(m_pEE, *m_object_type_name, m_object_type_id, bResult))
    return false;
  if (!bResult)
    return true;
  */

  //
  // - Check that the features exist for the given object type.
  //
  if (!m_features->SymbolFeaturesExist(m_pEE, m_object_type_id, bResult))
    return false;
  if (!bResult)
    return true;

  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return true;
}

bool CGetFeaturesStatement::Type(bool& bResult)
{
  //
  // - Get feature name and type
  //
  if (!m_features->TypeFeatureName(m_pEE, bResult))
    return false;
  if (!bResult)
    return true;

  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return true;
}

bool CGetFeaturesStatement::Exec()
{
  // Create result
  m_result = new CMQLResult();

  // Add headers
  m_result->append_header("id_d", kTCID_D);
  m_features->ExecMakeNameList(m_result);

  // Get a list of feature_infos
  std::list<feature_info> feature_infos;
  m_features->ExecMakeFeatureList(feature_infos);

  // Get a list of the id_ds
  std::list<id_d_t> id_ds;
  m_id_ds->ExecMakeList(id_ds);

  // Ask the database
  std::list<std::list<std::string> > results;
  if (!m_pEE->pDB->GetFeatures(*m_object_type_name,
			feature_infos,
			id_ds,
			results))
    return false;
  
  // Add the results
  std::list<std::list<std::string> >::const_iterator ci(results.begin());
  std::list<std::list<std::string> >::const_iterator cend(results.end());
  while (ci != cend) {
    // Append result
    m_result->append(*ci);

    // Advance iterator
    ++ci;
  }
  
  // If we got this far, there were no DB errors
  return true;
}


