/*
 * mql_object.cpp
 *
 * MQL representation of an object while processing the query
 *
 * Ulrik Petersen
 * Created: 4/2-2001
 * Last update: 6/21-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_object.h>
#include <mql_extern.h>
#include <assert.h>
#include <algorithm>

CMQLObject::CMQLObject() : m_feature_name_vec(0)
{
}

CMQLObject::~CMQLObject()
{
  delete m_feature_name_vec;
  unsigned int feature_value_vec_size = m_feature_value_vec.size();
  for (unsigned int i = 0; i < feature_value_vec_size; ++i) {
    delete m_feature_value_vec[i];
  }
}

void CMQLObject::add_feature(const feature_info& fi)
{
  // Get feature name
  std::string feature_name;
  feature_name = fi.get_name();

  // Only add if not there already
  std::list<feature_info>::const_iterator ci(m_feature_infos.begin());
  std::list<feature_info>::const_iterator cend(m_feature_infos.end());
  while (ci != cend) {
    // Check if this is the one
    if (strcmp_nocase(ci->get_name(), feature_name) == 0) {
      // It was, so return without adding.
      return;
    }

    // Advance iterator
    ++ci;
  }

  // It wasn't there, so add it at the back
  m_feature_infos.push_back(fi);

  // Resize m_feature_value_vec
  m_feature_value_vec.resize(m_feature_infos.size());
}

void CMQLObject::retrieve_feature_values(EMdFDB* pDB, const std::string& object_type_name, id_d_t object_id_d)
{
  // Declare list
  std::list<CEMdFValue *> feature_values;

  // Get values
  if (!pDB->GetFeatures(object_type_name, m_feature_infos, object_id_d, feature_values))
      throw CEMdFDBDBError();

  // Add them to the m_feature_value_vec
  std::list<CEMdFValue*>::const_iterator fvi;
  std::list<feature_info>::const_iterator fii;
  std::list<feature_info>::const_iterator fiend(m_feature_infos.end());
  fii = m_feature_infos.begin();
  fvi = feature_values.begin();
  int index = 0;
  while (fii != fiend) {
    delete m_feature_value_vec[index];
    m_feature_value_vec[index] = *fvi;
    ++index;
    ++fvi;
    ++fii;
  }
}

void CMQLObject::retrieve_feature_values(CInstObject *pOther)
{
  std::list<feature_info>::const_iterator fii(m_feature_infos.begin());
  std::list<feature_info>::const_iterator fiend(m_feature_infos.end());
  unsigned int index = 0;
  while (fii != fiend) {
    delete m_feature_value_vec[index];
    m_feature_value_vec[index] = new CEMdFValue(*(pOther->get_feature(index)));
    ++index;
    ++fii;
  }
}


//
// Get a vector of strings of feature-names in m_map
// Must only be called after all the adding of feature names has taken place.
// The order of the list is fixed: The order imposed on the strings by '<'.
const std::vector<std::string>& CMQLObject::get_feature_names(void) 
{
  // Should we create the name vector?
  if (m_feature_name_vec == 0) {
    // Yes

    m_feature_name_vec = new std::vector<std::string>;
    m_feature_name_vec->resize(m_feature_infos.size());
    unsigned int index = 0;
   
    std::list<feature_info>::const_iterator fi_i(m_feature_infos.begin());
    std::list<feature_info>::const_iterator fi_iend(m_feature_infos.end());
    while (fi_i != fi_iend) {
      (*m_feature_name_vec)[index] = fi_i->get_name();
      ++index;
      ++fi_i;
    }
    
    // Sort vector
    // std::sort(m_feature_name_vec->begin(), m_feature_name_vec->end());
  }
  return *m_feature_name_vec;
}

short int CMQLObject::get_feature_index(const std::string& feature_name) const
{
  std::list<feature_info>::const_iterator fi_i(m_feature_infos.begin());
  std::list<feature_info>::const_iterator fi_iend(m_feature_infos.end());
  short int index = 0;
  while (fi_i != fi_iend) {
    if (strcmp_nocase(fi_i->get_name(), feature_name) == 0)
      return index;
    ++index;
    ++fi_i;
  }
  // We shouldn't get here, since it should have been in m_feature_infos
  assert(false);
  return -1;
}



