/*
 * mql_symboltable.cpp
 *
 * MQL symbol table for object references
 *
 * Ulrik Petersen
 * Created: 3/16-2001
 * Last update: 3/28-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_query.h>
#include <mql_object.h>
#include <mql_symboltable.h>
#include <cassert>


//////////////////////////////////////////////////////////////
//
// CMQLSymbolTableEntry
//
//////////////////////////////////////////////////////////////
CMQLSymbolTableEntry::CMQLSymbolTableEntry(const std::string& object_reference,
					   const std::string& object_type_name,
					   id_d_t object_type_id)
  : m_object_type_id(object_type_id),
    m_object_type_name(object_type_name),
    m_vec(true),
    m_object_reference(object_reference),
    m_pFeature_infos(0)
{
}

CMQLSymbolTableEntry::~CMQLSymbolTableEntry()
{
  short int index;
  short int size = m_vec.size();
  for (index = 0; index < size; ++index) {
    delete m_vec[index]->first;
    m_vec[index]->first = 0;
  }
  delete m_pFeature_infos;
}

short int CMQLSymbolTableEntry::add_feature(const feature_info& fi)
{
  ValPair *pVal_pair = new ValPair((CEMdFValue *) 0, fi);
  bool bWasAdded;
  return m_vec.add_value(pVal_pair, valpair_equal(), bWasAdded);
  if (!bWasAdded)
    delete pVal_pair;
}

void CMQLSymbolTableEntry::set_feature(short int index, 
				       CEMdFValue *feature_value)
{
  delete m_vec[index]->first;
  m_vec[index]->first = feature_value;
}

const std::list<feature_info>& CMQLSymbolTableEntry::get_feature_infos(void)
{
  if (m_pFeature_infos == 0) {
    m_pFeature_infos = new std::list<feature_info>;
    short int index;
    short int size = m_vec.size();
    for (index = 0; index < size; ++index) {
      m_pFeature_infos->push_back(m_vec[index]->second);
    }
  }
  return *m_pFeature_infos;
}

void CMQLSymbolTableEntry::get_feature_value(short int index, 
					     const CEMdFValue **result)
{
  *result = m_vec[index]->first;
}

id_d_t CMQLSymbolTableEntry::get_feature_type_id(short int index)
{
  return m_vec[index]->second.get_type();
}

void CMQLSymbolTableEntry::set_id_d(id_d_t id_d_O)
{
  m_id_d = id_d_O;
}

void CMQLSymbolTableEntry::set_features(EMdFDB* pDB)
{
  // Get list of features
  const std::list<feature_info>& feature_infos = get_feature_infos();

  // Get values
  std::list<CEMdFValue*> feature_values;
  if (!pDB->GetFeatures(m_object_type_name,
			feature_infos,
			m_id_d,
			feature_values))
    throw CEMdFDBDBError();

  // Set feature values
  short int index;
  short int size = m_vec.size();
  std::list<CEMdFValue*>::const_iterator fvi(feature_values.begin());
  for (index = 0; index < size; ++index) {
    // Set feature
    set_feature(index, *fvi);

    // Advance iterator
    ++fvi;
  }
}


void CMQLSymbolTableEntry::set_features(CMQLObject *pMQLObject)
{
  // Get list of features
  const std::list<feature_info>& feature_infos = get_feature_infos();

  // Set feature values
  short int index;
  short int size = m_vec.size();
  std::list<feature_info>::const_iterator fici = feature_infos.begin();
  for (index = 0; index < size; ++index) {
    // Get index from MQL object
    short int mql_object_feature_index = pMQLObject->get_feature_index(fici->get_name());

    // Get EMdF value from MQL object
    const CEMdFValue *pEMdFValue = pMQLObject->get_feature_value(mql_object_feature_index);

    // Clone pEMdFValue
    CEMdFValue *pEMdFValueCloned = new CEMdFValue(*pEMdFValue);

    // Set feature
    set_feature(index, pEMdFValueCloned);

    // Advance iterator
    ++fici;
  }
}






//////////////////////////////////////////////////////////////
//
// CMQLSymbolTable
//
//////////////////////////////////////////////////////////////

CMQLSymbolTable::CMQLSymbolTable()
  : m_vec_ste(true), m_vec_object_block(true)
{
}

// Delete entries
CMQLSymbolTable::~CMQLSymbolTable()
{
  short int index;
  short int size = m_vec_ste.size();
  for (index = 0; index < size; ++index) {
    delete m_vec_ste[index]->second;
    m_vec_ste[index]->second = 0;
  }
}

short int CMQLSymbolTable::insert(CMQLSymbolTableEntry* pSTE, CObject_block_common *pObjectBlock)
{
  // Get object reference
  std::string object_reference = pSTE->get_object_reference();

  // Make sure it doesn't exist yet
  short int dummy_index;
  if (exists(object_reference, dummy_index)) {
    assert(false);
    throw CMQLSymbolTableException();
  }

  // Add to m_vec_object_block
  StrCOBCPair  *pStrCOBCPair = new StrCOBCPair(object_reference, pObjectBlock);
  bool bWasAdded;
  short int object_block_index = m_vec_object_block.add_value(pStrCOBCPair, strcobc_pair_equal(), bWasAdded);
  if (!bWasAdded) {
    delete pStrCOBCPair;
  }

  // Add to m_vec_ste
  StrSTEPair *pStrSTEPair = new StrSTEPair(object_reference, pSTE);
  short int ste_block_index = m_vec_ste.add_value(pStrSTEPair, strste_pair_equal(), bWasAdded);
  if (!bWasAdded) {
    delete pStrSTEPair;
  }

  // Make sure indexes are equal
  assert(ste_block_index == object_block_index);

  // Return index
  return ste_block_index;
}

bool CMQLSymbolTable::exists(const std::string& object_reference, short int& index)
{
  StrSTEPair p(object_reference, (CMQLSymbolTableEntry *) 0);
  index = m_vec_ste.get_index(&p, strste_pair_equal());
  return index != m_vec_ste.no_element;
}

CMQLSymbolTableEntry* CMQLSymbolTable::get(short int index)
{
  return m_vec_ste[index]->second;
}

CObject_block_common* CMQLSymbolTable::get_object_block(short int index)
{
  return m_vec_object_block[index]->second;
}


short int CMQLSymbolTable::get_index(const std::string& object_reference)
{
  StrSTEPair p(object_reference, (CMQLSymbolTableEntry *) 0);
  return m_vec_ste.get_index(&p, strste_pair_equal());
}
