/*
 * emdfdb.h
 *
 * EMdF database base class
 *
 * Ulrik Petersen
 * Created: 1/27-2001
 * Last update: 5/19-2004
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2004  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/



#ifndef EMDFDB__H__
#define EMDFDB__H__

#ifndef SWIG
#include <emdf.h>
#include <monads.h>
#include <conn.h>
#include <inst.h>
#include <infos.h>
#include <table.h>
#include <string>
#include <list>
#include <set>
#include <sstream>
#include <map>
#endif // !defined SWIG


class CEMdFDBException {};
class CEMdFDBDBError : public CEMdFDBException {};

#ifndef SWIG
typedef std::map<id_d_t, std::string> IDD2StringMap;
#endif 

class CEMdFValue; // forward declaration

class EMdFDB {
protected:
  EMdFConnection* pConn;
  std::string m_local_errormessage;
  CTable *m_enum_cache;
  CTable *m_object_type_cache_table;
  CTable *m_enum_const_cache;
  IDD2StringMap m_object_type_cache_map;
  CTable *m_feature_cache;
public:
  //
  // Construction and destruction
  EMdFDB();
  virtual ~EMdFDB();        

  // Get backend-name
  static std::string BackendName(void);


#ifndef SWIG
  //
  // Sequence manipulation
  virtual bool GetNextObjectID_D(/* out */ id_d_t& id_d);
  virtual bool GetNextOtherID(/* out */ id_d_t& id_d);
  virtual bool SetNextObjectID_DIfNotHigher(id_d_t next_id_d);
protected:
  virtual bool GetNextID(int sequence, id_d_t& out);
public:
#endif // !defined SWIG
  

  //
  // Database manipulation
#ifndef SWIG
  virtual bool CreateDatabase(const std::string& db_name);
  virtual bool DropDatabase(const std::string& db_name);
#endif
  virtual bool UseDatabase(const std::string& db_name);
  virtual bool DatabaseExists(const std::string& db_name, bool& result);
  virtual bool DBIsInitialized(bool& bIsInitialized);
               // Note: Do we really need DatabaseExists?


  //
  // Object type manipulation
#ifndef SWIG
  virtual bool CreateObjectType(const std::string& object_type_name,
				const std::list<feature_info>& feature_infos,
                                /* out */ id_d_t& object_type_id);
  virtual bool DropObjectType(const std::string& object_type_name, id_d_t object_type_id);
  virtual bool ObjectTypeExists(const std::string& object_type_name, 
                                /* out */ bool& result, 
                                /* out */ id_d_t& object_type_id); 
  virtual bool ObjectTypeExists(id_d_t object_type_id,
                                /* out */ bool& result, 
                                /* out */ std::string& object_type_name); 

  //
  // Enumeration manipulation
  virtual bool CreateEnum(const std::string& enum_name, 
                          /* out */ id_d_t& enum_id);
  virtual bool DropEnum(id_d_t enum_id);
  virtual bool EnumExists(const std::string& enum_name, 
                          /* out */ bool& result,
                          /* out */ id_d_t& enum_id);
  virtual bool EnumExists(id_d_t enum_id,
                          /* out */ bool& result,
			  /* out */ std::string& enum_name);
  virtual bool GetDefault(id_d_t enum_id,
			  /* out */ bool& result,
			  /* out */ std::string& default_constant_name);

  //
  // Enumeration constant manipulation
  virtual bool CreateEnumConstant(id_d_t enum_id,
                                  const enum_const_info& ec_info);
  virtual bool EnumConstExists(const std::string& enum_const_name, 
                               id_d_t enum_id, 
                               /* out */ bool& result,
                               /* out */ long& value,
                               /* out */ bool& is_default);
  virtual bool EnumConstExists(long value,
			       id_d_t enum_id,
			       bool& bExists,
			       /* out */ std::string& enum_const_name,
			       /* out */ bool& is_default);
  // Returns the string-representation of an enumeration constant in 
  // enum enum_name with the value value.  If more than one enum constant
  // has the same value, it is undefined which of them is returned.
  virtual bool GetEnumConstNameFromValue(long value,
					 const std::string& enum_name, 
					 /* out */ std::string& enum_const_name);
  virtual bool DropEnumConst(id_d_t enum_id, const std::string& enum_const_name, long value);
  virtual bool UpdateEnumConst(const std::string& enum_const_name, 
			       id_d_t enum_id,
			       long old_value,
			       long new_value);
  virtual bool SetDefaultEnumConst(id_d_t enum_id,
				   const std::string& enum_value_name);

  //
  // Feature manipulation
  virtual bool CreateFeature(const std::string& feature_name,
                             id_d_t object_type_id, 
                             id_d_t feature_type_id, 
                             const std::string& default_value, 
                             bool computed);  // Assumes it does not
                                              // exist
  virtual bool AddFeature(const std::string& object_type_name,
			  id_d_t object_type_id,
			  const feature_info& fi);
  virtual bool DropFeature(const std::string& object_type_name,
			   const std::string& feature_name,
                           id_d_t object_type_id);
  virtual bool FeatureExists(const std::string& feature_name,
                             id_d_t object_type_id,
                             /* out */ bool& result,
                             /* out */ id_d_t& feature_type_id,
                             /* out */ std::string& default_value,
                             /* out */ bool& is_computed);
  // Assumes everything exists 
  virtual bool GetFeatures(const std::string& object_type_name,
			   const std::list<feature_info>& feature_infos,
                           id_d_t object_id_d,
                           /* out */ std::list<std::string>& results); 
  virtual bool GetFeatures(const std::string& object_type_name,
			   const std::list<feature_info>& feature_infos,
                           id_d_t object_id_d,
                           /* out */ std::list<CEMdFValue*>& results); 
  virtual bool GetFeatures(const std::string& object_type_name,
			   const std::list<feature_info>& feature_infos,
			   const std::list<id_d_t> object_id_ds,
			   /* out */ std::list<std::list<std::string> >& results);


private:
  virtual bool GetFeaturesByQuery(const std::list<std::string>& stored_feature_names, 
				  const std::string& object_type_name,
				  const std::list<feature_info>& feature_infos,
				  const std::list<id_d_t> object_id_ds,
				  /* out */ std::list<std::list<std::string> >& results);
  virtual bool GetComputedFeatures(const std::string& object_type_name,
				   const std::list<feature_info>& feature_infos,
				   const std::list<id_d_t> object_id_ds,
				   /* out */ std::list<std::list<std::string> >& results);
  virtual bool GetComputedFeature(const std::string& object_type_name, 
				  const feature_info& fi, 
				  id_d_t object_id_d, 
				/* out */ std::string& comp_result);


public:
                           

  //
  // Segment manipulation
  virtual bool CreateSegment(const std::string& segment_name, monad_m
                              first, monad_m last);
  virtual bool DropSegment(const std::string& segment_name);
#endif // !defined SWIG
  virtual bool ThereAreSegments(/* out */ bool& Result);
  virtual bool SegmentExists(const std::string& segment_name, 
                             /* out */ bool& Result,
			     /* out */ monad_m& first,
			     /* out */ monad_m& last);
#ifndef SWIG
  virtual bool SegmentDoesNotOverlap(monad_m first, monad_m last,
				     /* out */ bool& Result);
  virtual bool MonadsFitInSegment(monad_m first, monad_m last,
				     /* out */ bool& Result);

  //
  // Covered_by and buildable_from manipulation
  virtual bool CreateCoveredByRelationship(id_d_t object_type_id,
                                           id_d_t covered_by_id);
  virtual bool CreateBuildableFromRelationship(id_d_t
                                          object_type_id,  
                                          id_d_t buildable_from_id); 
  virtual bool DropCoveredByRelationship(id_d_t object_type_id,
                                         id_d_t covered_by_id);
  virtual bool DropBuildableFromRelationship(id_d_t object_type_id,   
                                          id_d_t buildable_from_id);  
  virtual bool CoveredByRelationshipExists(id_d_t object_type_id,
                                           id_d_t covered_by_id,
                                           /* out */ bool& result);
  virtual bool BuildableFromRelationshipExists(id_d_t
                                           object_type_id, 
                                           id_d_t buildable_from_id, 
                                           /* out */ bool& result);


  //
  // Upgrading to 1.1.0 and beyond
  bool UpgradeDatabase(const std::string& database_name, int nCurrentVersion);
  bool GetCurrentDatabaseVersion(const std::string& database_name, 
				 std::string& /* out */ strCurrentVersion,
				 int& /* out */ nCurrentVersion,
				 bool& /* out */ bNeedsUpgrading,
				 bool& /* out */ bCanUpgrade);
protected:
  bool UpgradeObjectType(const std::string& object_type_name);
  bool TableExists(std::string table_name, bool& /* out */ bTableExists);
  bool UpgradeDatabaseTo110(const std::string& database_name);
  bool UpgradeDatabaseTo115(const std::string& database_name);
  bool UpgradeDatabaseTo118(const std::string& database_name);
#endif // !defined SWIG
public:

  // min_m and max_m
protected:  
  bool Create_min_max_m_table(void);
  bool Set_min_max_m_from_object_types(bool bBeVerbose);
  bool Set_min_max_m_from_object_type(const std::string& object_type_name, bool bBeVerbose);
  bool Set_max_m(monad_m max_m_candidate, bool bSetUnconditionally = false);
  bool Set_min_m(monad_m min_m_candidate, bool bSetUnconditionally = false);
public:
  bool Get_min_m(monad_m& /* out */ min_m);
  bool Get_max_m(monad_m& /* out */ max_m);
  bool Get_all_m_1(CSet_of_monad_ms& /* out */ all_m_1);

  // Indices
  bool CreateIndicesOnDatabase(const std::string& database_name);
  bool DropIndicesOnDatabase(const std::string& database_name);
  bool CreateIndicesOnObjectType(const std::string& object_type_name);
  bool DropIndicesOnObjectType(const std::string& object_type_name);
  bool CreateIndicesOnObjectType(const std::string& database_name, const std::string& object_type_name);
  bool DropIndicesOnObjectType(const std::string& database_name, const std::string& object_type_name);
    
  //
  // Querying
#ifndef SWIG
  virtual bool GetObjectsHavingMonadsInMonadSet(const std::string object_type_name,
						const CSet_of_monad_ms& monad_ms,
						std::set<id_d_t>& objects_already_seen,					      
						/* out */ CTable& result);
  virtual bool GetObjectsHavingMonadsInMonadSet(const std::string object_type_name,
						const CSet_of_monad_ms& monad_ms,
						/* out */ CTable& result);
  virtual bool GetObjectsInUniverse(const std::string& object_type_name,
				    monad_m first, monad_m last,
				    /* out */ IDDList& Result);
  virtual bool GetInst(const std::string& object_type_name,
		       const CSet_of_monad_ms& Su,
		       const CSet_of_monad_ms& all_m_1,
		       const std::string& pre_query_string,
		       const std::list<feature_info>& features_to_get,
		       /* out */ CInst& Result);
protected:
  virtual bool GetInst(const std::string& object_type_name,	
		       monad_m first_monad, monad_m last_monad,
		       const CSet_of_monad_ms& all_m_1,
		       const std::string& pre_query_string,
		       const std::list<feature_info>& features_to_get,
		       /* out */ CInst& Result);
public:

  virtual bool GetObjectsStartingAtSm(const std::string& object_type_name,
				      monad_m Sm,
				      /* out */ IDDList& Result);
  virtual bool GetObjectsAtM(const std::string& object_type_name,
			     monad_m at_monad,
			     /* out */ IDDList& Result);
  virtual bool GetSOMForObject(const std::string& object_type_name,
			       id_d_t object_id_d,
			       /* out */ CSet_of_monad_ms& Result);
  virtual bool GetMonadsFromID_D(id_d_t object_id_d, 
				 id_d_t object_type_id,
				 /* out */ CSet_of_monad_ms& monads);
  virtual bool GetMonadsFromID_Ds(const std::list<id_d_t>& id_ds, 
				  const std::string &object_type_name,
				  CTable *pOutTable);
  virtual bool GetObjectTypes(/* out */ std::list<std::string>& Result);
  virtual bool GetFeaturesForObjectType(id_d_t object_type_id,
					/* out */ std::list<feature_info>& Result);
  virtual bool GetEnumerations(/* out */ std::list<std::string>& Result);
  virtual bool GetEnumConstants(id_d_t enum_id,
				/* out */ std::list<enum_const_info>& Result);
  virtual bool GetObjectTypesUsingEnumeration(id_d_t enum_id,
					      /* out */ std::list<std::string>& Result);
  virtual bool GetSegments(/* out */ std::list<segment_info>& Result);
  virtual bool GetObjectsWithinMonads(const CSet_of_monad_ms& monads, 
				      const std::string& object_type_name,
				      /* out */ std::list<id_d_t>& result);

  // 
  // Object manipulation
  virtual bool ObjectID_DExistsInType(id_d_t object_id_d, 
				      const std::string& object_type_name,
				      /* out */ bool& Result);
  virtual bool ObjectID_DExists(id_d_t object_id_d, 
				/* out */ bool& Result,
				/* out */ id_d_t& object_type_id);
  virtual bool CreateObject(id_d_t object_id_d,
			    const std::string& object_type_name,
			    const CSet_of_monad_ms& monads,
			    const std::list<feature_info>& features);
  virtual bool UpdateObject(id_d_t object_id_d,
			    const std::string& object_type_name,
			    const std::list<feature_info>& features);
  virtual bool DropObject(id_d_t object_id_d, id_d_t object_type_id, const std::string& object_type_name);
  virtual bool DropObjectsInMonads(const CSet_of_monad_ms& monads, const std::string& object_type_name, id_d_t object_type_id, std::list<id_d_t>& result);


  //
  // Feature name encoding/decoding
  static std::string encode_feature_name(const std::string feature_name);
  static std::string decode_feature_name(const std::string coded_feature_name);
  static std::string encode_feature_name_for_prequery_string(const std::string feature_name);

  //
  // SQL production
  void MakeFeatureSQLType(std::ostringstream& OT_Objects_stream, 
			  const feature_info& fi,
			  bool add_not_null = true);
  std::string MakeMonadConstraintsFromSOM(const CSet_of_monad_ms& monads);
  std::string MakeMonadConstraintsFromSOMMonad_m(const CSet_of_monad_ms& monads);
  std::string MakeMonadConstraintsFromSOMmse_first(const CSet_of_monad_ms& monads);

#endif // !defined SWIG
  // DB maintenance
  bool Vacuum(bool bAnalyze);

  // 
  // Monad manipulation
  //    virtual bool InsertMonads(const MonadSet& ms);
  //    virtual bool DeleteMonads(const MonadSet& ms);

  //
  // Helper functions
#ifndef SWIG
  bool Type_id_to_string(id_d_t type_id, std::string& result);
  bool Type_id_to_TableColumnType(id_d_t type_id, TableColumnType& tc_type, std::string& enum_type);
#endif // !defined SWIG

  // Transactions
  bool BeginTransaction(void);
  bool CommitTransaction(void);
  bool AbortTransaction(void);

  //
  // Error messages
  virtual std::string ErrorMessage(void);
  virtual bool ConnectionOk(void);
  void clear_local_error(void);
  std::string get_local_error(void);

#ifndef SWIG
  void append_local_error(std::string error);
protected:
  virtual bool CreateIndex(const std::string& index_name, const std::string& table_name, const std::list<std::string>& columns);
  bool CreateIndicesOnOTObjects(const std::string& OTN);
  bool CreateIndicesOnOTMonadMs(const std::string& OTN);
  virtual void DropIndex(const std::string& index_name, const std::string& table_name);
  void DropIndicesOnOTObjects(const std::string& object_type_name);
  void DropIndicesOnOTMonadMs(const std::string& object_type_name);
  bool CreateSequenceTables110(bool bSeedTables);
  bool CreateSequenceTables118();
  bool MoveSequenceValue110(int sequence);
  bool MoveSequenceValue118(int sequence);
  void add_enum_to_cache(id_d_t enum_id, const std::string& enum_name);
  void add_object_type_to_cache(id_d_t object_type_id, const std::string& object_type_name);
  bool add_enum_const_to_cache(id_d_t enum_id, const std::string& enum_const_name, long value, bool is_default);
  void add_feature_to_cache(id_d_t object_type_id, const std::string& feature_name, id_d_t feature_type_id, const std::string& default_value, bool is_computed);
  void add_feature_to_cache_if_not_already_there(id_d_t object_type_id, const std::string& feature_name, id_d_t feature_type_id, const std::string& default_value, bool is_computed);
  void delete_object_type_from_cache(id_d_t object_type_id);
#endif // !defined SWIG
};

#ifndef SWIG
extern id_d_t encode_feature_type_id(id_d_t feature_type, long length = 0);
extern void decode_feature_type_id(id_d_t feature_type_id, id_d_t& feature_type, long& length);
extern bool feature_type_id_is_ASCII(id_d_t feature_type_id);
extern bool feature_type_id_is_STRING(id_d_t feature_type_id);
extern std::string feature_info2SQLvalue(const feature_info& fi);
extern std::string NormalizeOTName(const std::string& object_type_name);
#endif




#endif // EMDFDB__H__
