/*
 * mql_object_statements.h
 *
 * Definitions of AST-classes for MQL
 *
 * Ulrik Petersen
 * Created: 11/22-2002
 * Last update: 11/22-2002
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2002  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/




#ifndef MQL_OBJECT_STATEMENTS__H__
#define MQL_OBJECT_STATEMENTS__H__

#include <mql_types.h>
#include <mql_execution_environment.h>

// Helper class: object_creation_specification
class CObjectSpec {
private:
  std::string* m_object_type_name;
  CFeatureAssignment* m_feature_assignments;
  id_d_t m_object_type_id;
public:
  CObjectSpec(std::string* object_type_name,
	      CFeatureAssignment* feature_assignment_list);
  virtual ~CObjectSpec();
  void WeedSelfNotAssigned(CMQL_execution_environment *pEE, bool& bResult);
  void WeedObjectTypeNameNotSpecial(CMQL_execution_environment *pEE, bool& bResult);
  const std::string& get_object_type_name() { return *m_object_type_name; };
  id_d_t get_object_type_id(void) const { return m_object_type_id; };
  CFeatureAssignment* get_feature_assignments() { return m_feature_assignments; };
  bool SymbolObjectTypeExists(CMQL_execution_environment *pEE, bool& bTypeExists);
  bool SymbolAllFeaturesMustBeAssigned(CMQL_execution_environment *pEE);
  bool SymbolAllFeaturesMustExist(CMQL_execution_environment *pEE, bool& bResult);
  bool SymbolNoFeatureMayBeComputed(CMQL_execution_environment *pEE, bool& bResult);
  bool ExecMakeFeatureList(CMQL_execution_environment *pEE, std::list<feature_info>& features);
  bool TypeAssignTypes(CMQL_execution_environment *pEE);
  bool TypeTypeCompatibility(CMQL_execution_environment *pEE, bool& bResult);
};

// Base class for all CREATE OBJECT statements
class CCreateObjectStatement : public CStatement {
protected:
  id_d_t m_with_id_d;
  CObjectSpec* m_object_creation_spec;
  CSet_of_monad_ms m_monads;
public:
  CCreateObjectStatement(CMQL_execution_environment *pEE, 
			 id_d_t with_id_d, 
			 CObjectSpec* object_creation_spec);
  virtual ~CCreateObjectStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Monads(bool& bResult);
  virtual bool Exec();
};

// CREATE OBJECT FROM MONADS
class CCreateObjectFromMonadsStatement : public CCreateObjectStatement {
private:
  CMonadSetElement* m_monad_set;
public:
  CCreateObjectFromMonadsStatement(CMQL_execution_environment *pEE, 
				   CMonadSetElement* monad_set,
				   id_d_t with_id_d,
				   CObjectSpec* object_creation_spec);
  virtual ~CCreateObjectFromMonadsStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Monads(bool& bResult);
  virtual bool Exec();
};

// CREATE OBJECT FROM ID_DS
class CCreateObjectFromID_DsStatement : public CCreateObjectStatement {
private:
  CID_D* m_id_d_list;
public:
  CCreateObjectFromID_DsStatement(CMQL_execution_environment *pEE, 
				  CID_D* id_d_specification,
				  id_d_t with_id_d,
				  CObjectSpec* object_creation_spec);
  virtual ~CCreateObjectFromID_DsStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Monads(bool& bResult);
  virtual bool Exec();
};

// CREATE OBJECT FROM (focus|all|) QUERY
class CCreateObjectFromQueryStatement : public CCreateObjectStatement, public CQueryBase {
public:
  CCreateObjectFromQueryStatement(CMQL_execution_environment *pEE, 
				  eFocusSpec focus, 
				  id_d_t with_id_d,
				  CObjectSpec* object_creation_spec,
				  CMonadSetElement* in_spec, 
				  std::string* using_segment,
				  sheaf_return_type_t sheaf_return_type,
				  CTopograph* query);
  virtual ~CCreateObjectFromQueryStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Exec();
};


// Base class for all UPDATE OBJECTS statements
class CUpdateObjectsStatement : public CStatement {
protected:
  CObjectSpec* m_update_spec;
  std::list<id_d_t> m_update_candidates;
public:
  CUpdateObjectsStatement(CMQL_execution_environment *pEE, 
			  CObjectSpec* update_spec);
  virtual ~CUpdateObjectsStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Exec();
};

// By monads
class CByMonads {
protected:
  CMonadSetElement* m_monad_set;
  CSet_of_monad_ms m_monads;
public:
  CByMonads(CMonadSetElement* monad_set);
  virtual ~CByMonads();
  virtual void MyWeed(CMQL_execution_environment *pEE, bool& bResult);
  virtual bool MyMonads(CMQL_execution_environment *pEE, bool& bResult);
protected:  
  bool GetObjectsByMonads(CMQL_execution_environment *pEE, const std::string& object_type_name, std::list<id_d_t>& objects);
};

// UPDATE OBJECTS BY MONADS
class CUpdateObjectsByMonadsStatement : public CUpdateObjectsStatement, public CByMonads {
public:
  CUpdateObjectsByMonadsStatement(CMQL_execution_environment *pEE, 
				  CMonadSetElement* monad_set,
				  CObjectSpec* update_spec);
  virtual ~CUpdateObjectsByMonadsStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Monads(bool& bResult);
  virtual bool Exec();
};

// UPDATE OBJECTS BY ID_DS
class CUpdateObjectsByID_DsStatement : public CUpdateObjectsStatement {
private:
  CID_D* m_id_d_list;
public:
  CUpdateObjectsByID_DsStatement(CMQL_execution_environment *pEE, 
				 CID_D* id_d_specification,
				 CObjectSpec* update_spec);
  virtual ~CUpdateObjectsByID_DsStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Exec();
};

// UPDATE OBJECTS BY (focus|all|) QUERY
class CUpdateObjectsByQueryStatement : public CUpdateObjectsStatement, public CQueryBase {
public:
  CUpdateObjectsByQueryStatement(CMQL_execution_environment *pEE, 
				 eFocusSpec focus, 
				 CObjectSpec* update_spec,
				 CMonadSetElement* in_spec, 
				 std::string* using_segment,
				 sheaf_return_type_t sheaf_return_type,
				 CTopograph* query);
  virtual ~CUpdateObjectsByQueryStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Exec();
};


// Base class for all DELETE OBJECTS statements
class CDeleteObjectsStatement : public CStatement {
protected:
  std::string* m_object_type_name;
  std::list<id_d_t> m_delete_candidates;
  id_d_t m_object_type_id;
public:
  CDeleteObjectsStatement(CMQL_execution_environment *pEE, 
			  std::string* object_type_name);
  virtual ~CDeleteObjectsStatement();
  virtual bool Symbol(bool& bResult);
  virtual bool Exec();
};

// DELETE OBJECTS BY MONADS
class CDeleteObjectsByMonadsStatement : public CDeleteObjectsStatement, public CByMonads {
public:
  CDeleteObjectsByMonadsStatement(CMQL_execution_environment *pEE, 
				  CMonadSetElement* monad_set,
				  std::string* object_type_name);
  virtual ~CDeleteObjectsByMonadsStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Monads(bool& bResult);
  virtual bool Exec();
protected:
  bool ExecOneObjectType();
};

// DELETE OBJECTS BY ID_DS
class CDeleteObjectsByID_DsStatement : public CDeleteObjectsStatement {
private:
  CID_D* m_id_d_list;
public:
  CDeleteObjectsByID_DsStatement(CMQL_execution_environment *pEE, 
				 CID_D* id_d_specification,
				 std::string* object_type_name);
  virtual ~CDeleteObjectsByID_DsStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Exec();
};

// DELETE OBJECTS BY (focus|all|) QUERY
class CDeleteObjectsByQueryStatement : public CDeleteObjectsStatement, public CQueryBase {
public:
  CDeleteObjectsByQueryStatement(CMQL_execution_environment *pEE, 
				 eFocusSpec focus, 
				 std::string* object_type_name,
				 CMonadSetElement* in_spec, 
				 std::string* using_segment,
				 sheaf_return_type_t sheaf_return_type,
				 CTopograph* query);
  virtual ~CDeleteObjectsByQueryStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Exec();
};




#endif /* MQL_OBJECT_STATEMENTS__H__ */
