/*
 * mql_object_type_statements.h
 *
 * Definitions of AST-classes for MQL
 *
 * Ulrik Petersen
 * Created: 11/22-2002
 * Last update: 11/22-2002
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2002  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/




#ifndef MQL_OBJECT_TYPE_STATEMENTS__H__
#define MQL_OBJECT_TYPE_STATEMENTS__H__

#include <mql_types.h>
#include <mql_execution_environment.h>

// Base class for the OBJECT TYPE statements
class CObjectTypeStatement : public CStatement {
protected:
  std::string* m_object_type_name;
  id_d_t m_object_type_id;
public:
  CObjectTypeStatement(CMQL_execution_environment *pEE, std::string* object_type_name) : CStatement(pEE), m_object_type_name(object_type_name)  {};
  virtual ~CObjectTypeStatement();
protected:
  virtual bool SymbolObjectTypeExists(bool& bObjectTypeExists);
};




// Helper class
class CFeatureDeclaration {
private:
  CFeatureDeclaration* m_next;
  std::string* m_feature_name;
  CType* m_type;
  CExpression* m_default_specification;
  bool m_is_computed;
public:
  CFeatureDeclaration(std::string* feature_name,
		      CType* feature_type,
		      CExpression* default_specification,
		      bool is_computed,
		      CFeatureDeclaration* next);
  virtual ~CFeatureDeclaration();
  CFeatureDeclaration* Next() { return m_next; };
  void set_Next(CFeatureDeclaration* next) { m_next = next; };
  const std::string& get_feature_name() const { return *m_feature_name; };
  CType* get_type() { return m_type; };
  CExpression* get_default_specification() { return m_default_specification; };
  bool get_is_computed() { return m_is_computed; };
  virtual void Weed(CMQL_execution_environment *pEE, bool& bResult);
  bool SymbolEnumerationsExist(CMQL_execution_environment *pEE, bool& bResult);
  bool SymbolEnumConstantsExist(CMQL_execution_environment *pEE, bool& bResult);
  bool SymbolFeaturesMustNotExist(CMQL_execution_environment *pEE, bool& bResult, id_d_t object_type_id);
  bool get_feature_info(CMQL_execution_environment *pEE, feature_info& fi);
  bool TypeAssignTypeID(CMQL_execution_environment *pEE);
  bool TypeTypeCompatibility(CMQL_execution_environment *pEE, bool& bResult);
  bool ExecAdd(CMQL_execution_environment *pEE, const std::string& object_type_name, id_d_t object_type_id);
private:
  void WeedSelfNotDeclared(CMQL_execution_environment *pEE, bool& bResult);
  void WeedStringLength(CMQL_execution_environment *pEE, bool& bResult);
};

typedef enum { kCovered_by, kBuildable_from } eCBBF;


// Helper class
class CObjectTypeName {
private:
  CObjectTypeName* m_next;
  std::string* m_object_type_name;
  id_d_t m_object_type_id;
public:
  CObjectTypeName(std::string* object_type_name, CObjectTypeName* next);
  virtual ~CObjectTypeName();
  const std::string& get_object_type_name(void) { return *m_object_type_name; };
  CObjectTypeName* Next() { return m_next; };
  void set_Next(CObjectTypeName* next) { m_next = next; };
  bool SymbolTypesExist(CMQL_execution_environment *pEE, bool& bResult);
  virtual bool Exec(CMQL_execution_environment *pEE, id_d_t object_type_id, eAddUpdateRemove AddRemove, eCBBF cbbf);
};

class CCBBFList {
private:
  CCBBFList* m_next;
  eAddUpdateRemove m_AddRemove;
  eCBBF m_cbbf;
  CObjectTypeName* m_object_type_name_list;
public:
  CCBBFList(CObjectTypeName* object_type_name_list, 
	    eAddUpdateRemove AddRemove, 
	    eCBBF cbbf, 
	    CCBBFList* next); // only Add or Remove
  virtual ~CCBBFList();
  eAddUpdateRemove get_AddRemove() { return m_AddRemove; };
  eCBBF get_cbbf() { return m_cbbf; };
  CObjectTypeName* get_object_type_name_list() { return m_object_type_name_list; };
  CCBBFList* Next() { return m_next; };
  void set_Next(CCBBFList* next) { m_next = next; };
  bool SymbolTypesExist(CMQL_execution_environment *pEE, bool& bResult);
  virtual bool Exec(CMQL_execution_environment *pEE, const std::string& object_type_name, id_d_t object_type_id);
};

// CREATE OBJECT TYPE
class CCreateObjectTypeStatement : public CObjectTypeStatement {
private:
  CFeatureDeclaration* m_features;
  CCBBFList* m_cbbf_list;
public:
  CCreateObjectTypeStatement(CMQL_execution_environment *pEE, 
			     std::string* object_type_name, 
			     CFeatureDeclaration* feature_declaration_list,
			     CCBBFList* cbbf_list);
  virtual ~CCreateObjectTypeStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Exec();
};


// Helper class feature_update
class CFeatureUpdate {
private:
  CFeatureUpdate* m_next;
  CFeatureDeclaration* m_feature_addition;
  std::string* m_feature_removal;
  eAddUpdateRemove m_AddOrRemove;
public:
  CFeatureUpdate(CFeatureDeclaration* feature_declaration, CFeatureUpdate* next);
  CFeatureUpdate(std::string* feature_name, CFeatureUpdate* next);
  virtual ~CFeatureUpdate();
  eAddUpdateRemove get_kind() { return m_AddOrRemove; };
  const std::string& get_feature_removal() const { assert(m_AddOrRemove == kRemove); return *m_feature_removal; };
  CFeatureDeclaration* get_feature_addition() { assert(m_AddOrRemove == kAdd); return m_feature_addition; };
  CFeatureUpdate* Next() { return m_next; };
  void set_Next(CFeatureUpdate* next) { m_next = next; }
  virtual void Weed(CMQL_execution_environment *pEE, bool& bResult);
  bool SymbolExistenceOfFeatures(CMQL_execution_environment *pEE, bool& bResult, id_d_t object_type_id);
  bool SymbolEnumerations(CMQL_execution_environment *pEE, bool& bResult);
  bool TypeAssignTypeID(CMQL_execution_environment *pEE);
  bool TypeTypeCompatibility(CMQL_execution_environment *pEE, bool& bResult);
  virtual bool Exec(CMQL_execution_environment *pEE, const std::string& object_type_name, id_d_t object_type_id);
};

// UPDATE OBJECT TYPE
class CUpdateObjectTypeStatement : public CObjectTypeStatement {
private:
  CFeatureUpdate* m_features;
  CCBBFList* m_cbbf_list;
public:
  CUpdateObjectTypeStatement(CMQL_execution_environment *pEE, 
			     std::string* object_type_name, 
			     CFeatureUpdate* feature_update_list,
			     CCBBFList* cbbf_list);
  virtual ~CUpdateObjectTypeStatement();
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Exec();
};

// DROP OBJECT TYPE
class CDropObjectTypeStatement : public CObjectTypeStatement {
public:
  CDropObjectTypeStatement(CMQL_execution_environment *pEE, 
			   std::string* object_type_name);
  virtual ~CDropObjectTypeStatement() {};
  virtual bool Symbol(bool& bResult);
  virtual bool Exec();
};




#endif /* MQL_OBJECT_TYPE_STATEMENTS__H__ */
