/*
 * mysqlconn.h
 *
 * MySQL Database connection 
 *
 * Ulrik Petersen
 * Created: 1/27-2001
 * Last update: 6/9-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/




#ifndef EMDF_MYSQLCONN__H__
#define EMDF_MYSQLCONN__H__

#include <debug.h>
#include <utils.h>
#include <conn.h>
#include <cstdlib>
#include <mysql/mysql.h>


class MySQLEMdFConnection : public EMdFConnection {
private:
  MYSQL* pConn;
  MYSQL_RES *pResultSet;
  std::string m_host;
  std::string m_user;
  std::string m_passwd;
  std::string m_db;
  bool bTransactionInProgress;
  unsigned long m_command_rows;
  MYSQL_ROW m_Row;
  unsigned int m_num_fields;
  long m_current_tuple_no;
public:
  // Constructor and destructor
  MySQLEMdFConnection(const std::string& host, const std::string& user, const std::string& passwd, const std::string& db);
  virtual ~MySQLEMdFConnection();

  // Status of connection
  // Returns true on connection good, false on connection bad
  virtual bool ConnectionOk(void);

  // Executing
  virtual bool Exec(const std::string& query);

  // Get last error message
  // Returns true on error message available 
  // (even if it is "OK"), false on error 
  // message not available.
  virtual bool ErrorMessage(std::string& out);

  // Transactions
  virtual bool BeginTransaction(void);
  virtual bool CommitTransaction(void);
  virtual bool AbortTransaction(void);


  // Tuple-counts
  // Both return -1 on not appliccable or error
  virtual long NoOfResultTuples(void);  // SELECT
  virtual long NoOfCommandTuples(void); // INSERT, UPDATE, DELETE

  // Field count
  // Returns -1 on no fields.
  virtual int NoOfFields(void);

  // Tuple-access by field number
  // First tuple is 0
  // First field is 0.
  virtual bool AccessTuple(std::string& out, long tuple_no, int field_no) {
    if (!FetchRow(tuple_no))
      return false;
    else {
      if (m_Row == NULL) {
	// Return failure
	return false;
      } else {
	// Fetch data
	// Check for NULL value
	const char *pField = m_Row[field_no];
	if (pField == NULL) {
	  // It was NULL.  Throw exception
	  throw EMdFNULLValueException();
	} else {
	  // It wasn't NULL.  Convert result
	  out = pField;
	}
	
	// Return success      
	return true;
      }
    }
  };

  virtual bool AccessTuple(long tuple_no, 
			   int field_no, 
			   const char* calling_method,
			   bool& result) {
    if (!this->FetchRow(tuple_no)) {
      DEBUG_ACCESS_TUPLE_FAILED(calling_method, tuple_no, field_no);
      return false;
    } else {
      if (m_Row == NULL) {
	// Return failure
	return false;
      } else {
	// Declare field
	const char *pField = m_Row[field_no];

	// Check for NULL value
	if (pField == NULL) {
	  // It was NULL.  Throw exception
	  throw EMdFNULLValueException();
	} else {
	  // It wasn't NULL.  Convert result
	  result = char2bool(*pField);
	}
	
	// Return success      
	return true;
      }
    }
  };
  virtual bool AccessTuple(long tuple_no, 
			   int field_no, 
			   const char* calling_method,
			   long& result) {
    if (!this->FetchRow(tuple_no)) {
      DEBUG_ACCESS_TUPLE_FAILED(calling_method, tuple_no, field_no);
      return false;
    } else {
      if (m_Row == NULL) {
	// Return failure
	return false;
      } else {
	const char *pField = m_Row[field_no];
	// Convert result
	// NOTE: The following is MUCH slower than strtol or atol.
	//std::string strtmp = pField;
	//std::istringstream mystringstream(strtmp);
	//mystringstream >> result;
	
	// Check for NULL value
	if (pField == NULL) {
	  // It was NULL.  Throw exception
	  throw EMdFNULLValueException();
	} else {
	  // It wasn't NULL.  Convert result
	  result = atol(pField);
	}
	
	// Return success      
	return true;
      }
    }
  };

  virtual bool AccessTuple(long tuple_no, 
			   int field_no, 
			   const char* calling_method,
			   std::string& result) {

    if (!this->FetchRow(tuple_no)) {
      DEBUG_ACCESS_TUPLE_FAILED(calling_method, tuple_no, field_no);
      return false;
    } else {
      const char *pField = m_Row[field_no];

      // Check for NULL value
      if (pField == NULL) {
	// It was NULL.  Throw exception
	throw EMdFNULLValueException();
      } else {
	// It wasn't NULL.  Convert result
	result = pField;
      }
      
      // Return success
      return true;
    }
  };

  

  // Release results so they don't take up memory
  virtual void ReleaseResults(void);

  // UseDatabase
  virtual bool UseDatabase(const std::string& db_name);
protected:
  void FreeResult(void);
private:
  // Make the assignment operator and copy constructor private
  // because they are not supported.
  MySQLEMdFConnection(const MySQLEMdFConnection&);
  MySQLEMdFConnection& operator=(const MySQLEMdFConnection&);
  
  // Fetch row
  bool FetchRow(long tuple_no) {
    if (pConn == 0 || pResultSet == NULL)
      return false;
    else {
      // Fetch row if necessary
      if (m_current_tuple_no != tuple_no) {
	if (m_current_tuple_no != (tuple_no - 1)) {
	mysql_data_seek(pResultSet, tuple_no);
	}
	m_Row = mysql_fetch_row(pResultSet);
	m_current_tuple_no = tuple_no;
      }
      
      if (m_Row == NULL) {
	// Return failure
	return false;
      } else {
	// Return success      
	return true;
      }
    }
  }
};


#endif // EMDF_MYSQLCONN__H__
