/*
 * table.h
 *
 * Tables
 *
 * Ulrik Petersen
 * Created: 5/5-2001 
 * Last update: 6/21-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/




#ifndef TABLE__H__
#define TABLE__H__


#ifndef SWIG
#include <emdf_output.h>
#endif // !defined SWIG

#include <string>

#undef list
#include <list>

#ifdef SWIG
#pragma SWIG nowarn=401
#endif

typedef enum { kTCString, kTCInteger, kTCID_D, kTCMonad_m, kTCBool, kTCEnum } TableColumnType;


#ifndef SWIG
typedef std::list<TableColumnType> TCTList;

#endif

class CTable; // Forward declaration
class TableRow; // Forward declaration

class TableRowIterator {
private:
  friend class CTable;
  TableRow *m_pMotherTableRow;
  std::list<std::string>::iterator m_iter;
public:
  TableRowIterator(TableRow *pMotherTableRow);
  TableRowIterator();
  TableRowIterator(const TableRowIterator& other);
  ~TableRowIterator();
#ifndef SWIG
  TableRowIterator& operator=(const TableRowIterator& other);
#endif 
  bool hasNext() const; // Is the iterator == end iterator?  Doesn't alter iterator
  std::string next(); // Gets current and advances iterator afterwards
  std::string previous(); // Regresses iterator and then gets current
  std::string current(); // Gets current without altering iterator
  std::string getColumn(int column); // Gets column in current without touching iterator
private:
  void assign(const TableRowIterator& other);
};

class TableRow : public std::list<std::string> {
public:
  TableRow();
#ifndef SWIG
  TableRow(std::list<std::string> other);
#endif
  ~TableRow();
  TableRowIterator iterator();  // Gets iterator pointing to beginning
  std::string getColumn(int column); // Gets column
  int size() const;  // Gets number of columns
};

class TableIterator {
private:
  friend class CTable;
  CTable *m_pMotherTable;
  std::list<TableRow>::iterator m_iter;
public:
  TableIterator();
  TableIterator(CTable *pMotherTable);
  TableIterator(const TableIterator& other);
  ~TableIterator();
  bool hasNext() const; // Is the iterator == end iterator?  Doesn't alter iterator
  TableRow& next(); // Gets current and advances iterator afterwards
  TableRow& previous();  // Regresses iterator and then gets current
  TableRow& current(); // Gets current without altering iterator
  std::string next(int column);  // Gets column in current and then advances iterator
  std::string getColumn(int column); // Gets column in current without altering iterator
#ifndef SWIG
  TableIterator& operator=(const TableIterator& other);
#endif
private:
  void assign(const TableIterator& other);
};





class CTable {
private:
  typedef std::list<TableRow> TableRowList;
  friend class TableIterator;
  TableRowList *m_list;
  TableRow *m_header_names;
  TCTList *m_header_types;
  TableRow *m_header_enum_types;
public:
  CTable();
  ~CTable();
  void append(const std::string& str);
#ifndef SWIG
  void append(const std::list<std::string>& l);
#endif // !defined SWIG
  void newline();
  void append_header(std::string header_name, TableColumnType tc_type, std::string enum_type = "");
  void update_row(TableIterator& i, const std::string new_value, int column_index);
  void start_new_row();
  TableIterator iterator();
  TableIterator find(const std::string& str, int column_index, bool bCompareCaseInsensitively = true);
  TableIterator find(const std::string& str1, int column_index1, const std::string& str2, int column_index2, bool bCompareCaseInsensitively = true);
  void erase(TableIterator& d);
  void clear();
  std::string get_column(TableIterator& i, int column_index);
  void PrintConsole(CEMdFOutput *pOut) const;
  void PrintXML(CEMdFOutput *pOut) const;
  static void print_DTD(CEMdFOutput *pOut);
  long size() const; // Get number of rows

private:
  void create_headers(TableRow& headers) const;
  void PrintXML_Header(CEMdFOutput *pOut) const;
  void PrintXML_Rows(CEMdFOutput *pOut) const;
  void PrintXML_Row(CEMdFOutput *pOut, const TableRow& sl) const;
  void get_max_lengths_all(std::list<std::string::size_type>& lengths, const TableRow& headers) const;
  void get_max_lengths(const TableRow& strlist, std::list<std::string::size_type>& lengths) const;
  void dump_results_all(CEMdFOutput* pOut, const std::list<std::string::size_type>& max_lengths, const TableRow& headers) const;
  void dump_results(CEMdFOutput* pOut, const TableRow& strlist, const std::list<std::string::size_type>& max_lengths) const;
  void emit_horizontal_line(CEMdFOutput* pOut, const std::list<std::string::size_type>& max_lengths) const;
};


#endif /* TABLE__H__ */
