/*
 * manage_indices.cpp
 *
 *
 *
 * Tool to manage indices on Emdros databases.  Either creates or
 * drops all indices.  If a "--type object-type" argument is given,
 * then only that object type gets its indices created or dropped.  If
 * no such argument is given, all all object types in the database
 * have their indices created or dropped.
 *
 * Synopsis:
 *   manage_indices --create [options] database-name
 *   manage_indices --drop [options] database-name
 *   manage_indices --help
 *
 * Supports -t, -h, -u, and -p arguments (see output from --help).
 * 
 *
 *
 * Created: 11/13-2001
 * Last update: 4/7-2004
 *
 * Return codes:
 * 
 * 0 : Success
 * 1 : Wrong usage
 * 2 : Connection to backend server could not be established
 * 3 : An exception occurred
 * 5 : Database error
 * */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2004  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <emdfdb.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <assert.h>
#include <infos.h>
#include <emdf_output.h>
#include <default_password.h>


#if USE_POSTGRESQL
#include <pgemdfdb.h>
#elif USE_MYSQL
#include <mysqlemdfdb.h>
#endif

#define TASK_NONE    (0)
#define TASK_CREATE  (1)
#define TASK_DROP    (2)

void show_version(std::ostream& ostr)
{
  ostr << "manage_indices from Emdros version " << EMDROS_VERSION 
       << " on " << EMdFDB::BackendName() << std::endl;
}

void print_usage(std::ostream& ostr)
{
  show_version(ostr);
  ostr << "Usage:" << std::endl;
  ostr << "    manage_indices  --create  [options]  database-name  " << std::endl;
  ostr << "    manage_indices  --drop    [options]  database-name" << std::endl;
  ostr << "    manage_indices  --help" << std::endl;
  ostr << "    manage_indices  (-V|--version)" << std::endl;
  ostr << "OPTIONS:" << std::endl;
  ostr << "   --help               Show this help" << std::endl;
  ostr << "   -V , --version       Show version" << std::endl;
  ostr << "   --type object-type   " << std::endl;
  ostr << "   -h , --host host     Set hostname to connect to" << std::endl;
  ostr << "   -u , --user user     Set database user to connect as (default: 'emdf')" << std::endl;
  ostr << "   -p , --password pwd  Set password to use for database user" << std::endl;

  ostr << "DEFAULTS:" << std::endl;
  ostr << "   -h localhost" << std::endl;
  ostr << "   -u emdf" << std::endl;
}

bool parse_argument(int argc, char* argv[], int &i, std::string& strDatabase, bool& bObjectType, std::string& strObjectType, std::string& hostname, std::string& user, std::string& password)
{
  if (argv[i][0] != '-') {
    if (i == (argc-1)) {
      strDatabase = argv[i];
      return true;
    } else {
      // Databasename must come at the very end
      std::cerr << "ERROR: database name must be at the very end." << std::endl;
      return false;
    }
  } else { // [0] == '-'
    std::string argument(argv[i]);
    if (argument == "-t" || argument == "--type") {
      if (i < (argc-1)) {
	++i;
	strObjectType = argv[i];
	bObjectType = true;
      } else {
	std::cerr << "ERROR: " << argument << " must have an object type as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " word " << std::endl;
	return false;
      }
    } else if (argument == "-h" || argument == "--host") {
      if (i < (argc-1)) {
	++i;
	hostname = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a hostname as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " vlager" << std::endl;
	return false;
      }
    } else if (argument == "-u" || argument == "--user") {
      if (i < (argc-1)) {
	++i;
	user = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a user as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " emdf" << std::endl;
	return false;
      }
    } else if (argument == "-p" || argument == "--password") {
      if (i < (argc-1)) {
	++i;
	password = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a password as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " secret" << std::endl;
	return false;
      }
    } else {
      std::cerr << "ERROR: Unrecognized option " << argument << std::endl;
      return false;
    }
    return true;
  }
}

bool parse_parameters(int argc, char* argv[], int& nTask, std::string& strDatabase, bool& bObjectType, std::string& strObjectType, std::string &hostname, std::string& user, std::string& password)
{
  // Set defaults
  hostname = "localhost";
  bObjectType = false;
  strDatabase = "";
  user = "emdf";

  // Get default password
#if USE_DEFAULT_PASSWORD
  get_default_password(password);
#else
  password = "";
#endif
  
  if (argc == 1) 
    return false;
    

  // Check task
  std::string arg1 = std::string(argv[1]);
  if (arg1 == "--create") {
    nTask = TASK_CREATE;
  } else if (arg1 == "--drop") {
    nTask = TASK_DROP;
  } else if (arg1 == "--help") {
    print_usage(std::cout);
    exit(0);
  } else if (arg1 == "-V" || arg1 == "--version") {
    show_version(std::cout);
    exit(0);
  } else
    return false;

  for (int i=2; i < argc; ++i) {
    if (!parse_argument(argc, argv, i, strDatabase, bObjectType, strObjectType, hostname, user, password)) 
      return false;
  }

  if (strDatabase == "")
    return false;

  // Return success
  return true;
}

int main(int argc, char* argv[])
{
#if USE_POSTGRESQL
  PgEMdFDB* pDB;
#elif USE_MYSQL
  MySQLEMdFDB *pDB;
#endif

  int nTask;
  bool bObjectType;
  std::string strObjectType;
  std::string strDatabase;
  std::string hostname;
  std::string user;
  std::string password;
  if (!parse_parameters(argc, argv, nTask, strDatabase, bObjectType, strObjectType, hostname, user, password)) {
    print_usage(std::cerr);
    return 1;
  }
    
  // Make connection
#if USE_POSTGRESQL
  pDB = new PgEMdFDB(hostname, user, password, strDatabase);
#elif USE_MYSQL
  pDB = new MySQLEMdFDB(hostname, user, password, strDatabase);
#endif

  // Zero-fill password
  zero_fill_string(password);

  // Check that we connected
  if (!pDB->ConnectionOk()) {
    std::cerr << "Connection to backend server could not be established." << std::endl;
    std::cerr << "This usually has one of the following causes:" << std::endl;
    std::cerr << "  - The back-end is not running." << std::endl;
    std::cerr << "  - The back-end is not configured for incoming TCP/IP connections." << std::endl;
    std::cerr << "  - You did not supply a user/password with the -u/-p options." << std::endl;
    std::cerr << "    Note that the default user is 'emdf'." << std::endl;
    std::cerr << "  - You have not followed the instructions in the 'bootstrapping.txt'" << std::endl;
    std::cerr << "    document." << std::endl;
    std::cerr << "There are also hints in the HOW-TO-USE document on configuring your back-end." << std::endl;
    std::cerr << "     Please see      <http://emdros.org/whattodo.html>" << std::endl;
    std::cerr << "     for more information on this problem." << std::endl;
    std::cerr << "Also, please be sure to download the full documentation, available from" << std::endl;
    std::cerr << "<http://emdros.org/>.  You won't be able to make Emdros useful without it." << std::endl;
    delete pDB;
    return 2;
  }

  int nResult = 3;
  try {
    switch (nTask) {
    case TASK_CREATE:
      if (bObjectType) {
	nResult = (pDB->CreateIndicesOnObjectType(strDatabase, strObjectType)) ? 0 : 5;
      } else {
	nResult = (pDB->CreateIndicesOnDatabase(strDatabase)) ? 0 : 5;
      }
      break;
    case TASK_DROP:
      if (bObjectType) {
	nResult = (pDB->DropIndicesOnObjectType(strDatabase, strObjectType)) ? 0 : 5;
      } else {
	nResult = (pDB->DropIndicesOnDatabase(strDatabase)) ? 0 : 5;
      }
      break;
    default:
      assert(false);
      break;
    }
    if (nResult != 0) {
      nResult = 5;
      std::cerr << "ERROR: Database error..." << std::endl;
      std::cerr << pDB->ErrorMessage() << std::endl; 
    }
  } catch (CEMdFDBException e) {
    std::cerr << "ERROR: CEMdFDBException (Database error)..." << std::endl;
    std::cerr << pDB->ErrorMessage() << std::endl; 
  } catch (BadMonadsException e) {
    std::cerr << "BadMonadsException caught.  Program aborted." << std::endl;
  } catch (NoElementsException e) {
    std::cerr << "NoElementsException caught.  Program aborted." << std::endl;
  } catch (CEMdFOutputWrongCharactersetException e) {
    std::cerr << "CEMdFOutputWrongCharactersetException caught.  Program aborted." << std::endl;
  } catch (CEMdFOutputException e) {
    std::cerr << "CEMdFOutputException caught.  Program aborted." << std::endl;
  } catch (...) {
    std::cerr << "Unknown exception occurred.  Program aborted." << std::endl;
  }

  // Clean up
  delete pDB;
  return nResult;
}


