/*
 * mqldump.cpp
 *
 * A tool to dump Emdros databases in MQL
 * Created: 5/1-2001 (1st of May, 2001)
 * Last update: 6/21-2003
 *
 * Return codes:
 * 
 * 0 : Success
 * 1 : Wrong usage
 * 2 : Connection to backend server could not be established
 * 3 : An exception occurred
 * 4 : Could not open file
 * 5 : Database error
 * 6 : Compiler error
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#if HAVE_CONFIG_H
#include <emdros-config.h>
#endif 
#include <emdros-lconfig.h>
#include <emdros_environment.h>
#include <emdfdb.h>
#include <inst.h>
#include <inst_object.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <default_password.h>
#include <list>

typedef struct CDumpEnv {
public:
  std::string filename;
  std::string password;
  std::string hostname;
  std::string user;
  bool bCreateAndUseDB;
  bool bCreateSegments;
  bool bCreateObjectTypes;
  bool bCreateEnums;
  bool bCreateObjectData;
  bool bVerbose;
  std::string db;
  std::list<std::string> object_types;
  CDumpEnv() {
    hostname = "localhost";
    user = "emdf";
    bCreateAndUseDB = true;
    bCreateSegments = true;
    bCreateObjectTypes = true;
    bCreateEnums = true;
    bCreateObjectData = true;
    bVerbose = false;

    // Get default password
#if USE_DEFAULT_PASSWORD
    get_default_password(password);
#else
    password = "";
#endif
  };
  ~CDumpEnv() {};
} CDumpEnv;

bool ExecuteQuery(CEmdrosEnv *pEE, std::string query, bool& bCompilerResult) 
{
  return pEE->execute_string(query, bCompilerResult, false, true);
}

#define EXECUTE_QUERY(query) { \
    if (!ExecuteQuery(pEE, query, bCompilerResult)) { \
      return false; \
    } else { \
      if (!bCompilerResult) { \
	return true; \
      } \
    } \
 }

bool GetObjectTypeId(CEmdrosEnv *pEE, std::string object_type_name, id_d_t& object_type_id)
{
  // Get object type id
  bool bObjectTypeExists;
  if (!pEE->getMQLEE()->pDB->ObjectTypeExists(object_type_name,
					      bObjectTypeExists,
					      object_type_id)) {
    return false;
  } else {
    if (!bObjectTypeExists) {
      std::cerr << "ERROR: Object type " << object_type_name << " does not exist." << std::endl;
      return false;
    }
  }
  
  // If we got this far, there were no DB errors, and the object type existed
  return true;
}

bool GetEnumConst(CEmdrosEnv *pEE, long value, const std::string& enum_name, std::string& outstr)
{
  bool bDBOK = false;
  outstr= pEE->GetEnumConstNameFromValue(value,
					 enum_name,
					 bDBOK);
  if (!bDBOK) {
    std::cerr << "ERROR getting enumeration constant with value " << value << std::endl;
    std::cerr << "from enumeration " << enum_name << std::endl;
    return false;
  }

  // If we got this far, there were no DB errors
  return true;
}

bool GetObjectTypes(CDumpEnv& de, CEmdrosEnv *pEE, bool& bCompilerResult)
{
  // Do we have any object types from the command line?
  if (!de.object_types.empty()) {
    // Yes.  Check that they exist
    std::list<std::string>::const_iterator ci = de.object_types.begin();
    while (ci != de.object_types.end()) {
      // Check to see that OT exists

      bool bObjectTypeExists;
      id_d_t dummy_id_d;

      // Ask DB
      if (!pEE->getMQLEE()->pDB->ObjectTypeExists(*ci, bObjectTypeExists, dummy_id_d)) {
	return false;
      } else {
	// No DB error.
	// Does it exist?
	if (!bObjectTypeExists) {
	  // No. Report error
	  std::cerr << "ERROR: Object type " << *ci << " does not exist." << std::endl;
	  return false;
	}
      }

      // Advance iterator
      ci++;
    }
  } else {
    // No object types from the command line.

    // Get list from DB

    // Execute SELECT OBJECT TYPES query
    std::string query;
    query = "SELECT OBJECT TYPES GO";
    EXECUTE_QUERY(query);

    // Check for sanity of result
    assert(pEE->isTable());

    // Get result table
    CTable *pTable = pEE->takeOverTable();

    // Iterate over result
    TableIterator it = pTable->iterator();
    while (it.hasNext()) {
      // Add contents of column 1
      de.object_types.push_back(it.next(1));
    }

    // Clean up (delete table)
    delete pTable;
  }

  // If we got this far, then all object type names exist in the DB.
  return true;
}


void DumpCreateAndUseDB(CDumpEnv& de, std::ostream& ostr) {
  if (de.bVerbose) {
    std::cerr << "Dumping CREATE DATABASE and USE DATABASE..." << std::endl;
  }
  ostr << "// Create database\n";
  ostr << "CREATE DATABASE " << de.db << " GO\n\n";
  ostr << "// Switch to using database\n";
  ostr << "USE DATABASE " << de.db << " GO\n\n";
}

bool DumpSegments(CDumpEnv& de, CEmdrosEnv *pEE, std::ostream& ostr, bool& bCompilerResult) 
{
  if (de.bVerbose) {
    std::cerr << "Dumping segments..." << std::flush;
  }
  // Build and execute query
  std::string query = "SELECT SEGMENTS GO";
  EXECUTE_QUERY(query);

  // Assert that it is a table
  assert(pEE->isTable());

  // Iterate through table
  CTable *pTable = pEE->takeOverTable();
  TableIterator it = pTable->iterator();
  if (it.hasNext()) {
    ostr << "\n\n";
    ostr << "//////////////////////////////////////////////////// \n";
    ostr << "// \n";
    ostr << "// Create segments\n";
    ostr << "// \n";
    ostr << "//////////////////////////////////////////////////// \n\n";
  }
  while (it.hasNext()) {
    // Get segment info
    std::string segment_name = it.getColumn(1);
    std::string first_monad = it.getColumn(2);
    std::string last_monad = it.getColumn(3);

    // Write info to ostr
    ostr << "// Segment " << segment_name << '\n';
    ostr << "CREATE SEGMENT " << segment_name << '\n';
    ostr << "RANGE = " << first_monad << "-" << last_monad << '\n';
    ostr << "GO\n\n";

    // Advance iterator
    it.next();
  }
  
  // Clean up (delete table)
  delete pTable;

  if (de.bVerbose) {
    std::cerr << "done!" << std::endl;
  }

  // If we got this far, there were no DB errors
  return true;
}

bool DumpEnum(CDumpEnv& de, CEmdrosEnv *pEE, std::string enum_name, std::ostream& ostr, bool& bCompilerResult) 
{
  // Emit message if verbose
  if (de.bVerbose) {
    std::cerr << ">> Dumping enumeration " << enum_name << "..." << std::flush;
  }

  // Build and execute query
  std::string query = std::string("SELECT ENUMERATION CONSTANTS\n")
    + "FROM ENUM " + enum_name + " GO";
  EXECUTE_QUERY(query);

  // Assert that it is a table
  assert(pEE->isTable());

  // Emit pre-constant material
  ostr << "// Enumeration " << enum_name << '\n';
  ostr << "CREATE ENUMERATION " << enum_name << " = {\n";

  // Iterate through table
  CTable *pTable = pEE->takeOverTable();
  TableIterator it = pTable->iterator();

  while (it.hasNext()) {
    // Get enum constant info info
    std::string enum_const_name = it.getColumn(1);
    long value = string2long(it.getColumn(2));
    bool is_default = string2bool_alpha(it.getColumn(3));

    // Write info to ostr
    ostr << "  ";
    if (is_default) {
      ostr << "DEFAULT ";
    }
    ostr << enum_const_name << " = " << value;

    // Advance iterator
    it.next();

    // If there are more, then emit ","
    if (it.hasNext()) {
      ostr << ",";
    }

    // Emit trailing newline
    ostr << '\n';
  }
   
  // Emit post-constant material
  ostr << "}\n" << "GO\n\n";

  // Clean up (delete table)
  delete pTable;

  if (de.bVerbose) {
    std::cerr << "done!" << std::endl;
  }

  // If we got this far, there were no DB errors
  return true;
}

bool DumpEnums(CDumpEnv& de, CEmdrosEnv *pEE, std::ostream& ostr, bool& bCompilerResult) 
{
  // Emit message if verbose
  if (de.bVerbose) {
    std::cerr << "Dumping enumerations " << std::endl;
  }

  // Build and execute query
  std::string query = "SELECT ENUMERATIONS GO";
  EXECUTE_QUERY(query);

  // Assert that it is a table
  assert(pEE->isTable());

  // Iterate through table
  CTable *pTable = pEE->takeOverTable();
  TableIterator it = pTable->iterator();
  if (it.hasNext()) {
    ostr << "\n\n";
    ostr << "//////////////////////////////////////////////////// \n";
    ostr << "// \n";
    ostr << "// Create enumerations\n";
    ostr << "// \n";
    ostr << "//////////////////////////////////////////////////// \n\n";
  }
  while (it.hasNext()) {
    // Get enumeration name
    std::string enum_name = it.getColumn(1);

    if (!DumpEnum(de, pEE, enum_name, ostr, bCompilerResult)) {
      // DB error
      if (de.bVerbose) {
	std::cerr << "failure!" << std::endl;
      }
      return false;
    } else {
      // No DB error.  Compiler error?
      if (!bCompilerResult) {
	if (de.bVerbose) {
	  std::cerr << "failure!" << std::endl;
	}
	return true;
      }
    }


    // Advance iterator
    it.next();
  }

  // Clean up (delete table)
  delete pTable;

  if (de.bVerbose) {
    std::cerr << "Done dumping enums!" << std::endl;
  }

  // If we got this far, there were no DB errors
  return true;
}

bool DumpObjectType(CDumpEnv& de, CEmdrosEnv *pEE, std::string object_type_name, std::ostream& ostr, bool& bCompilerResult) 
{
  // Emit message if verbose
  if (de.bVerbose) {
    std::cerr << ">> Dumping object type " << object_type_name << "..." << std::flush;
  }

  // Get object type id
  id_d_t object_type_id;
  if (!GetObjectTypeId(pEE, object_type_name, object_type_id)) {
    // Error has already been reported
    return false;
  }

  // Get features
  std::list<feature_info> feature_infos;
  if (!pEE->getMQLEE()->pDB->GetFeaturesForObjectType(object_type_id,
						      feature_infos)) {
    std::cerr << std::endl << "ERROR: Could not get features for object type " << object_type_name << std::endl;
    return false;
  }

  // Emit pre-feature material
  ostr << "// Object type " << object_type_name << '\n';
  ostr << "CREATE OBJECT TYPE \n";
  ostr << "[" << object_type_name << '\n';

  // Iterate through list of feature_infos
  std::list<feature_info>::const_iterator ci(feature_infos.begin());
  std::list<feature_info>::const_iterator cend(feature_infos.end());
  while (ci != cend) {
    // Name
    std::string feature_name = ci->get_name();

    // Type
    id_d_t feature_type_id = ci->get_type();
    std::string strType_id;
    bool bTypeIDGottenOK = pEE->getMQLEE()->pDB->Type_id_to_string(feature_type_id, strType_id);
    if (!bTypeIDGottenOK) {
      std::cerr << std::endl << "ERROR: Could not get string corresponding to feature " << feature_name << std::endl;
      std::cerr << "from object type " << object_type_name << std::endl;
      return false;
    }

    // Default value
    std::string defval;
    std::string defval_unencoded = ci->get_default_value();

    // Decode feature type id
    id_d_t ft;
    long dummy_length;
    decode_feature_type_id(feature_type_id, ft, dummy_length);

    // Get default value based on feature type
    switch (ft) {
    case FEATURE_TYPE_INTEGER:
    case FEATURE_TYPE_ID_D:
      defval = defval_unencoded;
      break;
    case FEATURE_TYPE_ASCII:
    case FEATURE_TYPE_STRING:
      // This is a STRING or ASCII string
      defval = std::string("\"") + encodeSTRINGstring(defval_unencoded) + "\"";
      break;
    default: // Enumerations
      defval = defval_unencoded;
    }

    // Is_computed: Only added here for future use
    // bool is_computed = string2bool(bool_alpha2string(ci->get_is_computed()));

    // Emit feature if it isn't self
    if (feature_name != "self") {
      ostr << "  " << feature_name 
	   << " : " << strType_id
	   << " DEFAULT " << defval
	   << ";\n";
    }

    // Advance iterator
    ++ci;
  }

  // Emit post-feature material
  ostr << "]\n" << "GO\n\n";

  // Emit trailing "done!" if verbose
  if (de.bVerbose) {
    std::cerr << "done!" << std::endl;
  }

  // If we got this far, there were no DB errors
  return true;
}


bool DumpObjectTypes(CDumpEnv& de, CEmdrosEnv *pEE, std::ostream& ostr, bool& bCompilerResult) 
{
  // Emit message if verbose
  if (de.bVerbose) {
    std::cerr << "Dumping object types..." << std::endl;
  }

  // Emit pre-object types material
  ostr << "\n\n";
  ostr << "//////////////////////////////////////////////////// \n";
  ostr << "// \n";
  ostr << "// Create object types\n";
  ostr << "// \n";
  ostr << "//////////////////////////////////////////////////// \n\n";

  // Iterate through de.object_types
  std::list<std::string>::const_iterator ci;
  ci = de.object_types.begin();
  while (ci != de.object_types.end()) {
    if (!DumpObjectType(de, pEE, *ci, ostr, bCompilerResult)) {
      // DB error
      if (de.bVerbose) {
	std::cerr << "failure!" << std::endl;
      }

      return false;
    } else {
      // No DB error. Compiler error?
      if (!bCompilerResult) {
	// Emit trailing "failure!" if verbose
	if (de.bVerbose) {
	  std::cerr << "failure!" << std::endl;
	}

	return true;
      }
    }
    ci++;
  }
  
  // Emit trailing "done!" if verbose
  if (de.bVerbose) {
    std::cerr << "Done dumping object types!" << std::endl;
  }


  // If we got this far, there were no DB errors
  return true;
}

bool DumpObjectDataSingleObject(CDumpEnv& de, CEmdrosEnv *pEE, std::string object_type_name, const std::list<feature_info>& feature_infos, const CInstObject *pObject, std::ostream& ostr, bool& bCompilerResult)
{
  // Emit pre-feature material
  ostr << "CREATE OBJECT\n";
  ostr << "FROM MONADS = ";
  ostr << pObject->get_monads().ToString() << '\n';
  ostr << "WITH ID_D = " << id_d2string(pObject->get_id_d()) << '\n';
  ostr << "[" << object_type_name << '\n';

  // Iterate through list of feature_infos
  std::list<feature_info>::const_iterator ci(feature_infos.begin());
  std::list<feature_info>::const_iterator cend(feature_infos.end());
  unsigned int feature_index = 0;
  while (ci != cend) {
    // Feature name
    std::string feature_name = ci->get_name();
    
    // EMdFValue
    const CEMdFValue *pValue = pObject->get_feature(feature_index);

    // Sanity check
    assert(pValue != 0);

    // Type
    id_d_t feature_type_id = ci->get_type();
    std::string strType_id;
    bool bTypeIDGottenOK = pEE->getMQLEE()->pDB->Type_id_to_string(feature_type_id, strType_id);
    if (!bTypeIDGottenOK) {
      std::cerr << "ERROR: Could not get string corresponding to feature " << feature_name << std::endl;
      std::cerr << "from object type " << object_type_name << std::endl;
      return false;
    }

    // Decode feature type id
    id_d_t ft;
    long dummy_length;
    decode_feature_type_id(feature_type_id, ft, dummy_length);

    // Get value-string based on feature type
    std::string value_as_string;
    switch (ft) {
    case FEATURE_TYPE_INTEGER:
      assert(pValue->get_kind() == kEVInt);
      value_as_string = long2string(pValue->get_int());
      break;
    case FEATURE_TYPE_ID_D:
      assert(pValue->get_kind() == kEVID_D);
      value_as_string = id_d2string(pValue->get_id_d());
      break;
    case FEATURE_TYPE_ASCII:
    case FEATURE_TYPE_STRING:
      // This is a STRING or ASCII string
      assert(pValue->get_kind() == kEVString);
      value_as_string = std::string("\"") + encodeSTRINGstring(pValue->get_string()) + "\"";
      break;
    default: // Enumerations
      assert(pValue->get_kind() == kEVEnum);
      if (!GetEnumConst(pEE, pValue->get_enum(), strType_id, value_as_string)) {
	return false;
      }
      break;
    }

    // Emit feature if it isn't self
    if (feature_name != "self") {
      ostr << "  " << feature_name 
	   << " := " << value_as_string
	   << ";\n";
    }

    // Advance iterators
    ++ci;
    ++feature_index;
  }

  // Emit post-object-data material
  ostr << "]\n" << "GO\n\n";

  // If we got this far, there were no errors
  return true;
}

bool DumpObjectDataForObjectType(CDumpEnv& de, CEmdrosEnv *pEE, std::string object_type_name, std::ostream& ostr, bool& bCompilerResult)
{
  // Emit message if verbose
  if (de.bVerbose) {
    std::cerr << ">> Dumping object data for object type " << object_type_name << "..." << std::flush;
  }

  // Emit pre-object data material
  ostr << "\n\n";
  ostr << "//////////////////////////////////////////////////// \n";
  ostr << "// \n";
  ostr << "// Create object data for object type " << object_type_name << '\n';
  ostr << "// \n";
  ostr << "//////////////////////////////////////////////////// \n\n";

  // Get object type id
  id_d_t object_type_id;
  if (!GetObjectTypeId(pEE, object_type_name, object_type_id)) {
    // Error has already been reported
    return false;
  }

  // Get features
  std::list<feature_info> feature_infos;
  if (!pEE->getMQLEE()->pDB->GetFeaturesForObjectType(object_type_id,
						      feature_infos)) {
    std::cerr << std::endl << "ERROR: Could not get features for object type " << object_type_name << std::endl;
    return false;
  }

  // Transform list to vector
  std::vector<std::string> feature_name_vec;
  std::list<feature_info>::const_iterator ci(feature_infos.begin());
  std::list<feature_info>::const_iterator cend(feature_infos.end());
  while (ci != cend) {
    // Name
    std::string feature_name = ci->get_name();

    // Push back in feature_name_vec
    feature_name_vec.push_back(feature_name);

    // Advance iterator
    ++ci;
  }

  // Get objects
  std::string pre_query_string = "";
  CSet_of_monad_ms all_m_1;
  CSet_of_monad_ms Su;
  if (!pEE->Get_all_m_1(all_m_1)) {
    std::cerr << std::endl << "ERROR: Could not get all_m_1." << std::endl;
    return false;
  }
  Su = all_m_1;
  CInst *pInst = new CInst(feature_name_vec);

  if (!pEE->getMQLEE()->pDB->GetInst(object_type_name,
				     Su,
				     all_m_1,
				     pre_query_string,
				     feature_infos,
				     *pInst)) {
    std::cerr << std::endl << "ERROR: DB error in getting list of objects for object type " << object_type_name << std::endl;
    delete pInst;
    return false;
  }

  // Iterate over inst
  pInst->set_current_universe(all_m_1);
  CInst::const_iterator inst_cit = pInst->begin();
  int object_count = 0;
  while (inst_cit != pInst->end()) {
    // If verbose, emit "." for every 5000 objects
    if (de.bVerbose) {
      object_count++;
      if ((object_count % 5000) == 0) {
	std::cerr << "." << std::flush;
      }
    }

    // Get object pointer
    const CInstObject *pObject = inst_cit->second;

    // Process object
    if (!DumpObjectDataSingleObject(de, pEE, object_type_name, feature_infos, 
				    pObject, ostr, bCompilerResult)) {
      delete pInst;
      if (de.bVerbose) {
	std::cerr << "failure!" << std::endl;
      }
      return false;
    } else {
      if (!bCompilerResult) {
	delete pInst;
	if (de.bVerbose) {
	  std::cerr << "failure!" << std::endl;
	}
	return true;
      }
    }

    // Advance iterator
    inst_cit = pInst->next(inst_cit);
  }

  // Clean up pInst
  delete pInst;

  // Emit message if verbose
  if (de.bVerbose) {
    std::cerr << "done!" << std::endl;
  }

  // If we got this far, there were no DB errors
  return true;
}

bool DumpObjectData(CDumpEnv& de, CEmdrosEnv *pEE, std::ostream& ostr, bool& bCompilerResult) 
{
  // Emit message if verbose
  if (de.bVerbose) {
    std::cerr << "Dumping object data..." << std::endl;
  }
  
  // Emit pre-object data material
  ostr << "\n\n";
  ostr << "//////////////////////////////////////////////////// \n";
  ostr << "// \n";
  ostr << "// Create object data\n";
  ostr << "// \n";
  ostr << "//////////////////////////////////////////////////// \n\n";

  // Iterate through de.object_types
  std::list<std::string>::const_iterator ci;
  ci = de.object_types.begin();
  while (ci != de.object_types.end()) {
    if (!DumpObjectDataForObjectType(de, pEE, *ci, ostr, bCompilerResult)) {
      // DB error
      return false;
    } else {
      // No DB error. Compiler error?
      if (!bCompilerResult) {
	return true;
      }
    }
    ci++;
  }

  // Emit message if verbose
  if (de.bVerbose) {
    std::cerr << "Done dumping object data!" << std::endl;
  }
  
  // If we got this far, there were no DB errors
  return true;
}


bool do_dump(CDumpEnv& de, CEmdrosEnv *pEE, std::ostream& ostr, bool& bCompilerResult) 
{

  // Find out which object types to treat
  if (!GetObjectTypes(de, pEE, bCompilerResult)) {
    return false;
  } else {
    if (!bCompilerResult) {
      return true;
    }
  }

  // Dump CREATE DATABASE and USE DATABASE
  if (de.bCreateAndUseDB) {
    DumpCreateAndUseDB(de, ostr);
  }

  // Dump segments
  if (de.bCreateSegments) {
    if (!DumpSegments(de, pEE, ostr, bCompilerResult)) {
      return false;
    } else {
      if (!bCompilerResult) {
	return true;
      }
    }
  }

  // Dump enums
  if (de.bCreateEnums) {
    if (!DumpEnums(de, pEE, ostr, bCompilerResult)) {
      return false;
    } else {
      if (!bCompilerResult) {
	return true;
      }
    }
  }

  // Dump object types
  if (de.bCreateObjectTypes) {
    if (!DumpObjectTypes(de, pEE, ostr, bCompilerResult)) {
      return false;
    } else {
      if (!bCompilerResult) {
	return true;
      }
    }
  }

  // Dump object data
  if (de.bCreateObjectData) {
    if (!DumpObjectData(de, pEE, ostr, bCompilerResult)) {
      return false;
    } else {
      if (!bCompilerResult) {
	return true;
      }
    }
  }

  // If we got this far, there were no DB errors
  return true;
}

int dump_to_stream(CDumpEnv& de, CEmdrosEnv *pEE, std::ostream& ostr) 
{
  bool bResult = true;
  std::string strError;
  int nResult;
  if (!do_dump(de, pEE, ostr, bResult)) {
    std::cerr << "FAILURE: Database error dumping." << std::endl;
    nResult = 5;
  } else {
    if (!bResult) {
      std::cerr << "FAILURE: Compiler error dumping." << std::endl;
      nResult = 6;
    } else { 
      std::cerr << "SUCCESS dumping." << std::endl;
      nResult = 0;
    }
  }

  // Return result
  return nResult;
}

int dump_to_file(CDumpEnv& de, CEmdrosEnv *pEE, std::string filename)
{
  std::string strError;
  std::ofstream fout;
  int nResult;
  fout.open(filename.c_str());
  if (!fout) {
    std::cerr << "FAILURE: Could not open file " << filename << " for writing." << std::endl;
    std::cerr << "Action aborted." << std::endl;
    nResult = 4;
  } else {
    nResult = dump_to_stream(de, pEE, fout);
  }

  // Return result
  return nResult;
}


int dump_to_cout(CDumpEnv& de, CEmdrosEnv *pEE)
{
  // Return result
  return dump_to_stream(de, pEE, std::cout);
}

void show_version(std::ostream& ostr)
{
  ostr << "mqldump from Emdros version " << EMDROS_VERSION 
       << " on " << EMdFDB::BackendName() << std::endl;
}

void print_usage(std::ostream& ostr)
{
  show_version(ostr);
  ostr << "Usage:" << std::endl;
  ostr << "   mqldump [options] database [object type names...] " << std::endl;
  ostr << "   mqldump (--help|-V|--version)" << std::endl;
  ostr << "OPTIONS:" << std::endl;
  ostr << "   --help               Show this help" << std::endl;
  ostr << "   -V , --version       Show version" << std::endl;
  ostr << "   -v , --verbose       Be verbose" << std::endl;
  ostr << "   -o filename          Dump to file rather than stdout" << std::endl;
  ostr << "   -h , --host host     Set hostname to connect to" << std::endl;
  ostr << "   -u , --user user     Set database user to connect as (default: 'emdf')" << std::endl;
  ostr << "   -p , --password pwd  Set password to use for database user" << std::endl;
  ostr << "   -n , --no-create-db            Don't issue statements to create and use db" << std::endl;
  ostr << "   -i , --no-create-info          Don't issue statements to create segments," << std::endl;
  ostr << "                                  object types, and enums (i.e., meta-data)." << std::endl;
  ostr << "   -s , --no-create-segments      Don't issue statements to create segments" << std::endl;
  ostr << "   -t , --no-create-object-types  Don't issue statements to create object" << std::endl
       << "                                  types" << std::endl;
  ostr << "   -e , --no-create-enums         Don't issue statements to create enums" << std::endl;
  ostr << "   -d , --no-create-objects       Don't issue statements to create" << std::endl 
       << "                                  object data" << std::endl;
  ostr << "DEFAULTS:" << std::endl;
  ostr << "   -h localhost" << std::endl;
  ostr << "   -u emdf" << std::endl;
}

bool parse_argument(int argc, char* argv[], int &i, bool& bDBHasBeenMet, bool& bCreateAndUseDB, bool& bCreateSegments, bool& bCreateObjectTypes, bool& bCreateEnums, bool& bCreateObjectData, bool &bVerbose, std::string& db, std::list<std::string>& object_types, std::string& filename, std::string& hostname, std::string& user, std::string& password)
{
  std::string argument(argv[i]);
  if (argv[i][0] != '-') {
    if (bDBHasBeenMet) {
      // DB name has been seen.  This is an object type.
      object_types.push_back(argument);
      return true;
    } else {
      // DB name has not been seen. This is the DB name.
      bDBHasBeenMet = true;
      db = argument;
      return true;
    }
  } else { // [0] == '-'
    if (bDBHasBeenMet) {
      // No options after DB
      std::cerr << "ERROR: no options allowed after DB name." << std::endl;
      return false;
    }
    if (argument == "--help") {
      print_usage(std::cout);
      exit(0);
    } else if (argument == "-V" || argument == "--version") {
      show_version(std::cout);
      exit(0);
    } else if (argument == "-v" || argument == "--verbose") {
      bVerbose = true;
    } else if (argument == "-n" || argument == "--no-create-db") {
      bCreateAndUseDB = false;
    } else if (argument == "-i" || argument == "--no-create-info") {
      bCreateSegments = false;
      bCreateEnums = false;
      bCreateObjectTypes = false;
    } else if (argument == "-s" || argument == "--no-create-segments") {
      bCreateSegments = false;
    } else if (argument == "-e" || argument == "--no-create-enums") {
      bCreateEnums = false;
    } else if (argument == "-t" || argument == "--no-create-object-types") {
      bCreateObjectTypes = false;
    } else if (argument == "-d" || argument == "--no-create-objects") {
      bCreateObjectData = false;
    } else if (argument == "-o") {
      if (i < (argc-1)) {
	++i;
	filename = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a filename as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " db_dump.mql" << std::endl;
	return false;
      }
    } else if (argument == "-h"
	       || argument == "--host") {
      if (i < (argc-1)) {
	++i;
	hostname = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a hostname as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " vlager" << std::endl;
	return false;
      }
    } else if (argument == "-u" || argument == "--user") {
      if (i < (argc-1)) {
	++i;
	user = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a user as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " emdf" << std::endl;
	return false;
      }
    } else if (argument == "-p" || argument == "--password") {
      if (i < (argc-1)) {
	++i;
	password = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a password as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " secret" << std::endl;
	return false;
      }
    } else {
      std::cerr << "ERROR: Unrecognized option " << argument << std::endl;
      return false;
    }
    return true;
  }
}

bool parse_arguments(int argc, char* argv[], bool& bCreateAndUseDB, bool& bCreateSegments, bool& bCreateObjectTypes, bool& bCreateEnums, bool& bCreateObjectData, bool &bVerbose, std::string& db, std::list<std::string>& object_types, std::string& filename, std::string& hostname, std::string& user, std::string& password)
{
  bool bDBHasBeenMet = false;
  for (int i=1; i < argc; ++i) {
    if (!parse_argument(argc, argv, i, bDBHasBeenMet, bCreateAndUseDB, bCreateSegments, bCreateObjectTypes, bCreateEnums, bCreateObjectData, bVerbose, db, object_types, filename, hostname, user, password)) {
      return false;
    }
  }
  
  return true;
}

int main(int argc, char* argv[])
{
  // Set defaults
  eOutputKind output_kind = kOKConsole;
  CDumpEnv de;


  // Parse arguments
  if (!parse_arguments(argc, argv, 
		       de.bCreateAndUseDB, de.bCreateSegments, 
		       de.bCreateObjectTypes, de.bCreateEnums, 
		       de.bCreateObjectData, de.bVerbose,
		       de.db, de.object_types, de.filename, 
		       de.hostname, de.user, de.password)) {
    print_usage(std::cerr);
    return 1;
  }

  // Make connection
  CEmdrosEnv *pEE = new CEmdrosEnv(output_kind, kCSISO_8859_1,
				   de.hostname,
				   de.user, de.password,
				   de.db);
  // Zero-fill password
  zero_fill_string(de.password);

  // Check that we connected
  if (!pEE->ConnectionOk()) {
    std::cerr << "Connection to backend server could not be established." << std::endl;
    std::cerr << "This usually has one of the following causes:" << std::endl;
    std::cerr << "  - The back-end is not running." << std::endl;
    std::cerr << "  - The back-end is not configured for incoming TCP/IP connections." << std::endl;
    std::cerr << "  - You did not supply a user/password with the -u/-p options." << std::endl;
    std::cerr << "    Note that the default user is 'emdf'." << std::endl;
    std::cerr << "  - You have not followed the instructions in the 'bootstrapping.txt'" << std::endl;
    std::cerr << "    document." << std::endl;
    std::cerr << "There are also hints in the HOW-TO-USE document on configuring your back-end." << std::endl;
    std::cerr << "     Please see      <http://emdros.org/whattodo.html>" << std::endl;
    std::cerr << "     for more information on this problem." << std::endl;
    std::cerr << "Also, please be sure to download the full documentation, available from" << std::endl;
    std::cerr << "<http://emdros.org/>.  You won't be able to make Emdros useful without it." << std::endl;
    delete pEE;
    return 2;
  }

  int nResult = 3;
  try {
    if (de.filename == "") {
      nResult = dump_to_cout(de, pEE);
    } else {
      nResult = dump_to_file(de, pEE, de.filename);
    }
    if (nResult != 0) {
      std::cerr << pEE->get_DB_error() << std::endl; 
      std::cerr << pEE->get_compiler_error() << std::endl;
    }
  } catch (CEMdFDBException e) {
    std::cerr << "ERROR: CEMdFDBException (Database error)..." << std::endl;
    std::cerr << pEE->get_DB_error() << std::endl; 
    std::cerr << pEE->get_compiler_error() << std::endl;
  } catch (BadMonadsException e) {
    std::cerr << "BadMonadsException caught.  Program aborted." << std::endl;
  } catch (NoElementsException e) {
    std::cerr << "NoElementsException caught.  Program aborted." << std::endl;
  } catch (CMQLSymbolTableEntryException e) {
    std::cerr << "CMQLSymbolTableEntryException caught.  Program aborted." << std::endl;
  } catch (CMQLSymbolTableException e) {
    std::cerr << "CMQLSymbolTableException caught.  Program aborted." << std::endl;
  } catch (CEMdFOutputWrongCharactersetException e) {
    std::cerr << "CEMdFOutputWrongCharactersetException caught.  Program aborted." << std::endl;
  } catch (CEMdFOutputException e) {
    std::cerr << "CEMdFOutputException caught.  Program aborted." << std::endl;
  } catch (...) {
    std::cerr << "Unknown exception occurred.  Program aborted." << std::endl;
  }

  // Clean up
  delete pEE;
  return nResult;
}


