/*
 * Upgrade_db.cpp
 *
 *
 *
 * Tool to upgrade an Emdros database created with version 1.1.4 or 
 * earlier to one that works with Emdros release 1.1.8 and later.
 *
 * Run it with the name of the database to upgrade as the last
 * argument.
 * 
 * Supports -h, -u, and -p arguments (see output from --help).
 *
 * For example:
 *    ./upgrade_db book
 * 
 *    ./upgrade_db --host isaiah.somedomain.com mydb
 * 
 *
 *
 * Created: 11/11-2001
 * Last update: 4/7-2004
 *
 * Return codes:
 * 
 * 0 : Success
 * 1 : Wrong usage
 * 2 : Connection to backend server could not be established
 * 3 : An exception occurred
 * 4 : We could not upgrade
 * 5 : Database error
 * */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2004  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <emdfdb.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <infos.h>
#include <emdf_output.h>
#include <default_password.h>


#if USE_POSTGRESQL
#include <pgemdfdb.h>
#elif USE_MYSQL
#include <mysqlemdfdb.h>
#endif

void show_version(std::ostream& ostr)
{
  ostr << "upgrade_db from Emdros version " << EMDROS_VERSION 
       << " on " << EMdFDB::BackendName() << std::endl;
}

void print_usage(std::ostream& ostr)
{
  show_version(ostr);
  ostr << "Usage:" << std::endl;
  ostr << "   upgrade_db [options] databasename " << std::endl;
  ostr << "   upgrade_db --help" << std::endl;
  ostr << "   upgrade_db (-V|--version)" << std::endl;
  ostr << "OPTIONS:" << std::endl;
  ostr << "   --help               Show this help" << std::endl;
  ostr << "   -V , --version       Show version" << std::endl;
  ostr << "   -h , --host host     Set hostname to connect to" << std::endl;
  ostr << "   -u , --user user     Set database user to connect as (default: 'emdf')" << std::endl;
  ostr << "   -p , --password pwd  Set password to use for database user" << std::endl;
  ostr << "DEFAULTS:" << std::endl;
  ostr << "   -h localhost" << std::endl;
  ostr << "   -u emdf" << std::endl;
}

bool show_warning(void)
{
  std::cout << std::endl;
  std::cout << "upgrade_db:" << std::endl;
  std::cout << "  This program upgrades an Emdros database created with a version" << std::endl;
  std::cout << "  of Emdros prior to 1.1.8 to a database that can be used with Emdros" << std::endl;
  std::cout << "  version 1.1.8 and later." << std::endl;
  std::cout << std::endl;
  std::cout << "  WARNING: This software may make a lot of changes to the database." << std::endl;
  std::cout << "           It is recommended that you backup your database before proceeding." << std::endl;
  std::cout << std::endl;
  std::cout << "  Are you sure you want to proceed? (y/n)" << std::endl;
  char c;
  std::cin >> c;
  return toupper(c) == 'Y';
}

bool parse_argument(int argc, char* argv[], int &i, std::string& database_to_upgrade, std::string& hostname, std::string& user, std::string& password)
{
  if (argv[i][0] != '-') {
    if (i == (argc-1)) {
      database_to_upgrade = argv[i];
      return true;
    } else {
      // Database must come at the very end
      std::cerr << "ERROR: database to upgrade must be at the very end." << std::endl;
      return false;
    }
  } else { // [0] == '-'
    std::string argument(argv[i]);
    if (argument == "--help") {
      print_usage(std::cout);
      exit(0);
    } else if (argument == "-V" || argument == "--version") {
      show_version(std::cout);
      exit(0);
    } else if (argument == "-h"	|| argument == "--host") {
      if (i < (argc-1)) {
	++i;
	hostname = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a hostname as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " vlager" << std::endl;
	return false;
      }
    } else if (argument == "-u" || argument == "--user") {
      if (i < (argc-1)) {
	++i;
	user = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a user as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " emdf" << std::endl;
	return false;
      }
    } else if (argument == "-p" || argument == "--password") {
      if (i < (argc-1)) {
	++i;
	password = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a password as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " secret" << std::endl;
	return false;
      }
    } else {
      std::cerr << "ERROR: Unrecognized option " << argument << std::endl;
      return false;
    }
    return true;
  }
}

bool parse_arguments(int argc, char* argv[], std::string& database_to_upgrade, std::string& hostname, std::string& user, std::string& password)
{
  // Set defaults
  database_to_upgrade = "";
  hostname = "localhost";
  user = "emdf";

  // Get default password
#if USE_DEFAULT_PASSWORD
  get_default_password(password);
#else
  password = "";
#endif
  
  for (int i=1; i < argc; ++i) {
    if (!parse_argument(argc, argv, i, database_to_upgrade, hostname, user, password)) 
      return false;
  }

  if (database_to_upgrade == "") {
    return false;
  }

  return true;
}

int main(int argc, char* argv[])
{
#if USE_POSTGRESQL
  PgEMdFDB* pDB;
#elif USE_MYSQL
  MySQLEMdFDB *pDB;
#endif

  // Declare option-strings
  std::string database_to_upgrade;
  std::string password;
  std::string hostname;
  std::string user;

  // Parse arguments
  if (!parse_arguments(argc, argv, database_to_upgrade, hostname, user, password)) {
    print_usage(std::cerr);
    return 1;
  }


  // Make connection
#if USE_POSTGRESQL
  pDB = new PgEMdFDB(hostname, user, password, database_to_upgrade);
#elif USE_MYSQL
  pDB = new MySQLEMdFDB(hostname, user, password, database_to_upgrade);
#endif

  // Zero-fill password
  zero_fill_string(password);

  int nResult = 3;
  try {

    // Check that we connected
    if (!pDB->ConnectionOk()) {
      std::cerr << "Connection to backend server could not be established." << std::endl;
      std::cerr << "This usually has one of the following causes:" << std::endl;
      std::cerr << "  - The back-end is not running." << std::endl;
      std::cerr << "  - The back-end is not configured for incoming TCP/IP connections." << std::endl;
      std::cerr << "  - You did not supply a user/password with the -u/-p options." << std::endl;
      std::cerr << "    Note that the default user is 'emdf'." << std::endl;
      std::cerr << "  - You have not followed the instructions in the 'bootstrapping.txt'" << std::endl;
      std::cerr << "    document." << std::endl;
      std::cerr << "There are also hints in the HOW-TO-USE document on configuring your back-end." << std::endl;
      std::cerr << "     Please see      <http://emdros.org/whattodo.html>" << std::endl;
      std::cerr << "     for more information on this problem." << std::endl;
      std::cerr << "Also, please be sure to download the full documentation, available from" << std::endl;
      std::cerr << "<http://emdros.org/>.  You won't be able to make Emdros useful without it." << std::endl;
      delete pDB;
      return 2;
    }

    // Check whether we need upgrading
    std::string strCurrentDBVersion;
    bool bNeedsUpgrading;
    bool bCanUpgrade;
    int nCurrentVersion;
    if (!pDB->GetCurrentDatabaseVersion(database_to_upgrade, strCurrentDBVersion, nCurrentVersion, bNeedsUpgrading, bCanUpgrade)) {
      std::cerr << "Could not determine current database version. Aborting without modifying DB." << std::endl;
      delete pDB;
      return 5;
    }
    
    // Show current version
    std::cout << "Current db version:" << std::endl;
    std::cout << "    Current database version is " << strCurrentDBVersion << "." << std::endl;

    // Check to see whether we should upgrade
    if (!bNeedsUpgrading) {
      std::cout << "The database is compatible with the current version of Emdros," << std::endl
		<< "and doesn't need upgrading. Aborting without modifying database." << std::endl;
      delete pDB;
      return 0;
    }

    if (!bCanUpgrade) {
      std::cout << "ERROR: We cannot upgrade the database." << std::endl;
      std::cout << "Exiting without doing anything to the database." << std::endl;
      delete pDB;
      return 4;
    }

    // Show warning
    if (!show_warning()) {
      std::cout << "OK, aborting without modifying DB." << std::endl;
      delete pDB;
      return 0;
    }


    // Show our intention
    std::cout << "Upgrading database to version 1.1.8 and beyond..." << std::endl;

    if (pDB->UpgradeDatabase(database_to_upgrade, nCurrentVersion)) {
      nResult = 0;
    } else {
      nResult = 5;
      std::cerr << "ERROR: Database error..." << std::endl;
      std::cerr << pDB->ErrorMessage() << std::endl; 
    }
  } catch (CEMdFDBException e) {
    std::cerr << "ERROR: CEMdFDBException (Database error)..." << std::endl;
    std::cerr << pDB->ErrorMessage() << std::endl; 
  } catch (BadMonadsException e) {
    std::cerr << "BadMonadsException caught.  Program aborted." << std::endl;
  } catch (NoElementsException e) {
    std::cerr << "NoElementsException caught.  Program aborted." << std::endl;
  } catch (CEMdFOutputWrongCharactersetException e) {
    std::cerr << "CEMdFOutputWrongCharactersetException caught.  Program aborted." << std::endl;
  } catch (CEMdFOutputException e) {
    std::cerr << "CEMdFOutputException caught.  Program aborted." << std::endl;
  } catch (...) {
    std::cerr << "Unknown exception occurred.  Program aborted." << std::endl;
  }

  // Clean up
  delete pDB;

  if (nResult == 0) {
    std::cout << "Database " << database_to_upgrade << " was successfully upgraded." << std::endl;
  } else {
    std::cerr << "An error occurred while upgrading database " << database_to_upgrade << "." << std::endl;
    std::cerr << "Reason: ";
    if (nResult == 3) {
      std::cerr << "An exception occurred. (See above.)" << std::endl;
    } else if (nResult == 5) {
      std::cerr << "A database error occurred. (See above.)" << std::endl;
    }
    std::cerr << "Database was not successfully upgraded." << std::endl;
    std::cerr << "Try restoring from a backup and trying once more." << std::endl;
    std::cerr << "If this does not work, please contact the Emdros maintainer," << std::endl;
    std::cerr << "Ulrik Petersen, at ulrikp@users.sourceforge.net with a problem description." << std::endl;
  }
  return nResult;
}


