/*
 * mqltry.cpp
 *
 * Trial of MQL and EMdF framework
 * Created: 3/6-2001 (March 6, 2001)
 * Last update: 5/19-2004
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2004  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <default_password.h>
#include <emdros_environment.h>
#include <mql_sheaf.h>
#include <iostream>
#include <sstream>
#include <string>
#include <string_func.h>

static bool bAllOK; // Global variable -- ugly, but it works

void show_version(std::ostream& ostr)
{
  ostr << "mqltry from Emdros version " << EMDROS_VERSION 
       << " on " << EMdFDB::BackendName() << std::endl;
}

void print_usage(std::ostream& ostr)
{
  ostr << "Usage:" << std::endl;
  ostr << "   mqltry [options] " << std::endl;
  ostr << "OPTIONS:" << std::endl;
  ostr << "   --help               Show this help" << std::endl;
  ostr << "   -V , --version       Show version" << std::endl;
  ostr << "   -h , --host host     Set hostname to connect to" << std::endl;
  ostr << "   -u , --user user     Set database user to connect as (default: 'emdf')" << std::endl;
  ostr << "   -p , --password pwd  Set password to use for database user" << std::endl;
  ostr << "DEFAULTS:" << std::endl;
  ostr << "   -h localhost" << std::endl;
  ostr << "   -u emdf" << std::endl;
}

bool parse_argument(int argc, char* argv[], int &i, std::string& hostname, std::string& user, std::string& password)
{
  if (argv[i][0] != '-') {
    return false;
  } else { // [0] == '-'
    std::string argument(argv[i]);
    if (argument == "--help") {
      print_usage(std::cout);
      exit(0);
    } else if (argument == "-V" || argument == "--version") {
      show_version(std::cout);
      exit(0);
    } else if (argument == "-h"
	       || argument == "--host") {
      if (i < (argc-1)) {
	++i;
	hostname = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a hostname as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " vlager" << std::endl;
	return false;
      }
    } else if (argument == "-u" || argument == "--user") {
      if (i < (argc-1)) {
	++i;
	user = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a user as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " emdf" << std::endl;
	return false;
      }
    } else if (argument == "-p" || argument == "--password") {
      if (i < (argc-1)) {
	++i;
	password = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a password as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " secret" << std::endl;
	return false;
      }
    } else {
      std::cerr << "ERROR: Unrecognized option " << argument << std::endl;
      return false;
    }
    return true;
  }
}

bool parse_arguments(int argc, char* argv[], std::string& hostname, std::string& user, std::string& password)
{
  // Set defaults
  hostname = "localhost";
  user = "emdf";

  // Get the default password
#if USE_DEFAULT_PASSWORD
  get_default_password(password);
#else
  password = "";
#endif


  for (int i=1; i < argc; ++i) {
    if (!parse_argument(argc, argv, i, hostname, user, password)) 
      return false;
  }
  
  return true;

}


#define DB_ERROR { \
    std::cerr << "ERROR: Database error in " << success << " while executing query " << std::endl; \
    std::cerr << "-------------------" << std::endl; \
    std::cerr << mql_query << std::endl; \
    std::cerr << "-------------------" << std::endl; \
    std::cerr << pEE->get_DB_error() << std::endl;  \
    std::cerr << pEE->get_compiler_error() << std::endl; \
    bAllOK = false; \
}


#define QUERY_WAS { \
	  std::cerr << "The query was:" << std::endl; \
	  std::cerr << "------------------------------------------------" << std::endl; \
	  std::cerr << mql_query << std::endl; \
	  std::cerr << "------------------------------------------------" << std::endl; \
	  std::cerr << pEE->get_compiler_error() << std::endl; \
	  std::cerr << "while attempting " << success << std::endl; \
}

#define COMPILE_ERROR { \
      std::cerr << "ERROR: Compile error in " << success << "." << std::endl; \
      QUERY_WAS; \
      bAllOK = false; \
}



#define ROW_ERROR { \
	std::cerr << "Error: Expecting " << no_of_expected_rows << " rows, got ." << no_of_rows << "." << std::endl; \
        QUERY_WAS; \
        bAllOK = false; \
}

bool exec_string(CEmdrosEnv *pEE, std::string mql_query, std::string success, bool bExpectedResult = true)
{
  bool bResult;
  if (!pEE->execute_string(mql_query, bResult, true, false)) {
    DB_ERROR;
    bResult= false;
  } else {
    if (bExpectedResult != bResult) {
      COMPILE_ERROR;
      bResult= false;
    } else { 
      std::cout << "SUCCESS in " << success << std::endl;
      bResult= true;
    }
  }

  // Clean up
  pEE->clean();

  // Return
  return bResult;
}

bool exec_string_get_id_d(CEmdrosEnv *pEE, std::string mql_query, std::string success, id_d_t& result_id_d)
{
  bool bResult;
  if (!pEE->execute_string(mql_query, bResult, true, false)) {
    DB_ERROR;
    bResult = false;
  } else {
    if (!bResult) {
      COMPILE_ERROR;
    } else {
      std::cout << "SUCCESS in " << success << std::endl;
    }
  }

  // Get resulting id_d
  if (bResult) {
    CTable *pResult = pEE->getTable();
    TableIterator ti = pResult->iterator();
    std::string strId_d = pResult->get_column(ti, 1);
    result_id_d = string2id_d(strId_d);
  }

  // Clean up
  pEE->clean();

  // Return result
  return bResult;
}

bool exec_string_check_no_of_rows(CEmdrosEnv *pEE, std::string mql_query, std::string success, long no_of_expected_rows)
{
  bool bResult;
  if (!pEE->execute_string(mql_query, bResult, true, false)) {
    DB_ERROR;
    bResult = false;
  } else {
    if (!bResult) {
      COMPILE_ERROR;
    } else {
      // Check no of expected rows
      long no_of_rows = pEE->getTable()->size();
      if (no_of_rows != no_of_expected_rows) {
	bResult = false;
	ROW_ERROR;
      } else {
	std::cout << "SUCCESS in " << success << std::endl;
      }
    }
  }

  // Clean up
  pEE->clean();

  // Return result
  return bResult;
}

bool exec_string_check_monad(CEmdrosEnv *pEE, std::string mql_query, std::string success, monad_m expected_monad)
{
  bool bResult = false;
  if (!pEE->execute_string(mql_query, bResult, true, false)) {
    DB_ERROR;
    bResult = false;
  } else {
    if (!bResult) {
      COMPILE_ERROR;
    } else {
      // Check monad
      long no_of_expected_rows = 1;
      long no_of_rows = pEE->getTable()->size();
      if (no_of_rows != no_of_expected_rows) {
	bResult = false;
	ROW_ERROR;
      } else {
	monad_m monad_m_result;
	CMQLResult *pResult = pEE->getResult();
	TableIterator ti = pResult->iterator();
	std::string strMonad_m = pResult->get_column(ti, 1);
	monad_m_result = string2monad_m(strMonad_m);
	if (monad_m_result != expected_monad) {
	  std::cerr << "Error: Expecting resulting monad to be " << expected_monad << " but it was " << monad_m_result << "." << std::endl;
	  QUERY_WAS;
	  bResult = false;
	  bAllOK = false;
	} else {
	  std::cout << "SUCCESS in " << success << std::endl;
	  bResult = true;
	}
      }
    }
  }

  // Clean up
  pEE->clean();

  // Return result
  return bResult;
}


void exec_string_expect_error(CEmdrosEnv *pEE, std::string mql_query, std::string success)
{
  bool bResult;
  if (!pEE->execute_string(mql_query, bResult, false, false)) {
    // Do nothing
    // ESPECIALLY don't call DB_ERROR.  This will set bAllOK to false,
    // which is not the intention!
  } else {
    std::cout << "SUCCESS in " << success << std::endl;
  }

  // Clean up
  pEE->clean();
}


int testall(CEmdrosEnv *pEE)
{
  std::string mql_query;
  int result = 0;
  CSheaf *pSheaf1 = 0;
  SheafConstIterator ci1;
  SheafIterator i1;

  // Drop database EMdF4
  mql_query = "DROP DATABASE EMdF4 GO";
  exec_string_expect_error(pEE, mql_query, "dropping database EMdF4");

  // Drop database EMdF3
  mql_query = "DROP DATABASE EMdF3 GO";
  exec_string_expect_error(pEE, mql_query, "dropping database EMdF3");

  bAllOK = true; // Global variable

  // Create database
  mql_query = "CREATE DATABASE EMdF3 GO";
  if (!exec_string(pEE, mql_query, "creating database EMdF3"))
    goto end2;

  // Create database
  mql_query = "CREATE DATABASE EMdF4 GO";
  if (!exec_string(pEE, mql_query, "creating database EMdF4"))
    goto end1;

  // Use database
  mql_query = "USE DATABASE EMdF3 GO";
  if (!exec_string(pEE, mql_query, "using database EMdF3"))
    goto end;

  // Create enumeration
  mql_query = 
    "CREATE ENUMERATION "
    "psp_t = { "
    " default psp_None = -1, verb = 1, noun, proper_noun, adverb, preposition, \n"
    " conjunction = 7, foobar, pronoun "
    "} GO";
  if (!exec_string(pEE, mql_query, "creating enumeration psp_t"))
    goto end;

  // Update enumeration
  mql_query = 
    "UPDATE ENUMERATION "
    "psp_t = { "
    " remove foobar, add default negation = 11, add demonstrative = 8, update conjunction = 6 "
    "} GO";
  if (!exec_string(pEE, mql_query, "updating enumeration psp_t"))
    goto end;

  // Create enumeration
  mql_query = 
    "CREATE ENUMERATION "
    "sploinks_t = { "
    " habib, habab, default habub, WxQt = 142, WXQt = 132, xYqt = 121, XYqt = 111 "
    "} GO";
  if (!exec_string(pEE, mql_query, "creating enumeration sploinks_t"))
    goto end;

  // Selecting enumeration constants
  mql_query = 
    "SELECT ENUMERATION CONSTANTS FROM ENUMERATION sploinks_t GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting enumeration constants", 7))
    goto end;

  // Dropping enumeration constants
  mql_query = 
    "UPDATE ENUMERATION "
    "sploinks_t = { "
    "  remove xYqt"
    "} GO";
  if (!exec_string(pEE, mql_query, "dropping enumeration constant"))
    goto end;

  // Selecting enumeration constants
  mql_query = 
    "SELECT ENUMERATION CONSTANTS FROM ENUMERATION sploinks_t GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting enumeration constants", 6))
    goto end;

  // Updating enumeration constant with case-insensitive duplicate
  mql_query = 
    "UPDATE ENUMERATION "
    "sploinks_t = { "
    "  UPDATE WXQt = -2 "
    "} GO";
  if (!exec_string(pEE, mql_query, "updating enumeration constant"))
    goto end;

  // Create enumeration
  mql_query = 
    "DROP ENUMERATION sploinks_t GO";
  if (!exec_string(pEE, mql_query, "dropping enumeration sploinks_t"))
    goto end;

  // Create object type
  mql_query = 
    "CREATE OBJECT TYPE\n"
    "[Word\n"
    "  surface : string(34);\n"
    "  lemma : string(127);\n"
    "  half_verse : ascii(1);\n"
    "  psp : psp_t;\n"
    "  number : integer;\n"
    "] GO";
  if (!exec_string(pEE, mql_query, "creating object type Word"))
    goto end;

  // Create object type
  mql_query = 
    "CREATE OBJECT TYPE\n"
    "[Sploinks] GO";
  if (!exec_string(pEE, mql_query, "creating object type Sploinks"))
    goto end;

  // Create object type
  mql_query = 
    "CREATE OBJECT TYPE\n"
    "[Phalanges\n"
    "  fog : ASCII(1) DEFAULT '?';\n"
    "  \n"
    "]\n"
    "COVERED_BY sploinks\n"
    "BUILDABLE_FROM sploinks\n"
    " GO";
  if (!exec_string(pEE, mql_query, "creating object type Phalanges"))
    goto end;

  // Update object type
  mql_query = 
    "UPDATE OBJECT TYPE\n"
    "[Phalanges\n"
    "  ADD fax : ASCII(32);\n"
    "]\n"
    "ADD COVERED_BY word\n"
    "ADD BUILDABLE_FROM word\n"
    "REMOVE COVERED_BY sploinks\n"
    "REMOVE BUILDABLE_FROM sploinks\n"
    " GO";
  if (!exec_string(pEE, mql_query, "updating object type Phalanges"))
    goto end;

  // Create object type
  mql_query = 
    "CREATE OBJECT TYPE\n"
    "[Phrase\n"
    "  dummy : ASCII(1) default 'C';\n"
    "] GO";
  if (!exec_string(pEE, mql_query, "creating object type Phrase"))
    goto end;

  // Create segment
  mql_query = 
    "CREATE SEGMENT Old_Testament\n"
    "RANGE = 1 - 500000 GO";
  if (!exec_string(pEE, mql_query, "creating segment Old_Testament"))
    goto end;

  // Create segment
  mql_query = 
    "CREATE SEGMENT New_Testament\n"
    "RANGE = 1000000 - 1200000 GO";
  if (!exec_string(pEE, mql_query, "creating segment New_Testament"))
    goto end;

  // Select object types
  mql_query = 
    "SELECT OBJECT TYPES GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting object types",4))
    goto end;

  // Select features
  mql_query = 
    "SELECT FEATURES FROM OBJECT TYPE [Word] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting features",6))
    goto end;

  // Select features
  mql_query = 
    "SELECT FEATURES FROM OBJECT TYPE [Phalanges] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting features", 3))
    goto end;

  // Select enumerations
  mql_query = 
    "SELECT ENUMERATIONS GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting enumerations",1))
    goto end;

  // Select enumeration constants
  mql_query = 
    "SELECT ENUMERATION CONSTANTS FROM ENUM psp_t GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting enumeration constants", 10))
    goto end;
  std::cout << "Check that conjunction has value 6 and that negation is the default." << std::endl;

  // Select object types using enumeration
  mql_query = 
    "SELECT OBJECT TYPES USING ENUMERATION psp_t GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting object types using enumeration",1))
    goto end;

  // Select segments
  mql_query = 
    "SELECT SEGMENTS GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting segments",2))
    goto end;

  // Create object from monads
  mql_query = 
    "CREATE OBJECT\n"
    "FROM MONADS = { 1-2 }\n"
    "[Word\n"
    "  surface := 'KAI\\';\n"
    "  lemma := 'KAI/';\n"
    "  psp := conjunction;\n"
    "  number := 1;\n"
    "] GO";
  id_d_t object1;
  if (!exec_string_get_id_d(pEE, mql_query, "creating object from monads", object1))
    goto end;

  // Create object from monads
  mql_query = 
    "CREATE OBJECT\n"
    "FROM MONADS = { 4-5 }\n"
    "[Word\n"
    "  surface := \"U(MEI=S\";\n"
    "  lemma := 'E)GW/';\n"
    "  psp := pronoun;\n"
    "  number := 2;\n"
    "] GO";
  id_d_t object2;
  if (!exec_string_get_id_d(pEE, mql_query, "creating object from monads", object2))
    goto end;

  // Create object from id_ds
  mql_query = 
    "CREATE OBJECT\n"
    "FROM ID_DS = " + id_d2string(object1) + ", " + id_d2string(object2) + "\n" +
    "[Phrase] GO\n";
  id_d_t object3;
  if (!exec_string_get_id_d(pEE, mql_query, "creating object from id_ds", object3))
    goto end;

  // Create object from monads
  mql_query = 
    "CREATE OBJECT\n"
    "FROM MONADS = { 3 }\n"
    "[Word\n"
    "  surface := \"DE\\\\\";\n"
    "  lemma := 'DE/';\n"
    "  psp := conjunction;\n"
    "] GO";
  id_d_t object4;
  if (!exec_string_get_id_d(pEE, mql_query, "creating object from monads", object4))
    goto end;

  // Select objects at
  mql_query = "SELECT OBJECTS\n"
    "AT MONAD = 1\n"
    "[Word] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting objects at monad",1))
    goto end;

  // Get features
  mql_query = "GET FEATURES surface, self, number\n"
    "FROM OBJECTS WITH ID_DS = " + id_d2string(object1) + ", " + id_d2string(object2) + "\n"
    "[Word] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "getting features",2))
    goto end;

  // Get features
  mql_query = "GET FEATURES self\n"
    "FROM OBJECT WITH ID_D = " + id_d2string(object3) + "\n"
    "[Phrase] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "getting features",1))
    goto end;

  // Update object
  mql_query = "UPDATE OBJECT\n"
    "BY MONADS = { 4-5 }\n"
    "[Word\n"
    "  surface := \"H(MEI=S\";\n"
    "  number := 1;\n"
    "] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "updating object by monads",1))
    goto end;

  // Update object
  mql_query = "UPDATE OBJECT\n"
    "BY ID_D = " + id_d2string(object3) + "\n"
    "[Phrase\n"
    "  dummy := 'D';\n"
    "] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "updating object by id_ds",1))
    goto end;

  // min_m
  mql_query = "SELECT MIN_M\n"
    "GO";
  if (!exec_string_check_monad(pEE, mql_query, "checking min_m", 1))
    goto end;
  
  // max_m
  mql_query = "SELECT MAX_M\n"
    "GO";
  if (!exec_string_check_monad(pEE, mql_query, "checking max_m", 5))
    goto end;

  // Query
  mql_query = "SELECT FOCUS OBJECTS\n"
    "WHERE\n"
    "[Phrase\n"
    "  [Word\n"
    "    psp = conjunction\n"
    "  ]\n"
    "  [gap retrieve\n"
    "     [Word\n"
    "       psp = conjunction\n"
    "     ]\n"
    "  ]\n"
    "  [Word\n"
    "     number = 1\n"
    "  ]\n"
    "] GO";
  if (!exec_string(pEE, mql_query, "querying using MQL's query-subset"))
    goto end;

  // Query
  mql_query = "SELECT OBJECTS\n"
    "HAVING MONADS IN { 1-2, 4-5 }\n"
    "[ALL] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "selecting objects having monads in monad set", 3))
    goto end;


  // Delete object
  mql_query = "DELETE OBJECTS\n"
    "BY MONADS = { 1-2, 4-5 }\n"
    "[Word] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "deleting objects by monads", 2))
    goto end;

  // min_m
  mql_query = "SELECT MIN_M\n"
    "GO";
  if (!exec_string_check_monad(pEE, mql_query, "checking min_m", 1))
    goto end;
  
  // max_m
  mql_query = "SELECT MAX_M\n"
    "GO";
  if (!exec_string_check_monad(pEE, mql_query, "checking max_m", 5))
    goto end;

  // Delete object
  mql_query = "DELETE OBJECTS\n"
    "BY ID_D = " + id_d2string(object3) + "\n" 
    "[Phrase] GO";
  if (!exec_string_check_no_of_rows(pEE, mql_query, "deleting object by id_ds", 1))
    goto end;

  // Create object with id_d
  mql_query = "CREATE OBJECT\n"
    "FROM MONADS = { 1-10 }\n"
    "WITH ID_D = 12000\n"
    "[Phalanges]\n"
    "GO";
  id_d_t phalanges_object_id_d;
  if (!exec_string_get_id_d(pEE, mql_query, "creating object from monads with specific id_d", phalanges_object_id_d))
    goto end;

  if (phalanges_object_id_d == 12000) {
    std::cout << "SUCCESS: Phalanges object had the correct id_d." << std::endl;
  } else {
    std::cerr << "FAILURE: Phalanges object did not have the correct id_d." << std::endl;
    std::cerr << "         The id_d was: " << phalanges_object_id_d << std::endl;
    bAllOK = false;
    goto end;
  }

  // Check that next auto-generated id_d is correct
  mql_query = "CREATE OBJECT\n"
    "FROM MONADS = { 1-11 }\n"
    "[Phalanges]\n"
    "GO";
  if (!exec_string_get_id_d(pEE, mql_query, "creating object from monads without specific id_d", phalanges_object_id_d))
    goto end;

  if (phalanges_object_id_d == 12001) {
    std::cout << "SUCCESS: Object had the correct auto-generated id_d." << std::endl;
  } else {
    std::cerr << "FAILURE: Object did not have the correct auto-generated id_d." << std::endl;
    std::cerr << "         The id_d was: " << phalanges_object_id_d << std::endl;
    bAllOK = false;
    goto end;
  }



  // Drop object type
  mql_query = 
    "DROP OBJECT TYPE\n"
    "[Word] GO";
  if (!exec_string(pEE, mql_query, "dropping object type Word"))
    goto end;

  // Drop segment
  mql_query = 
    "DROP SEGMENT Old_Testament GO";
  if (!exec_string(pEE, mql_query, "dropping segment Old_Testament"))
    goto end;

  // Monad set calculation
  mql_query = 
    "MONAD SET CALCULATION\n"
    "{ 1-3, 6-8, 10-14 }\n"
    "UNION\n"
    "{ 4-7 }\n"
    "INTERSECT\n"
    "{ 2-13 }\n"
    "DIFFERENCE\n"
    "{ 2-3, 12 }\n"
    "GO";
  if (!exec_string(pEE, mql_query, "performing monad set calculation"))
    goto end;

  // Monad set calculation
  mql_query = 
    "MONAD SET CALCULATION\n"
    "{ 1-3, 6-8, 10-12 }\n"
    "UNION\n"
    "{ 4-7 }\n"
    "GO";
  if (!exec_string(pEE, mql_query, "performing monad set calculation"))
    goto end;

  // Const iterator on empty sheaf
  pSheaf1 = new CSheaf(false);
  ci1 = pSheaf1->const_iterator();
  if (ci1.hasNext() == false) {
    std::cout << "SUCCESS: const iterator on empty sheaf tested OK." << std::endl;
    delete pSheaf1;

  } else {
    std::cerr << "ERROR: const iterator on empty sheaf not right." << std::endl;
    delete pSheaf1;
    bAllOK = false;
    goto end;
  }

  // Const iterator on failed sheaf
  pSheaf1 = new CSheaf(true);
  ci1 = pSheaf1->const_iterator();
  if (ci1.hasNext() == false) {
    std::cout << "SUCCESS: const iterator on failed sheaf tested OK." << std::endl;
    delete pSheaf1;

  } else {
    std::cerr << "ERROR: const iterator on failed  sheaf not right." << std::endl;
    delete pSheaf1;
    bAllOK = false;
    goto end;
  }

  // Iterator on empty sheaf
  pSheaf1 = new CSheaf(false);
  i1 = pSheaf1->iterator();
  if (i1.hasNext() == false) {
    std::cout << "SUCCESS: iterator on empty sheaf tested OK." << std::endl;
    delete pSheaf1;

  } else {
    std::cerr << "ERROR: iterator on empty sheaf not right." << std::endl;
    delete pSheaf1;
    bAllOK = false;
    goto end;
  }

  // Iterator on failed sheaf
  pSheaf1 = new CSheaf(true);
  i1 = pSheaf1->iterator();
  if (i1.hasNext() == false) {
    std::cout << "SUCCESS: iterator on failed sheaf tested OK." << std::endl;
    delete pSheaf1;

  } else {
    std::cerr << "ERROR: iterator on failed  sheaf not right." << std::endl;
    delete pSheaf1;
    bAllOK = false;
    goto end;
  }


end:
  // Drop database
  mql_query = "DROP DATABASE EMdF4 GO";
  exec_string(pEE, mql_query, "dropping database EMdF4");

end1:
  // Drop database
  mql_query = "DROP DATABASE EMdF3 GO";
  exec_string(pEE, mql_query, "dropping database EMdF3");

end2:
  result = bAllOK ? 0 : 1;
  return result;
}

int main(int argc, char* argv[])
{
  int result = 0;

  std::string hostname;
  std::string password;
  std::string user;
  std::string initial_db("emdf");
  if (!parse_arguments(argc, argv, hostname, user, password)) {
    print_usage(std::cerr);
    return 1;
  }

  // Make execution environment
  CEmdrosEnv *pEE;
  pEE = new CEmdrosEnv(kOKConsole, kCSISO_8859_1,
		       hostname,
		       user, password,
		       initial_db);
  

  // Zero-fill the password
  zero_fill_string(password);


  // Test connection
  if (!pEE->ConnectionOk()) {
    std::cerr << "DB error message: '";
    std::cerr << pEE->get_DB_error() << "'" << std::endl;  
    std::cerr << "FAILURE: Could not connect to back-end server." << std::endl;
    std::cerr << "This usually has one of the following causes:" << std::endl;
    std::cerr << "  - The back-end is not running." << std::endl;
    std::cerr << "  - The back-end is not configured for incoming TCP/IP connections." << std::endl;
    std::cerr << "  - You did not supply a user/password with the -u/-p options." << std::endl;
    std::cerr << "    Note that the default user is 'emdf'." << std::endl;
    std::cerr << "  - You have not followed the instructions in the 'bootstrapping.txt'" << std::endl;
    std::cerr << "    document." << std::endl;
    std::cerr << "There are also hints in the HOW-TO-USE document on configuring your back-end." << std::endl;
    std::cerr << "     Please see      <http://emdros.org/whattodo.html>" << std::endl;
    std::cerr << "     for more information on this problem." << std::endl;
    std::cerr << "Also, please be sure to download the full documentation, available from" << std::endl;
    std::cerr << "<http://emdros.org/>.  You won't be able to make Emdros useful without it." << std::endl;

    delete pEE;
    return 1;
  }


  try {
    result = testall(pEE);
  } catch (CEMdFDBException e) {
    std::cerr << "ERROR: CEMdFDBException (Database error)..." << std::endl;
    std::cerr << pEE->get_DB_error() << std::endl; 
    std::cerr << pEE->get_compiler_error() << std::endl;
    bAllOK = false;
  } catch (...) {
    std::cerr << "ERROR: Unknown exception..." << std::endl;
    std::cerr << pEE->get_DB_error() << std::endl; 
    std::cerr << pEE->get_compiler_error() << std::endl;
    bAllOK = false;
  }

  if (bAllOK) {
    std::cout << "SUCCESS: All tests completed successfully." << std::endl;
  } else {
    std::cerr << "FAILURE: At least one test failed." << std::endl;
  }

  // Clean up
  delete pEE;
  return result;
}
