/*
 * emdfdb.cpp
 *
 * EMdF database base class
 *
 * Ulrik Petersen
 * Created: 1/27-2001
 * Last update: 5/29-2004
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2004  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#if HAVE_CONFIG_H
#include <emdros-config.h>
#endif 

#include <emdf.h>
#include <emdfdb.h>
#include <emdf_value.h>
#include <utils.h>
#include <string_func.h>
#include <cstdio>
#include <debug.h>
#include <iostream>
#include <sstream>
#include <vector>
#include <set>
#include <logging.h>
#include <cassert>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef WIN32
#include <windows.h>
#endif


#ifdef WIN32
void sleep(DWORD milliseconds)
{
  SleepEx(milliseconds, FALSE);
}
#endif

EMdFDB::EMdFDB() : pConn(0)
{
  m_enum_const_cache = new CTable;
  m_object_type_cache_table = new CTable;
  m_enum_cache = new CTable;
  m_feature_cache = new CTable;
}

EMdFDB::~EMdFDB()        
{
  if (pConn != 0)
    delete pConn;
  delete m_enum_cache;
  delete m_enum_const_cache;
  delete m_object_type_cache_table;
  delete m_feature_cache;
}



//
// Sequence manipulation


bool EMdFDB::SetNextObjectID_DIfNotHigher(id_d_t next_id_d)
{
  return false; // Must be overridden
}


bool EMdFDB::GetNextID(int sequence, id_d_t& out)
{
  if (pConn == 0)
    return false;
  else {
    // Declare szQuery
    char szQuery[512];

    // Begin transaction
    bool bDoCommit = pConn->BeginTransaction();

    try {

#if USE_POSTGRESQL
      // Ask what the next value is
      sprintf(szQuery, "SELECT nextval('sequence_118_%d');", sequence);
      if (!pConn->Exec(szQuery)) {
	DEBUG_QUERY_FAILED("EMdFDB::GetNextID", szQuery);
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }

      // Access result
      if (!pConn->AccessTuple(0, 0, "EMdFDB::GetNextID", out)) {
	// DEBUG information has already been printed if it is on.
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }

#elif USE_MYSQL

      // Update value
      sprintf(szQuery, 
	      "UPDATE sequence_118_%d\n"
	      "SET sequence_value = LAST_INSERT_ID(sequence_value + 1);",
	      sequence);
      if (!pConn->Exec(szQuery)) {
	DEBUG_QUERY_FAILED("EMdFDB::GetNextID", szQuery);
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }

      // Ask what the current value is
      sprintf(szQuery, "SELECT LAST_INSERT_ID();");
      if (!pConn->Exec(szQuery)) {
	DEBUG_QUERY_FAILED("EMdFDB::GetNextID", szQuery);
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }

      // Access result
      if (!pConn->AccessTuple(0, 0, "EMdFDB::GetNextID", out)) {
	// DEBUG information has already been printed if it is on.
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
#endif

      // Commit transaction
      if (bDoCommit) {
	if (!pConn->CommitTransaction()) {
	  DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::GetNextID");
	  return false;
	}
      }
    } catch (EMdFNULLValueException e) {
      // Abort transaction if in progress
      if (bDoCommit)
	pConn->AbortTransaction();

      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetNextID");

      // There was a NULL value exception, hence we return an error
      return false;
    }

    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::GetNextObjectID_D(/* out */ id_d_t& id_d)
{
  if (pConn == 0)
    return false;
  else {
    return GetNextID(SEQUENCE_OBJECT_ID_DS, id_d);
  }
}

bool EMdFDB::GetNextOtherID(/* out */ id_d_t& id_d)
{
  if (pConn == 0)
    return false;
  else {
    return GetNextID(SEQUENCE_OTHER_IDS, id_d);
  }
}

//
// Database manipulation
bool EMdFDB::CreateDatabase(const std::string& db_name)
{
  if (pConn == 0)
    return false;
  else {
    // Get a DBname which we can use
    std::string NewDbName;
    ConvertDBName(db_name, NewDbName);

    // Declare szQuery
    char szQuery[512];

    // Create database
    std::ostringstream query_stream;
    query_stream << "CREATE DATABASE " << NewDbName;
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateDatabase", query_stream.str().c_str());
      return false;
    }

    // Switch to new database
    if (!UseDatabase(db_name)) {
      DEBUG_X_FAILED("EMdFDB::CreateDatabase", "UseDatabase");
      return false;
    }

    // Begin transaction
    bool bDoCommit;
    bDoCommit = pConn->BeginTransaction();

    if (!CreateSequenceTables118()) {
      DEBUG_X_FAILED("EMdFDB::CreateDatabase", "CreateSequenceTables118");
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }


    // Create table 'enumerations'
    sprintf(szQuery,
	    "CREATE TABLE enumerations (\n"
	    "    enum_id INT NOT NULL,\n"
	    "    enum_name VARCHAR(%d) NOT NULL,\n"
	    "    PRIMARY KEY (enum_id)\n"
	    ")", MAX_IDENTIFIER_CHARS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateDatabase", szQuery);
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create table 'enumeration_constants'
    sprintf(szQuery,
	    "CREATE TABLE enumeration_constants (\n"
	    "    enum_id INT NOT NULL,\n"
	    "    enum_value_name VARCHAR(%d) NOT NULL,\n"
	    "    value INT NOT NULL,\n"
	    "    is_default CHAR(1) NOT NULL,\n"
	    "    PRIMARY KEY (enum_id, enum_value_name"
#if USE_MYSQL
	    ", value"
#endif
	    ")\n"
	    ")", MAX_IDENTIFIER_CHARS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateDatabase", szQuery);
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create table 'segments'
    sprintf(szQuery,
	    "CREATE TABLE segments (\n"
	    "    segment_name VARCHAR(%d),\n"
	    "    first_monad INT NOT NULL,\n"
	    "    last_monad INT NOT NULL,\n"
	    "    PRIMARY KEY (first_monad)\n"
	    ")", MAX_IDENTIFIER_CHARS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateDatabase", szQuery);
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create table 'object_types'
    sprintf(szQuery,
	    "CREATE TABLE object_types (\n"
	    "    object_type_id INT NOT NULL,\n"
	    "    object_type_name VARCHAR(%d) NOT NULL,\n"
	    "    PRIMARY KEY (object_type_id)\n"
	    ")", MAX_IDENTIFIER_CHARS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateDatabase", szQuery);
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create table 'features'
    sprintf(szQuery,
	    "CREATE TABLE features (\n"
	    "    object_type_id INT NOT NULL,\n"
	    "    feature_name VARCHAR(%d) NOT NULL,\n"
	    "    feature_type_id INT NOT NULL,\n"
	    "    default_value VARCHAR(%d) NOT NULL,\n"
	    "    computed CHAR(1) NOT NULL DEFAULT 'N',\n"
	    "    PRIMARY KEY (object_type_id, feature_name)\n"
	    ")", MAX_IDENTIFIER_CHARS, (MAX_STRING_LENGTH*2));
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateDatabase", szQuery);
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }


    strcpy(szQuery,
	   "CREATE TABLE covered_by (\n"
	   "    object_type_id INT NOT NULL,\n"
	   "    covered_by_id INT NOT NULL,\n"
	   "    PRIMARY KEY (object_type_id, covered_by_id)\n"
	   ")");
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateDatabase", szQuery);
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }



    strcpy(szQuery,
	   "CREATE TABLE buildable_from (\n"
	   "    object_type_id INT NOT NULL,\n"
	   "    buildable_from_id INT NOT NULL,\n"
	   "    PRIMARY KEY (object_type_id, buildable_from_id)\n"
	   ")");
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateDatabase", szQuery);
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create min_m/max_m table
    if (!Create_min_max_m_table()) {
      DEBUG_X_FAILED("EMdFDB::CreateDatabase", "Creating min_m/max_m table");
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }


    // Commit transaction if necessary
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::CreateDatabase");
	return false;
      }
    }

    return true;
  }
}

// Must be called from subclass
bool EMdFDB::UseDatabase(const std::string& db_name)
{
  // Clean out our tables
  m_enum_cache->clear();
  m_object_type_cache_table->clear();
  m_enum_const_cache->clear();
  m_feature_cache->clear();
  m_object_type_cache_map.clear();

  return true;
}

bool EMdFDB::DropDatabase(const std::string& db_name)
{
  if (pConn == 0)
    return false;
  else {
    // Switch database context to "EMdF"
    // So that we can drop.
    if (!UseDatabase("EMdF")) {
      DEBUG_X_FAILED("EMdFDB::DropDatabase", "UseDatabase 'EMdF'");
      return false;
    }
    
    // We need to give the server some time to delete the database
    sleep(2);

    // Get a DBname which we can use
    std::string NewDbName;
    ConvertDBName(db_name, NewDbName);

    std::ostringstream query_stream;
    query_stream << "DROP DATABASE " << NewDbName;
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::DropDatabase", query_stream.str().c_str());
      return false;
    }

    /* Everything was OK. */
    return true;
  }
}

bool EMdFDB::DatabaseExists(const std::string& db_name, bool& result)
  // Note: Do we really need DatabaseExists?
{
  // TODO: Implement
  return false;
}

bool EMdFDB::DBIsInitialized(bool& bIsInitialized)
{
  // Set it to false initially, so this is returned even if 
  // we return false as the return value
  bIsInitialized = false;

  // If pConn isn't initialized, return false
  if (pConn == 0)
    return false;
  else {
    // pConn is initialized.

    // Check that table object_types exists
    bool bTableExists;
    if (!TableExists("object_types", bTableExists)) {
      return false;
    }

    // Assume that DB is initialized (with CreateDatbase())
    // if object_types exists
    bIsInitialized = bTableExists;

    // If we got this far, there were no errors
    return true;
  }
}

void EMdFDB::MakeFeatureSQLType(std::ostringstream& OT_Objects_stream, 
				const feature_info& fi,
				bool add_not_null)
{
  if (!fi.get_is_computed()) {
    // Output feature name
    OT_Objects_stream << "    " << encode_feature_name(fi.get_name()) << " ";
    
    // Output feature type
    long feature_length;
    id_d_t feature_type;
    decode_feature_type_id(fi.get_type(), feature_type, feature_length);
    switch(feature_type) {
    case FEATURE_TYPE_INTEGER:
      OT_Objects_stream << "INT";
      break;
    case FEATURE_TYPE_ID_D:
      OT_Objects_stream << "INT";
      break;
    case FEATURE_TYPE_STRING:
      OT_Objects_stream << "VARCHAR(" << feature_length << ")";
      break;
    case FEATURE_TYPE_ASCII:
      OT_Objects_stream << "VARCHAR(" << feature_length << ")";
      break;
    default: // Enumerations: same as ID_D
      OT_Objects_stream << "INT";
    }
    
    // Output NOT NULL
    if (add_not_null) {
      OT_Objects_stream << " NOT NULL ";
    }
  }
}

//
// Build an SQL string which can be used to constrain a search on
// OT_objects to a particular set of monads.
//
// The string has the form
// (
//  (first_monad >= mse1.first() AND last_monad <= mse1.last())
//   OR 
//  (first_monad >= mse2.first() AND last_monad <= mse2.last())
//   OR
//   ...
//   OR
//  (first_monad >= msen.first() AND last_monad <= msen.last())
// )
//
std::string EMdFDB::MakeMonadConstraintsFromSOM(const CSet_of_monad_ms& monads)
{
  std::ostringstream monad_constraints_stream;
  monad_constraints_stream << "(";
  SOMConstIterator ci = monads.const_iterator();
  while (ci.hasNext()) {
    Monad_Set_Element mse = ci.current();
    monad_constraints_stream 
      << "(first_monad >= " 
      << mse.first() 
      << " AND "
      << "last_monad <= "
      << mse.last()
      << ")";
    ci.next();
    if (ci.hasNext()) {
      monad_constraints_stream 
	<< " OR ";
    }
  }
  monad_constraints_stream << ")";
  return monad_constraints_stream.str();
}

//
// Build an SQL string which can be used to constrain a search on
// OT_objects to a particular set of monads.
//
// The string has the form
// (
//  (mse_first >= mse1.first() AND mse_last <= mse1.last())
//   OR 
//  (mse_first >= mse2.first() AND mse_last <= mse2.last())
//   OR
//   ...
//   OR
//  (mse_first >= msen.first() AND mse_last <= msen.last())
// )
//
std::string EMdFDB::MakeMonadConstraintsFromSOMmse_first(const CSet_of_monad_ms& monads)
{
  std::ostringstream monad_constraints_stream;
  monad_constraints_stream << "(";
  SOMConstIterator ci = monads.const_iterator();
  while (ci.hasNext()) {
    Monad_Set_Element mse = ci.current();
    monad_constraints_stream 
      << "(mse_first >= " 
      << mse.first() 
      << " AND "
      << "mse_last <= "
      << mse.last()
      << ")";
    ci.next();
    if (ci.hasNext()) {
      monad_constraints_stream 
	<< " OR ";
    }
  }
  monad_constraints_stream << ")";
  return monad_constraints_stream.str();
}

//
// Build an SQL string which can be used to constrain a search on
// OT_objects to a particular set of monads.
//
// The string has the form
// (
//  (monad_m >= mse1.first() AND monad_m <= mse1.last())
//   OR 
//  (monad_m >= mse2.first() AND monad_m <= mse2.last())
//   OR
//   ...
//   OR
//  (monad_m >= msen.first() AND monad_m <= msen.last())
// )
//
std::string EMdFDB::MakeMonadConstraintsFromSOMMonad_m(const CSet_of_monad_ms& monads)
{
  std::ostringstream monad_constraints_stream;
  monad_constraints_stream << "(";
  SOMConstIterator ci = monads.const_iterator();
  while (ci.hasNext()) {
    Monad_Set_Element mse = ci.next();
    monad_constraints_stream 
      << "(monad_m >= " 
      << mse.first() 
      << " AND "
      << "monad_m <= "
      << mse.last()
      << ")";
    if (ci.hasNext()) {
      monad_constraints_stream 
	<< " OR ";
    }
  }
  monad_constraints_stream << ")";
  return monad_constraints_stream.str();
}


//
// Object type manipulation
bool EMdFDB::CreateObjectType(const std::string& object_type_name,
			      const std::list<feature_info>& feature_infos,
                              /* out */ id_d_t& object_type_id)
{
  if (pConn == 0)
    return false;
  else {
    // Begin transaction
    bool bDoCommit;
    bDoCommit = pConn->BeginTransaction();

    // Generate auto-generated ID
    if (!GetNextOtherID(object_type_id)) {
      DEBUG_X_FAILED("EMdFDB::CreateObjectType", "GetNextOtherID");
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Insert new values 
    {
      std::ostringstream query_stream;
      query_stream 
	<< "INSERT INTO object_types (object_type_id, object_type_name)\n"
	<< "VALUES ( " << object_type_id << ", '" 
	<< OTN << "' )\n";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::CreateObjectType", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Build query for creating OT_objects, with list of features
    std::ostringstream OT_Objects_stream;
    OT_Objects_stream << "CREATE TABLE " << OTN << "_objects(\n"
      "    object_id_d INT NOT NULL,\n"
      "    monad_ms_id INT NOT NULL,\n"
      "    first_monad INT NOT NULL,\n"
      "    last_monad INT NOT NULL,\n";
    std::list<feature_info>::const_iterator f_it(feature_infos.begin());
    std::list<feature_info>::const_iterator f_iend(feature_infos.end());
    while (f_it != f_iend) {
      // Make SQL type
      MakeFeatureSQLType(OT_Objects_stream, *f_it, true);

      // Make newline
      OT_Objects_stream << ",\n";

      // Advance iterator
      ++f_it;
    }

    OT_Objects_stream << 
      "    PRIMARY KEY (object_id_d)\n"
      ")";

    // Create OT_Objects
    if (!pConn->Exec(OT_Objects_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateObjectType", OT_Objects_stream.str().c_str());
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create indices on OT_objects    
    if (!CreateIndicesOnOTObjects(OTN)) {
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create OT_monad_ms
    std::ostringstream query_stream;
    query_stream << "CREATE TABLE " << OTN << "_monad_ms (\n"
		 << "    monad_ms_id INT NOT NULL,\n"
		 << "    object_id_d INT NOT NULL,\n"
		 << "    mse_first INT NOT NULL,\n"
		 << "    mse_last INT NOT NULL,\n"
		 << "    is_first CHAR(1) NOT NULL DEFAULT 'N',\n"
		 << "    PRIMARY KEY (monad_ms_id, mse_first)\n" // We need both mse_first and monad_ms_id, since monad_ms id is not a unique identifier
		 << ")";
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateObjectType", query_stream.str().c_str());
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create indices on OT_monad_ms
    if (!CreateIndicesOnOTMonadMs(OTN)) {
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create features
    f_it = feature_infos.begin();
    while (f_it != f_iend) {
      if (!CreateFeature(f_it->get_name(),
			 object_type_id,
			 f_it->get_type(),
			 f_it->get_default_value(),
			 f_it->get_is_computed() != 0)) {
	DEBUG_X_FAILED("EMdFDB::CreateObjectType", "Creating feature");
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }


      // Advance iterator
      ++f_it;
    }

    // Create "self" feature
    if (!CreateFeature("self",
		       object_type_id,
		       encode_feature_type_id(FEATURE_TYPE_ID_D),
		       NIL_AS_STRING,
		       true)) {
      DEBUG_X_FAILED("EMdFDB::CreateObjectType", "Creating feature 'self'");
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }
    

    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::CreateObjectType");
	return false;
      }
    }

    // Add to cache 
    add_object_type_to_cache(object_type_id, OTN);

    // The command was a success
    return true;
  }
}

void EMdFDB::add_object_type_to_cache(id_d_t object_type_id, const std::string& object_type_name)
{
  // Normalize object type name
  std::string OTN = NormalizeOTName(object_type_name);

  // Add to cache
  m_object_type_cache_table->start_new_row();
  m_object_type_cache_table->append(id_d2string(object_type_id));
  m_object_type_cache_table->append(OTN);

  // Add to map
  m_object_type_cache_map.insert(IDD2StringMap::value_type(object_type_id, OTN));
}

bool EMdFDB::DropObjectType(const std::string& object_type_name, id_d_t object_type_id)
{
  if (pConn == 0)
    return false;
  else {
    // Begin transaction
    bool bDoCommit;
    bDoCommit = pConn->BeginTransaction();

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Drop table OT_objects
    {
      std::ostringstream query_stream;
      query_stream << "DROP TABLE " << OTN << "_objects";
      if (!pConn->Exec(query_stream.str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropObjectType", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Drop table OT_monad_ms
    {
      std::ostringstream query_stream;
      query_stream << "DROP TABLE " << OTN << "_monad_ms";
      if (!pConn->Exec(query_stream.str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropObjectType", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Delete all features
    {
      std::ostringstream query_stream;
      query_stream
	<< "DELETE FROM features\n"
	<< "WHERE object_type_id = " << object_type_id;
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropObjectType", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Delete from table 'object_types'
    {
      std::ostringstream query_stream;
      query_stream
	<< "DELETE FROM object_types\n"
	<< "WHERE object_type_id = " << object_type_id;
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropObjectType", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Delete all traces from table 'covered_by'
    {
      std::ostringstream query_stream;
      query_stream
	<< "DELETE FROM covered_by\n"
	<< "WHERE covered_by_id = " << object_type_id << "\n" 
	<< "      OR object_type_id = " << object_type_id;
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropObjectType", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Delete all traces from table 'buildable_from'
    {
      std::ostringstream query_stream;
      query_stream
	<< "DELETE FROM buildable_from\n"
	<< "WHERE buildable_from_id = " << object_type_id << "\n" 
	<< "      OR object_type_id = " << object_type_id;
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropObjectType", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Commit transaction
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::DropObjectType");
	return false;
      }
    }

    // Delete from cache
    delete_object_type_from_cache(object_type_id);

    // If we got this far, the command was a success
    return true;
  }
}

void EMdFDB::delete_object_type_from_cache(id_d_t object_type_id)
{
  // Delete from table cache
  // First try the table cache
  TableIterator i;
  i = m_object_type_cache_table->find(id_d2string(object_type_id), 1);
  if (i.hasNext()) {
    m_object_type_cache_table->erase(i);
  }

  // Delete from the map cache
  IDD2StringMap::iterator mi;
  mi = m_object_type_cache_map.find(object_type_id);
  if (mi != m_object_type_cache_map.end()) {
    m_object_type_cache_map.erase(mi);
  }
}

bool EMdFDB::ObjectTypeExists(id_d_t object_type_id,
			      /* out */ bool& result, 
			      /* out */ std::string& object_type_name)
{
  if (pConn == 0)
    return false;
  else {
    // First try the map cache
    IDD2StringMap::const_iterator mi;
    mi = m_object_type_cache_map.find(object_type_id);
    if (mi != m_object_type_cache_map.end()) {
      object_type_name = mi->second;
      result = true;
      return true;
    }

    // Then try the table cache
    TableIterator i;
    i = m_object_type_cache_table->find(id_d2string(object_type_id), 1);
    if (i.hasNext()) {
      object_type_name = m_object_type_cache_table->get_column(i, 2);
      result = true;
      return true;
    }

    // Declare szQuery
    char szQuery[512];

    // Build and execute query
    sprintf(szQuery, 
	    "SELECT object_type_name\n"
	    "FROM object_types\n"
	    "WHERE object_type_id = '%ld'", object_type_id);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::ObjectTypeExists", szQuery);

      // Release results
      pConn->ReleaseResults();

      return false;
    }

    // Is there at least one tuple? (There should be either 1 or none)
    if (pConn->NoOfResultTuples() > 1) {
      std::cerr << "EMdFDB::ObjectTypeExists: There were more than one result tuples." << std::endl;
      assert(false);
    } else if (pConn->NoOfResultTuples() == 1) {
      try {
	// Access result
	if (!pConn->AccessTuple(0, 0, "EMdFDB::ObjectTypeExists", object_type_name)) {
	  // DEBUG information has already been printed if it is on.

	  // Release results
	  pConn->ReleaseResults();

	  return false;
	}
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::ObjectTypeExists");

	// Release results
	pConn->ReleaseResults();


	// There was a NULL value exception, hence we return an error
	return false;
      }
      
      // We did find the object type
      result = true;
    } else { 

      // Release results
      pConn->ReleaseResults();

      // We did not find the object type
      result = false;
    }

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Add to cache if we found it
    if (result) {
      add_object_type_to_cache(object_type_id, OTN);
    }

    // Release results
    pConn->ReleaseResults();

    // If we got this far, then the command as a whole was a success, 
    // no matter whether we found the object type or not.
    return true;
  }
}

bool EMdFDB::ObjectTypeExists(const std::string& object_type_name, 
                              /* out */ bool& result, 
                              /* out */ id_d_t& object_type_id) 
{
  if (pConn == 0)
    return false;
  else {
    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // First try the cache
    TableIterator i;
    i = m_object_type_cache_table->find(OTN, 2);
    if (i.hasNext()) {
      object_type_id = string2id_d(m_object_type_cache_table->get_column(i, 1));
      result = true;
      return true;
    }

    // Build and execute query
    std::ostringstream query_stream;
    query_stream << "SELECT object_type_id\n"
		 << "FROM object_types\n"
		 << "WHERE object_type_name = '" << OTN << "'";
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::ObjectTypeExists", query_stream.str().c_str());
      // Release results
      pConn->ReleaseResults();

      return false;
    }

    // Is there at least one tuple? (There should be either 1 or none)
    if (pConn->NoOfResultTuples() > 1) {
      std::cerr << "EMdFDB::ObjectTypeExists: There was more than one result tuple." << std::endl;
      assert(false);
    } else if (pConn->NoOfResultTuples() == 1) {
      // Access result
      try {
	if (!pConn->AccessTuple(0, 0, "EMdFDB::ObjectTypeExists", object_type_id)) {
	  // DEBUG information has already been printed if it is on.
	  // Release results
	  pConn->ReleaseResults();

	  return false;
	}
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::ObjectTypeExists");

	// There was a NULL value exception, hence we return an error
	return false;
      }
     
 
      // We did find the object type
      result = true;
    } else { 
      // We did not find the object type
      result = false;
    }

    // Add results to cache if found
    if (result) {
      add_object_type_to_cache(object_type_id, OTN);
    }

    // Release results
    pConn->ReleaseResults();

    // If we got this far, then the command as a whole was a success, 
    // no matter whether we found the object type or not.
    return true;
  }
}


//
// Enumeration manipulation
bool EMdFDB::CreateEnum(const std::string& enum_name, 
                        /* out */ id_d_t& enum_id)
{
  if (pConn == 0)
    return false;
  else {
    // Begin transaction
    bool bDoCommit;
    bDoCommit = pConn->BeginTransaction();

    // Get auto-generated ID
    if (!GetNextOtherID(enum_id)) {
      DEBUG_X_FAILED("EMdFDB::CreateEnum", "GetNextOtherID");
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Build and execute query
    std::ostringstream query_stream;
    query_stream 
      << "INSERT INTO enumerations (enum_id, enum_name)\n"
      << "VALUES ( " << enum_id << ", '" << enum_name << "' )";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateEnum", query_stream.str().c_str());
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Commit transaction
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::CreateEnum");
	return false;
      }
    }

    // Add to cache
    add_enum_to_cache(enum_id, std::string(enum_name));

    // If we got this far, the command succeeded.
    return true;
  }
}

void EMdFDB::add_enum_to_cache(id_d_t enum_id, const std::string& enum_name)
{
  m_enum_cache->start_new_row();
  m_enum_cache->append(id_d2string(enum_id));
  m_enum_cache->append(enum_name);
}

bool EMdFDB::DropEnum(id_d_t enum_id)
{
  if (pConn == 0)
    return false;
  else {
    // Begin transaction
    bool bDoCommit;
    bDoCommit = pConn->BeginTransaction();

    // Build and execute query
    std::ostringstream query_stream1;
    query_stream1
      << "DELETE\n"
      << "FROM enumeration_constants\n"
      << "WHERE enum_id = " << enum_id;
    if (!pConn->Exec(query_stream1.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::DropEnum", query_stream1.str().c_str());
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Build and execute query
    std::ostringstream query_stream2;
    query_stream2 
      << "DELETE\n"
      << "FROM enumerations\n"
      << "WHERE enum_id = " << enum_id;
    if (!pConn->Exec(query_stream2.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::DropEnum", query_stream2.str().c_str());
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Commit transaction
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::DropEnum");
	return false;
      }
    }

    // Delete from cache
    TableIterator i;
    i = m_enum_cache->find(id_d2string(enum_id), 1);
    if (i.hasNext()) {
      m_enum_cache->erase(i);
    }

    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::EnumExists(id_d_t enum_id,
			/* out */ bool& result,
			/* out */ std::string& enum_name)
{
  if (pConn == 0)
    return false;
  else {
    // First try the cache
    TableIterator i;
    i = m_enum_cache->find(id_d2string(enum_id), 1);
    if (i.hasNext()) {
      enum_name = m_enum_cache->get_column(i, 2);
      result = true;
      return true;
    }

    // Build and execute query
    std::ostringstream query_stream;
    query_stream 
      << "SELECT enum_name\n"
      << "FROM enumerations\n"
      << "WHERE enum_id = " << enum_id;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::EnumExists", query_stream.str().c_str());
      return false;
    }

    // Is there at least one tuple? (There should be either 1 or none)
    if (pConn->NoOfResultTuples() > 1) {
      std::cerr << "EMdFDB::EnumExists: There were more than one result tuples." << std::endl;
      assert(false);
    } else if (pConn->NoOfResultTuples() == 1) {
      // Access result
      try {
	if (!pConn->AccessTuple(0, 0, "EMdFDB::EnumExists", enum_name)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::EnumExists");

	// Release results
	pConn->ReleaseResults();

	// There was a NULL value exception, hence we return an error
	return false;
      }

      
      // We did find the enumeration
      result = true;
    } else { 
      // We did not find the enumeration
      result = false;
    }

    // Add results to cache if found
    if (result) {
      add_enum_to_cache(enum_id, std::string(enum_name));
    }

    // Release results
    pConn->ReleaseResults();

    // If we got this far, then the command as a whole was a success, 
    // no matter whether we found the enum or not.
    return true;
  }
}

bool EMdFDB::EnumExists(const std::string& enum_name, 
                        /* out */ bool& result,
                        /* out */ id_d_t& enum_id)
{
  if (pConn == 0)
    return false;
  else {
    // First try the cache
    TableIterator i;
    i = m_enum_cache->find(std::string(enum_name), 2);
    if (i.hasNext()) {
      enum_id = string2id_d(m_enum_cache->get_column(i, 1));
      result = true;
      return true;
    }

    // Build and execute query
    std::ostringstream query_stream;
    query_stream << "SELECT enum_id\n"
		 << "FROM enumerations\n"
		 << "WHERE enum_name = '" << enum_name << "'";
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::EnumExists", query_stream.str());
      return false;
    }

    // Is there at least one tuple? (There should be either 1 or none)
    if (pConn->NoOfResultTuples() > 1) {
      std::cerr << "EMdFDB::EnumExists: There were more than one result tuples." << std::endl;
      assert(false);
    } else if (pConn->NoOfResultTuples() == 1) {
      // Access result
      try {
	if (!pConn->AccessTuple(0, 0, "EMdFDB::EnumExists", enum_id)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::EnumExists");

	// Release results
	pConn->ReleaseResults();

	// There was a NULL value exception, hence we return an error
	return false;
      }

      
      // We did find the enumeration
      result = true;
    } else { 
      // We did not find the enumeration
      result = false;
    }

    // Add results to cache if found
    if (result) {
      add_enum_to_cache(enum_id, std::string(enum_name));
    }

    // Release results
    pConn->ReleaseResults();

    // If we got this far, then the command as a whole was a success, 
    // no matter whether we found the enum or not.
    return true;
  }
}

bool EMdFDB::GetDefault(id_d_t enum_id, 
                        /* out */ bool& result,
			/* out */ std::string& default_constant_name)
{
  if (pConn == 0)
    return false;
  else {
    // First try cache
    TableIterator i;
    i = m_enum_const_cache->find(id_d2string(enum_id), 1, bool2string(true), 4, false);
    if (i.hasNext()) {
      default_constant_name = m_enum_const_cache->get_column(i, 2);
      result = true;
      return true;
    }

    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT enum_value_name\n"
      << "FROM enumeration_constants\n"
      << "WHERE enum_id = " << enum_id << "\n"
      << "      AND is_default = 'Y'";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetDefault", query_stream.str().c_str());
      return false;
    }

    // Is there at least one tuple? (There should be either 1 or none)
    if (pConn->NoOfResultTuples() > 1) {
      std::cerr << "EMdFDB::GetDefault: There were more than one result tuples." << std::endl;
      assert(false);
    } else if (pConn->NoOfResultTuples() == 1) {
      // Access result
      try {
	if (!pConn->AccessTuple(0, 0, "EMdFDB::GetDefault", default_constant_name)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetDefault");

	// Release results
	pConn->ReleaseResults();

	// There was a NULL value exception, hence we return an error
	return false;
      }

      
      // We did find the default
      result = true;
    } else { 
      // We did not find the default
      result = false;
    }

    // Release results
    pConn->ReleaseResults();

    // If we got this far, then the command as a whole was a success, 
    // no matter whether we found the default or not.
    return true;
  }
}


//
// Enumeration constant manipulation
bool EMdFDB::CreateEnumConstant(id_d_t enum_id,
				const enum_const_info& ec_info)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "INSERT INTO enumeration_constants (\n"
      << "    enum_id,\n"
      << "    enum_value_name,\n"
      << "    value,\n"
      << "    is_default\n"
      << ")\n"
      << "VALUES (\n"
      << "    "  << enum_id
      << ", '" << ec_info.get_name()
      << "', " << ec_info.get_value()
      << ", '" << bool2char(ec_info.get_is_default()) << "'\n"
      << ")";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateEnumConst", query_stream.str().c_str());
      return false;
    }

    // Add to cache
    if (!add_enum_const_to_cache(enum_id, ec_info.get_name(), ec_info.get_value(), ec_info.get_is_default()))
      return false;

    // If we got this far, the command succeeded.
    return true;
  }
}


bool EMdFDB::add_enum_const_to_cache(id_d_t enum_id, const std::string& enum_const_name, long value, bool is_default)
{
  m_enum_const_cache->start_new_row();
  m_enum_const_cache->append(id_d2string(enum_id));
  m_enum_const_cache->append(enum_const_name);
  m_enum_const_cache->append(long2string(value));
  m_enum_const_cache->append(bool2string(is_default));

  // Get enum-name
  std::string enum_name;
  bool bEnumExists;
  if (!EnumExists(enum_id, bEnumExists, enum_name)) {
    DEBUG_X_FAILED("EMdFDB::add_enum_const_to_cache", "EnumExists returned with DB error");
    return false;
  }
  if (!bEnumExists) {
    DEBUG_X_FAILED("EMdFDB::add_enum_const_to_cache", "EnumExists said that enum didn't exist.");
    return false;
  }

  // Append enum_name
  m_enum_const_cache->append(enum_name);

  return true;
}

bool EMdFDB::EnumConstExists(const std::string& enum_const_name, 
			     id_d_t enum_id, 
                             /* out */ bool& result,
                             /* out */ long& value,
                             /* out */ bool& is_default)
{
  if (pConn == 0)
    return false;
  else {
    result = false; // Assume it isn't there
    // First try the cache
    TableIterator i;
    std::string enum_id_string = id_d2string(enum_id);
    i = m_enum_const_cache->find(enum_id_string, 1, enum_const_name, 2, false);
    if (i.hasNext()) {
      value = string2long(m_enum_const_cache->get_column(i, 3));
      is_default = string2bool(m_enum_const_cache->get_column(i, 4));
      result = true;
      return true;
    }

    // Build and execute query
    std::ostringstream query_stream;
    query_stream << "SELECT enum_value_name, value, is_default\n"
		 << "FROM enumeration_constants\n"
		 << "WHERE enum_id = " << enum_id 
		 << " AND enum_value_name = '" << enum_const_name << "'\n";
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::EnumConstExists", query_stream.str());
      return false;
    }

    std::string enum_const_name_from_db;
    // Is there at least one tuple? 
    // MySQL will return all that have the same enum_value_name
    // case-insensitively.
    if (pConn->NoOfResultTuples() > 0) {
      for (int i = 0; !result && (i < pConn->NoOfResultTuples()); ++i) {
	try {
	  // Get constant name
	  if (!pConn->AccessTuple(i, 0, "EMdFDB::EnumConstExists", enum_const_name_from_db)) {
	    // DEBUG information has already been printed if it is on.
	    return false;
	  }

	  // Get value
	  if (!pConn->AccessTuple(i, 1, "EMdFDB::EnumConstExists", value)) {
	    // DEBUG information has already been printed if it is on.
	    return false;
	  }
	
	  // Get is_default
	  if (!pConn->AccessTuple(i, 2, "EMdFDB::EnumConstExists", is_default)) {
	    // DEBUG information has already been printed if it is on.
	    return false;
	  }
	
	} catch (EMdFNULLValueException e) {
	  DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::EnumConstExists");

	  // Release results
	  pConn->ReleaseResults();

	  // There was a NULL value exception, hence we return an error
	  return false;
	}

	// Check whether we have really found it.
	// Do this case-sensitively.
	if (strcmp_case(enum_const_name_from_db, enum_const_name) == 0) {
	  // We did find the enumeration constant
	  result = true;
	}
      }
    } else { 
      // We did not find the enumeration constant
      result = false;
    }

    // Add to cache if found
    if (result) {
      if (!add_enum_const_to_cache(enum_id, enum_const_name, value, is_default))
	return false;
    }

    // Release results
    pConn->ReleaseResults();

    // If we got this far, then the command as a whole was a success, 
    // no matter whether we found the enumeration constant or not.
    return true;
  }
}



// Returns the string-representation of an enumeration constant in 
// enum enum_name with the value value.  If more than one enum constant
// has the same value, it is undefined which of them is returned.
bool EMdFDB::GetEnumConstNameFromValue(long value,
				       const std::string& enum_name,
				       /* out */ std::string& enum_const_name)
  
{
  if (pConn == 0)
    return false;
  else {
    // Declare bExists
    bool bExists = false;
      
    // First try the cache
    TableIterator i;
    std::string enum_value_string = long2string(value);
    i = m_enum_const_cache->find(enum_value_string, 3, enum_name, 5, false);
    if (i.hasNext()) {
      enum_const_name = i.getColumn(2);
      return true;
    }
      
    // Get enum_id
    id_d_t enum_id;
    bool bEnumExists;
    if (!EnumExists(enum_name, bEnumExists, enum_id)) {
      DEBUG_X_FAILED("EMdFDB::GetEnumConstNameFromValue", "EnumExists returned with DB error.");
      return false;
    }
    if (!bEnumExists) {
      DEBUG_X_FAILED("EMdFDB::GetEnumConstNameFromValue", "enumeration did not exist");
      return false;
    }
      
    // Build and execute query
    std::ostringstream query_stream;
    query_stream << "SELECT enum_value_name, is_default\n"
		 << "FROM enumeration_constants\n"
		 << "WHERE enum_id = '" << enum_id << "'"
		 << " AND value = '" << value << "'\n";
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetEnumConstNameFromValue", query_stream.str());
      return false;
    }
      
    // Is there at least one tuple? (There may be none, 1, or more)
    bool is_default;
    if (pConn->NoOfResultTuples() > 0) {
      try {
	// Get value
	if (!pConn->AccessTuple(0, 0, "EMdFDB::GetEnumConstNameFromValue", enum_const_name)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}
	
	// Get is_default
	if (!pConn->AccessTuple(0, 1, "EMdFDB::EnumConstExists", is_default)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::EnumConstExists");

	// Release results
	pConn->ReleaseResults();

	// There was a NULL value exception, hence we return an error
	return false;
      }
	
      // So it existed
      bExists = true;
    } else { 
      bExists = false;
    }
      
    // Release results
    pConn->ReleaseResults();

    if (bExists) {
      if (!add_enum_const_to_cache(enum_id, enum_const_name, value, is_default))

	return false;
    } else {
      DEBUG_X_FAILED("EMdFDB::GetEnumConstNameFromValue", "enumeration constant with value " + enum_value_string + " did not exist");
      return false;
    }
      
    // If we got this far, then the command as a whole was a success, 
    // no matter whether we found the enumeration constant or not.
    return true;
  }
}


// Only gets one of them if more than one
// TODO: Fix this
bool EMdFDB::EnumConstExists(long value,
			     id_d_t enum_id,
			     bool& bExists,
			     /* out */ std::string& enum_const_name,
			     /* out */ bool& is_default)

{
  if (pConn == 0)
    return false;
  else {
    // Assume it isn't there
    bExists = false;

    // First try the cache
    TableIterator i;
    std::string enum_value_string = long2string(value);
    std::string enum_id_string = id_d2string(enum_id);
    i = m_enum_const_cache->find(enum_id_string, 1, enum_value_string, 3, false);
    if (i.hasNext()) {
      enum_const_name = i.getColumn(2);
      bExists = true;
      return true;
    }

    // Build and execute query
    std::ostringstream query_stream;
    query_stream << "SELECT enum_value_name, is_default\n"
		 << "FROM enumeration_constants\n"
		 << "WHERE enum_id = '" << enum_id << "'"
		 << " AND value = '" << value << "'\n";
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::EnumConstExists", query_stream.str());
      return false;
    }

    // Is there at least one tuple? (There may be none, 1, or more)
    if (pConn->NoOfResultTuples() > 0) {
      try {
	// Get value
	if (!pConn->AccessTuple(0, 0, "EMdFDB::EnumConstExists", enum_const_name)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}
	
	// Get is_default
	if (!pConn->AccessTuple(0, 1, "EMdFDB::EnumConstExists", is_default)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}

      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::EnumConstExists");

	// Release results
	pConn->ReleaseResults();

	// There was a NULL value exception, hence we return an error
	return false;
      }

      // We did find the enumeration constant
      bExists = true;
    } else { 
      // We did not find the enumeration constant
      bExists = false;
    }

    // Release results
    pConn->ReleaseResults();

    // Add to cache if found
    if (bExists) {
      if (!add_enum_const_to_cache(enum_id, enum_const_name, value, is_default))
	return false;
    }

    // If we got this far, then the command as a whole was a success, 
    // no matter whether we found the enumeration constant or not.
    return true;
  }
}

bool EMdFDB::DropEnumConst(id_d_t enum_id, const std::string& enum_const_name, long value)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream << "DELETE\n"
		 << "FROM enumeration_constants\n"
		 << "WHERE enum_id = " << enum_id 
		 << " AND enum_value_name = '" << enum_const_name << "'"
		 << " AND value = " << value;
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::DropEnumConst", query_stream.str());
      return false;
    }

    // Delete from cache
    TableIterator i;
    i = m_enum_const_cache->find(id_d2string(enum_id), 1, enum_const_name, 2, false);
    if (i.hasNext()) {
      m_enum_const_cache->erase(i);
    }

    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::UpdateEnumConst(const std::string& enum_const_name, 
			     id_d_t enum_id,
			     long old_value,
			     long new_value)
{
  if (pConn == 0)
    return false;
  else {
    //
    // Build and execute query
    //
    std::ostringstream query_stream;
    query_stream 
      << "UPDATE enumeration_constants\n"
      << "SET value = " << new_value << "\n"
      << "WHERE enum_id = " << enum_id << "\n"
      << "      AND enum_value_name = '" << enum_const_name << "'\n"
      << "      AND value = " << old_value;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::UpdateEnumConst", query_stream.str().c_str());
      return false;
    }

    //
    // Update in cache, if there
    //
    TableIterator i;
    i = m_enum_const_cache->find(std::string(enum_const_name), 2, false);
    if (i.hasNext()) {
      m_enum_const_cache->update_row(i, long2string(new_value), 3);
    }


    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::SetDefaultEnumConst(id_d_t enum_id,
				 const std::string& enum_value_name)
{
  if (pConn == 0)
    return false;
  else {
    // Declare bDoCommit
    bool bDoCommit;

    // Start transaction
    bDoCommit = pConn->BeginTransaction();

    // Remove current default
    std::ostringstream query_stream1;
    query_stream1
      << "UPDATE enumeration_constants\n"
      << "SET is_default = 'N'\n"
      << "WHERE enum_id = " << enum_id;
    if (!pConn->Exec(query_stream1.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::SetDefaultEnumConst", query_stream1.str().c_str());
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Set new default
    std::ostringstream query_stream2;
    query_stream2
      << "UPDATE enumeration_constants\n"
      << "SET is_default = 'Y'\n"
      << "WHERE enum_id = " << enum_id << "\n"
      << "      AND enum_value_name = '" << enum_value_name << "'";
    if (!pConn->Exec(query_stream2.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::SetDefaultEnumConst", query_stream2.str().c_str());
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Commit transaction
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::SetDefaultEnumConst");
	return false;
      }
    }

    // Update cache
    TableIterator i = m_enum_const_cache->iterator();
    std::string enum_id_as_string = id_d2string(enum_id);
    std::string enum_value_name_as_string = std::string(enum_value_name);
    while (i.hasNext()) {
      TableRowIterator sli;
      sli = i.current().iterator();
      // Is this the right enumeration?
      if (strcmp_nocase(sli.current(), enum_id_as_string) == 0) {
	sli.next();
	// Compare enum_value_name
	if (strcmp_case(sli.current(), enum_value_name) == 0)
	  m_enum_const_cache->update_row(i, bool2string(true), 4);
	else
	  m_enum_const_cache->update_row(i, bool2string(false), 4);
      }
      i.next();
    }


    // If we got this far, the command succeeded.
    return true;
  }
}

//
// Feature manipulation
bool EMdFDB::CreateFeature(const std::string& feature_name,
                           id_d_t object_type_id, 
                           id_d_t feature_type_id, 
                           const std::string& default_value, 
                           bool computed)  // Assumes it does not
  // exist
{
  if (pConn == 0)
    return false;
  else {
    // Build feature_info
    feature_info fi(feature_name, feature_type_id, 
		    default_value, computed);

    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "INSERT INTO features (\n"
      << "    object_type_id,\n"
      << "    feature_name,\n"
      << "    feature_type_id,\n"
      << "    default_value,\n"
      << "    computed\n"
      << ")\n"
      << "VALUES (\n"
      << "    " << object_type_id
      << ", '" << encode_feature_name(feature_name)
      << "', " << feature_type_id
      << ", '" << string2dbstring(default_value)
      << "', '" << bool2char(computed)
      << "'\n"
      << ")";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateFeature", query_stream.str().c_str());
      return false;
    }

    // Make sure we have all other features in cache
    std::list<feature_info> dummy;
    if (!GetFeaturesForObjectType(object_type_id, dummy))
      return false;

    // Add to cache
    add_feature_to_cache_if_not_already_there(object_type_id, feature_name, feature_type_id, default_value, computed);

    // If we got this far, the command succeeded.
    return true;
  }
}


void EMdFDB::add_feature_to_cache_if_not_already_there(id_d_t object_type_id, const std::string& feature_name, id_d_t feature_type_id, const std::string& default_value, bool is_computed)
{
  // Add it to the feature cache if not already there.
  TableIterator i;
  i = m_feature_cache->find(feature_name, 2, id_d2string(object_type_id), 1);
  if (!i.hasNext()) { // We didn't find it, so add it.
    add_feature_to_cache(object_type_id, 
			 feature_name,
			 feature_type_id, 
			 default_value, 
			 is_computed);
  }
}

void EMdFDB::add_feature_to_cache(id_d_t object_type_id, const std::string& feature_name, id_d_t feature_type_id, const std::string& default_value, bool is_computed)
{
  m_feature_cache->start_new_row();
  m_feature_cache->append(id_d2string(object_type_id));
  m_feature_cache->append(std::string(feature_name));
  m_feature_cache->append(id_d2string(feature_type_id));
  m_feature_cache->append(default_value);
  m_feature_cache->append(bool2string(is_computed));
}

bool EMdFDB::AddFeature(const std::string& object_type_name,
			id_d_t object_type_id,
			const feature_info& fi)
{
  if (pConn == 0)
    return false;
  else {
    bool bDoCommit;

    // Begin transaction
    bDoCommit = pConn->BeginTransaction();

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Add feature to "OT_objects" table
    {
      // Build and execute query
      std::ostringstream query_stream;
      query_stream
	<< "ALTER TABLE " << OTN << "_objects ADD ";
      // Last arg is false because PostgreSQL does not support NOT NULL
      // constraint on ALTER TABLE ADD COLUMN.
      MakeFeatureSQLType(query_stream, fi, false);

      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::AddFeature", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Add feature to "features" table
    if (!CreateFeature(fi.get_name(),
		       object_type_id,
		       fi.get_type(),
		       fi.get_default_value(),
		       fi.get_is_computed() != 0)) {
      DEBUG_X_FAILED("EMdFDB::AddFeature", "Creating feature");
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Commit transaction
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::AddFeature");
	return false;
      }
    }

    // Make sure we have all other features in cache
    std::list<feature_info> dummy;
    if (!GetFeaturesForObjectType(object_type_id, dummy))
      return false;

    // Add to cache
    add_feature_to_cache_if_not_already_there(object_type_id, fi.get_name(), fi.get_type(), fi.get_default_value(), fi.get_is_computed());

    // If we got this far, the command succeeded.
    return true;
  }

}

bool EMdFDB::DropFeature(const std::string& object_type_name,
                         const std::string& feature_name,
                         id_d_t object_type_id)
{
  if (pConn == 0)
    return false;
  else {
    bool bDoCommit;

    // Begin transaction
    bDoCommit = pConn->BeginTransaction();

    /*
     * PostgreSQL (7.0) doesn't know how to ALTER TABLE DROP column. 
     */
    /*

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Remove feature from "OT_objects" table
    {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream
    << "ALTER TABLE " << OTN << "_objects DROP "
    << encode_feature_name(feature_name);
    if (!pConn->Exec(query_stream.str().c_str())) {
    DEBUG_QUERY_FAILED("EMdFDB::DropFeature", query_stream.str().c_str());
    if (bDoCommit)
    pConn->AbortTransaction();
    return false;
    }
    }
    */

    // Remove feature from "features" table
    {
      // Build and execute query
      std::ostringstream query_stream;
      query_stream
	<< "DELETE FROM features\n"
	<< "WHERE object_type_id = " << object_type_id << "\n"
	<< "      AND feature_name = '" << encode_feature_name(feature_name) << "'";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropFeature", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Commit transaction
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::DropFeature");
	return false;
      }
    }

    // Delete from cache
    TableIterator i;
    i = m_feature_cache->find(std::string(feature_name), 2, id_d2string(object_type_id), 1);
    if (i.hasNext()) {
      m_feature_cache->erase(i);
    }

    // If we got this far, the command succeeded.
    return true;
  }
}


bool EMdFDB::FeatureExists(const std::string& feature_name,
                           id_d_t object_type_id,
                           /* out */ bool& result,
                           /* out */ id_d_t& feature_type_id,
                           /* out */ std::string& default_value,
                           /* out */ bool& is_computed)
{
  if (pConn == 0)
    return false;
  else {
    // First try the cache
    TableIterator i;
    i = m_feature_cache->find(std::string(feature_name), 2, id_d2string(object_type_id), 1);
    if (i.hasNext()) {
      feature_type_id = string2id_d(m_feature_cache->get_column(i, 3));
      default_value = m_feature_cache->get_column(i, 4);
      is_computed = string2bool(m_feature_cache->get_column(i, 5));
      result = true;
      return true;
    }

    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT feature_type_id, default_value, computed\n"
      << "FROM features\n"
      << "WHERE object_type_id = " << object_type_id << "\n"
      << "      AND feature_name = '" << encode_feature_name(feature_name) << "'";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::FeatureExists", query_stream.str().c_str());
      return false;
    }

    // Is there at least one tuple? (There should be either 1 or none)
    if (pConn->NoOfResultTuples() > 1) {
      std::cerr << "EMdFDB::FeatureExists: There were more than one result tuples." << std::endl;
      assert(false);
    } else if (pConn->NoOfResultTuples() == 1) {
      try {
	// Get feature_type_id
	if (!pConn->AccessTuple(0, 0, "EMdFDB::FeatureExists", feature_type_id)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}

	// Decode feature type
	id_d_t ft;
	long dummy_length;
	decode_feature_type_id(feature_type_id, ft, dummy_length);

	// Get default_value
	std::string defval;
	if (!pConn->AccessTuple(0, 1, "EMdFDB::FeatureExists", defval)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}

	// Convert default value
	default_value = dbstring2string(defval);
      
	// Get is_computed
	if (!pConn->AccessTuple(0, 2, "EMdFDB::FeatureExists", is_computed)) {
	  // DEBUG information has already been printed if it is on.
	  return false;
	}
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::FeatureExists");

	// Release results
	pConn->ReleaseResults();

	// There was a NULL value exception, hence we return an error
	return false;
      }
      
      // We did find the feature
      result = true;
    } else { 
      // We did not find the feature
      result = false;
    }

    // Release results
    pConn->ReleaseResults();

    // If found, add to cache
    if (result) {
      add_feature_to_cache_if_not_already_there(object_type_id, feature_name, feature_type_id, default_value, is_computed);
    }

    // If we got this far, then the command as a whole was a success, 
    // no matter whether we found the feature or not.
    return true;
  }
}

// Assumes everything exists 
// TODO: Make this work with computed features
bool EMdFDB::GetFeatures(const std::string& object_type_name,
			 const std::list<feature_info>& feature_infos,
			 id_d_t object_id_d,
			 /* out */ std::list<std::string>& results)
{
  if (pConn == 0)
    return false;
  else {
    // Make id_d list
    std::list<id_d_t> oid_list;
    oid_list.push_back(object_id_d);

    // Make string list
    std::list<std::list<std::string> > list_results;
    if (!GetFeatures(object_type_name, feature_infos, oid_list, list_results))
      return false;

    // Get results
    results = *(list_results.begin());

    // If we got this far, there were no DB errors
    return true;
  }
}

// Assumes everything exists 
// TODO: Make this work with computed features
bool EMdFDB::GetFeatures(const std::string& object_type_name,
			 const std::list<feature_info>& feature_infos,
			 id_d_t object_id_d,
			 /* out */ std::list<CEMdFValue*>& results)
{
  if (pConn == 0)
    return false;
  else {
    //
    // Build query
    //
    std::ostringstream query_stream;
    query_stream 
      << "SELECT object_id_d";
  
    // Add the names of all the stored features
    std::list<feature_info>::const_iterator fi_i(feature_infos.begin());
    std::list<feature_info>::const_iterator fi_iend(feature_infos.end());
    while (fi_i != fi_iend) {
      // Emit comma 
      query_stream << ", ";

      // Get lowercase name
      std::string locase_name;
      str_tolower(fi_i->get_name(), locase_name);
      
      // Is it self?
      // TODO: We should check for computed status when we add
      // ability to do computed features.
      if (locase_name == "self") {
	query_stream << "object_id_d";
      } else {
	query_stream << encode_feature_name(locase_name);
      }
      
      // Advance iterator
      ++fi_i;
    }

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);  

    // Add the FROM and WHERE clauses
    query_stream << "\nFROM " << OTN << "_objects"
		 << "\nWHERE "
		 << "object_id_d = " << object_id_d
		 << "\n";

    // Execute query
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetFeatures", query_stream.str().c_str());
      return false;
    }
  
    // Check that there were results
    long lNoOfResultTuples = pConn->NoOfResultTuples();
    if (lNoOfResultTuples == 0) {
      DEBUG_X_FAILED("EMdFDB::GetFeatures", "No results from SQL-query");
      return false;
    } else if (lNoOfResultTuples != 1) {
      DEBUG_X_FAILED("EMdFDB::GetFeatures", "Too many results from query");
      return false;
    }
  
    //
    // Get results
    //
    int field_no = 1;
    
    // Iterate through features
    fi_i = feature_infos.begin();
    while (fi_i != fi_iend) {
      // Get type of fi_i
      id_d_t feature_type;
      long dummy_length;
      decode_feature_type_id(fi_i->get_type(), feature_type, dummy_length);
      
	

      // Get value from DB
      std::string valueStr;
      long valueLong;
      id_d_t valueID_D;
      
      CEMdFValue *pValue;
      try {
	switch (feature_type) {
	case FEATURE_TYPE_ASCII:
	case FEATURE_TYPE_STRING:
	  if (!pConn->AccessTuple(0, field_no, "EMdFDB::GetFeatures", valueStr))
	    return false;

	  // Decode string if FEATURE_TYPE_STRING
	  if (feature_type == FEATURE_TYPE_STRING) {
	    valueStr = dbstring2string(valueStr);
	  }

	  pValue = new CEMdFValue(valueStr);
	  break;
	case FEATURE_TYPE_INTEGER:
	  if (!pConn->AccessTuple(0, field_no, "EMdFDB::GetFeatures", valueLong))
	    return false;
	  pValue = new CEMdFValue(kEVInt, valueLong);
	  break;
	case FEATURE_TYPE_ID_D:
	  if (!pConn->AccessTuple(0, field_no, "EMdFDB::GetFeatures", valueID_D))
	    return false;
	  pValue = new CEMdFValue(kEVID_D, valueID_D);
	  break;
	default: // ENUM
	  if (!pConn->AccessTuple(0, field_no, "EMdFDB::GetFeatures", valueLong))
	    return false;
	  pValue = new CEMdFValue(kEVEnum, valueLong);
	  break;
	}
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetFeatures");

	// Release results
	pConn->ReleaseResults();

	// There was a NULL value exception, hence we return an error
	return false;
      }


      // Push back the value
      results.push_back(pValue);

      // Advance field_no
      ++field_no;

      // Advance iterator
      ++fi_i;
    }

    // Release results
    pConn->ReleaseResults();

    // If we got this far, there were no DB errors
    return true;
  }
}

void local_make_id_d_list(const std::list<id_d_t> object_id_ds, std::list<std::list<std::string> >& results)
{
  std::list<id_d_t>::const_iterator ci(object_id_ds.begin());
  std::list<id_d_t>::const_iterator cend(object_id_ds.end());
  while (ci != cend) {
    // Add list so we have something to fill
    results.push_back(std::list<std::string>());

    // Add id_d
    results.rbegin()->push_back(id_d2string(*ci));

    // Advance iterator
    ++ci;
  }
}


// Assumes everything exists 
// TODO: Make this work with computed features
bool EMdFDB::GetFeatures(const std::string& object_type_name,
			 const std::list<feature_info>& feature_infos,
			 const std::list<id_d_t> object_id_ds,
			 /* out */ std::list<std::list<std::string> >& results)
{
  if (pConn == 0)
    return false;
  else {
    // Make sure the results are pristine
    results.clear();
    
    // Return now if the list of features is empty
    if (feature_infos.empty()) {
      local_make_id_d_list(object_id_ds, results);
      return true;
    }

    // Make list of stored features
    std::list<std::string> stored_feature_names;
    std::list<feature_info>::const_iterator fii(feature_infos.begin());
    std::list<feature_info>::const_iterator fiiend(feature_infos.end());
    while (fii != fiiend) {
      if (!fii->get_is_computed()) {
	stored_feature_names.push_back(encode_feature_name(fii->get_name()));
      }
      ++fii;
    }

    // Should we run the query?
    bool bRunQuery = !stored_feature_names.empty();
    
    if (bRunQuery) {
      if (!GetFeaturesByQuery(stored_feature_names, object_type_name,
			      feature_infos, object_id_ds, results))
	return false;
      else
	return true;
    } else {
      if (!GetComputedFeatures(object_type_name,
			       feature_infos, object_id_ds, results))
	return false;
      else
	return true;
    }
  }
}

bool EMdFDB::GetFeaturesByQuery(const std::list<std::string>& stored_feature_names, 
				const std::string& object_type_name,
				const std::list<feature_info>& feature_infos,
				const std::list<id_d_t> object_id_ds,
				/* out */ std::list<std::list<std::string> >& results)
{
  //
  // Build query
  //
  std::ostringstream query_stream;
  query_stream 
    << "SELECT object_id_d";
  
  // Add the names of all the stored features
  std::list<std::string>::const_iterator sfni(stored_feature_names.begin());
  std::list<std::string>::const_iterator sfniend(stored_feature_names.end());
  while (sfni != sfniend) {
    query_stream << ", " << *sfni;
    ++sfni;
  }

  // Normalize object type name
  std::string OTN = NormalizeOTName(object_type_name);  

  // Add the FROM and WHERE clauses
  query_stream << "\nFROM " << OTN << "_objects"
	       << "\nWHERE ";
  std::list<id_d_t>::const_iterator id_di(object_id_ds.begin());
  std::list<id_d_t>::const_iterator id_diend(object_id_ds.end());
  while (id_di != id_diend) {
    query_stream << "object_id_d = " << *id_di;
    ++id_di;
    if (id_di != id_diend) {
      query_stream << "\n or ";
    } else {
      query_stream << '\n';
    }
  }
  
  // Execute query
  if (!pConn->Exec(query_stream.str().c_str())) {
    DEBUG_QUERY_FAILED("EMdFDB::GetFeaturesByQuery", query_stream.str().c_str());
    return false;
  }
  
  // Check that there were results
  long lNoOfResultTuples = pConn->NoOfResultTuples();
  if (lNoOfResultTuples == 0) {
    DEBUG_X_FAILED("EMdFDB::GetFeaturesByQuery", "No results from SQL-query");
    return false;
  }
  
  try {  
    // Get results
    std::list<feature_info>::const_iterator fii;
    std::list<feature_info>::const_iterator fiiend(feature_infos.end());
    for (long TupleNo = 0; TupleNo < lNoOfResultTuples; ++TupleNo) {
      // Add list so that we have something to fill
      results.push_back(std::list<std::string>());

      // Get object id_d from DB
      long field_no = 0;
      std::string object_id_d;
      if (!pConn->AccessTuple(TupleNo, field_no, "EMdFDB::GetFeaturesByQuery", object_id_d))
	return false;

      results.rbegin()->push_back(object_id_d);
      ++field_no;

      // Iterate through features
      fii = feature_infos.begin();
      while (fii != fiiend) {
	if (fii->get_is_computed() == true) {
	  // Get result
	  std::string comp_result;
	  if (!GetComputedFeature(object_type_name, *fii, string2id_d(object_id_d), comp_result)) {
	    DEBUG_X_FAILED("EMdFDB::GetComputedFeatures", "");
	    return false;
	  }
	
	  // Add result
	  results.rbegin()->push_back(comp_result);
	} else {
	  // Get value from DB
	  std::string value;
	  if (!pConn->AccessTuple(TupleNo, field_no, "EMdFDB::GetFeaturesByQuery", value))
	    return false;

	  // Push back the real value
	  if (feature_type_id_is_STRING(fii->get_type())) {
	    results.rbegin()->push_back(dbstring2string(value));
	  } else {
	    results.rbegin()->push_back(value);
	  }
      
	  // Advance field_no
	  ++field_no;
	}
    
	// Advance iterator
	++fii;
      }
    }
  } catch (EMdFNULLValueException e) {
    DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetFeaturesByQuery");

    // Release results
    pConn->ReleaseResults();

    // There was a NULL value exception, hence we return an error
    return false;
  }


  // Release results
  pConn->ReleaseResults();

  // If we got this far, there were no DB errors
  return true;
}

//
// Private function called from GetFeatures when there are no 
// stored features to retrieve.
//
bool EMdFDB::GetComputedFeatures(const std::string& object_type_name,
				 const std::list<feature_info>& feature_infos,
				 const std::list<id_d_t> object_id_ds,
				 /* out */ std::list<std::list<std::string> >& results)
{
  // Get object id_d from list
  std::list<id_d_t>::const_iterator id_di(object_id_ds.begin());
  std::list<id_d_t>::const_iterator id_diend(object_id_ds.end());
  while (id_di != id_diend) {
    // Add list so we have something to fill
    results.push_back(std::list<std::string>());

    // Add id_d
    results.rbegin()->push_back(id_d2string(*id_di));

    // Iterate through features
    std::list<feature_info>::const_iterator fii(feature_infos.begin());
    std::list<feature_info>::const_iterator fiiend(feature_infos.end());
    while (fii != fiiend) {
      // Get result
      std::string comp_result;
      if (!GetComputedFeature(object_type_name, *fii, *id_di, comp_result)) {
	DEBUG_X_FAILED("EMdFDB::GetComputedFeatures", "");
	return false;
      }

      // Add result
      results.rbegin()->push_back(comp_result);

      // Advance iterator
      ++fii;
    }

    // Advance iterator
    ++id_di;
  }
  
  // If we got this far, there were no DB errors
  return true;
}

bool EMdFDB::GetComputedFeature(const std::string& object_type_name, 
				const feature_info& fi, 
				id_d_t object_id_d, 
				/* out */ std::string& comp_result)
{
  // We must be computed
  if (!fi.get_is_computed()) {
    DEBUG_X_FAILED("EMdFDB::GetComputedFeature", 
		   "Function called with a non-computed feature.\n"
		   "This is an internal error.  Please contact Ulrik Petersen\n"
		   "at <mailto:ulrikp@th.vu.nl> with a description\n"
		   "of the problem.");
    return false;
  }

  if (strcmp_nocase(fi.get_name(), "self") == 0) {
    // Feature is "self"
    // Push back id_d
    comp_result = id_d2string(object_id_d);

    // If we got this far, there were no DB errors
    return true;
  } else {
    // We don't yet support computed features other than "self"
    DEBUG_X_FAILED("EMdFDB::GetComputedFeature", 
		   "Computed features other than 'self' not yet supported.");
    return false;
  }
}

//
// Segment manipulation
// Assumes that segment does not exist
bool EMdFDB::CreateSegment(const std::string& segment_name, monad_m
			   first, monad_m last)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "INSERT INTO segments (segment_name, first_monad, last_monad)\n"
      << "VALUES ('" << segment_name << "', " << first << ", " << last << ")";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateSegment", query_stream.str().c_str());
      return false;
    }

    return true;
  }
}


bool EMdFDB::ThereAreSegments(/* out */ bool& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Declare szQuery
    char szQuery[512];

    // Build and execute query
    sprintf(szQuery, 
	    "SELECT segment_name\n"
	    "FROM segments\n"
	    );
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::ThereAreSegments", szQuery);
      return false;
    }

    /* Get Result */
    Result = pConn->NoOfResultTuples() != 0;

    /* If we got here, we were a success. */
    return true;
  }
}

//
// Checks that the range first..last fits in one segment.
// If there are no segments, Result is true.
//
bool EMdFDB::MonadsFitInSegment(monad_m first, monad_m last,
				/* out */ bool& Result)
{
  if (pConn == 0)
    return false;
  else {
    // First check whether there are segments
    bool bSegmentsExist;
    if (!ThereAreSegments(bSegmentsExist))
      return false;

    if (!bSegmentsExist) {
      Result = true;
      return true;
    }

    // Get segments
    std::list<segment_info> ls;
    if (!GetSegments(ls))
      return false;

    // Build set of monads
    CSet_of_monad_ms monads;
    std::list<segment_info>::const_iterator ci(ls.begin());
    std::list<segment_info>::const_iterator cend(ls.end());
    while (ci != cend) {
      monads.add(ci->get_first(), ci->get_last());
      ++ci;
    }

    // Check that first..last is part_of set of monads
    CSet_of_monad_ms input;
    input.add(first, last);
    Result = input.part_of(monads);

    /* If we got here, there were no DB errors. */
    return true;
  }
}


//
// Checks that the segment first..last does not overlap
// with any other segment.
// If there are no segments, Result is true.
//
bool EMdFDB::SegmentDoesNotOverlap(monad_m first, monad_m last,
				   /* out */ bool& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream 
      << "SELECT segment_name\n"
      << "FROM segments\n"
      << "WHERE NOT (" << last << " < first_monad\n"
      << "           OR " << first << " > last_monad)";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::SegmentDoesNotOverlap", query_stream.str().c_str());
      return false;
    }

    /* If there were no tuples, then the segment did not overlap 
     * with any other segment. */
    Result = pConn->NoOfResultTuples() == 0;

    /* If we got here, we were a success. */
    return true;
  }
}


bool EMdFDB::DropSegment(const std::string& segment_name)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream << "DELETE\n"
		 << "FROM segments\n"
		 << "WHERE segment_name = '" << segment_name << "'";
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::DropSegment", query_stream.str());
      return false;
    }

    /* If we got here, we were a success. */
    return true;
  }
}

bool EMdFDB::SegmentExists(const std::string& segment_name, 
			   /* out */ bool& Result,
			   /* out */ monad_m& first,
			   /* out */ monad_m& last)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream << "SELECT first_monad, last_monad\n"
		 << "FROM segments\n"
		 << "WHERE segment_name = '" << segment_name << "'";
    if (!pConn->Exec(query_stream.str())) {
      DEBUG_QUERY_FAILED("EMdFDB::SegmentExists", query_stream.str().c_str());
      return false;
    }

    /* If there were no tuples, the segment does not exist. */
    Result = pConn->NoOfResultTuples() != 0;

    /* If we did have tuples, we must get first and last. */    
    if (Result) {
      try {
	/* Get first */
	if (!pConn->AccessTuple(0, 0, "EMdFDB::SegmentExists", first))
	  return false;

	/* Get last */
	if (!pConn->AccessTuple(0, 1, "EMdFDB::SegmentExists", last))
	  return false;
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::SegmentExists");

	// Release results
	pConn->ReleaseResults();

	// There was a NULL value exception, hence we return an error
	return false;
      }

    }

    // Release results
    pConn->ReleaseResults();
    /* If we got here, we were a success. */
    return true;
  }
}






//
// Covered_by and buildable_from manipulation
bool EMdFDB::CreateCoveredByRelationship(id_d_t object_type_id,
                                         id_d_t covered_by_id)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "INSERT INTO covered_by (\n"
      << "    object_type_id,\n"
      << "    covered_by_id\n"
      << ")\n"
      << "VALUES (\n"
      << "    " << object_type_id << ", " << covered_by_id << "\n"
      << ")";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateCoveredByRelationship", query_stream.str().c_str());
      return false;
    }

    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::CreateBuildableFromRelationship(id_d_t object_type_id,  
					     id_d_t buildable_from_id) 
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "INSERT INTO buildable_from (\n"
      << "    object_type_id,\n"
      << "    buildable_from_id\n"
      << ")\n"
      << "VALUES (\n"
      <<"    " << object_type_id << ", " << buildable_from_id << "\n"
      <<")";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateBuildableFromRelationship", query_stream.str().c_str());
      return false;
    }

    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::DropCoveredByRelationship(id_d_t object_type_id,
                                       id_d_t covered_by_id)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream 
      << "DELETE FROM covered_by\n"
      << "WHERE object_type_id = " << object_type_id << "\n"
      << "      AND covered_by_id = " << covered_by_id;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::DropCoveredByRelationship", query_stream.str().c_str());
      return false;
    }

    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::DropBuildableFromRelationship(id_d_t object_type_id,   
					   id_d_t buildable_from_id)  
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "DELETE FROM buildable_from\n"
      << "WHERE object_type_id = " << object_type_id << "\n"
      << "      AND buildable_from_id = " << buildable_from_id;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::DropBuildableFromRelationship", query_stream.str().c_str());
      return false;
    }

    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::CoveredByRelationshipExists(id_d_t object_type_id,
                                         id_d_t covered_by_id,
                                         /* out */ bool& result)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT covered_by_id\n"
      << "FROM covered_by\n"
      << "WHERE object_type_id = " << object_type_id << "\n"
      << "      AND covered_by_id = " << covered_by_id;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CoveredByRelationshipExists", query_stream.str().c_str());
      return false;
    }

    // Is there at least one tuple? (There should be either 1 or none)
    if (pConn->NoOfResultTuples() > 1) {
      std::cerr << "EMdFDB::CoveredByRelationshipExists: There were more than one result tuples." << std::endl;
      assert(false);
    } else if (pConn->NoOfResultTuples() == 1) {
      result = true;
    } else {
      result = false;
    }

    // If we got this far, the command was as success,
    // regardless of whether the relationship existed or not.
    return true;
  }
}

bool EMdFDB::BuildableFromRelationshipExists(id_d_t object_type_id, 
					     id_d_t buildable_from_id, 
					     /* out */ bool& result)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT buildable_from_id\n"
      << "FROM buildable_from\n"
      << "WHERE object_type_id = " << object_type_id << "\n"
      << "      AND buildable_from_id = " << buildable_from_id;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::BuildableFromRelationshipExists", query_stream.str().c_str());
      return false;
    }

    // Is there at least one tuple? (There should be either 1 or none)
    if (pConn->NoOfResultTuples() > 1) {
      std::cerr << "EMdFDB::BuildableFromRelationshipExists: There were more than one result tuples." << std::endl;
      assert(false);
    } else if (pConn->NoOfResultTuples() == 1) {
      result = true;
    } else {
      result = false;
    }

    // If we got this far, the command was as success,
    // regardless of whether the relationship existed or not.
    return true;
  }
}

// 
// Monad manipulation
//    bool EMdFDB::InsertMonads(const MonadSet& ms)
//    bool EMdFDB::DeleteMonads(const MonadSet& ms)


std::string EMdFDB::ErrorMessage(void)
{
  if (pConn == 0) {
    return ""; 
  } else {
    std::string remote_error;
    std::string local_error;
    std::string out;
    pConn->ErrorMessage(remote_error);
    local_error = get_local_error();
    if ((remote_error.length() + local_error.length()) > 0) {
      out = "Remote error message:\n"
	+ remote_error + ((remote_error.length() > 0) ? "\n" : "")
	+ "---------------------\n"
	+ "Local error message:\n"
	+ local_error + ((local_error.length() > 0) ? "\n" : "")
	+ "---------------------\n";
    }
    return out;
  }
}

bool EMdFDB::ConnectionOk(void)
{
  if (pConn == 0)
    return false;
  else {
    return pConn->ConnectionOk();
  }
}

void EMdFDB::append_local_error(std::string error)
{
  m_local_errormessage += error;
}

void EMdFDB::clear_local_error(void)
{
  m_local_errormessage = "";
  if (pConn) {
    pConn->clear_local_error();
  }
}

std::string EMdFDB::get_local_error(void) 
{ 
  std::string result;
  result = m_local_errormessage + "\n";
  if (pConn) {
    result += pConn->get_local_error();
  }
  return result; 
}

bool EMdFDB::GetObjectsHavingMonadsInMonadSet(const std::string object_type_name,
					      const CSet_of_monad_ms& monad_ms,
					      std::set<id_d_t>& objects_already_seen,					      
					      /* out */ CTable& result)
{
  // Normalize object type name
  std::string OTN = NormalizeOTName(object_type_name);


  //
  // Build WHERE clause
  //
  std::ostringstream where_clause;
  where_clause << "(mse_first <= " << monad_ms.last()
	       << " AND mse_last >= " << monad_ms.first()
	       << ") AND (" << '\n';

  SOMConstIterator monad_ci = monad_ms.const_iterator();
  while (monad_ci.hasNext()) {
    Monad_Set_Element mse = monad_ci.next();
    where_clause 
      << "(mse_last >= " 
      << mse.first() 
      << " AND "
      << "mse_first <= "
      << mse.last()
      << ")";
    if (monad_ci.hasNext()) {
      where_clause 
	<< "\n OR ";
    }
  }

  where_clause << ")";



  // Build and execute query
  std::ostringstream query_stream;
  query_stream
    << "SELECT object_id_d, mse_first, mse_last\n"
    << "FROM " << OTN << "_monad_ms\n"
    << "WHERE " << where_clause.str();
  if (!pConn->Exec(query_stream.str().c_str())) {
    DEBUG_QUERY_FAILED("EMdFDB::GetObjectsHavingMonadsInMonadSet", query_stream.str().c_str());
    return false;
  }

    
  int NoOfTuples = pConn->NoOfResultTuples();
  int tuple;
  try {
    for (tuple = 0; tuple < NoOfTuples; ++tuple) {
      // Get data from tuple
      id_d_t oid;
      if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetObjectsHavingMonadsInMonadSet", oid))
	return false;
      monad_m mse_first;
      if (!pConn->AccessTuple(tuple, 1, "EMdFDB::GetObjectsHavingMonadsInMonadSet", mse_first))
	return false;
      monad_m mse_last;
      if (!pConn->AccessTuple(tuple, 2, "EMdFDB::GetObjectsHavingMonadsInMonadSet", mse_last))
	return false;

      // Construct data
      // Only add if not already seen
      if (objects_already_seen.find(oid) == objects_already_seen.end()) {
	// Add to set of already-seen objects
	objects_already_seen.insert(oid);

	// Figure out which monad to add
	CSet_of_monad_ms temp_m;
	temp_m.add(mse_first, mse_last);
	monad_m monad_to_add = CSet_of_monad_ms::intersect(temp_m, monad_ms).first();

	// Add to result
	result.start_new_row();
	result.append(object_type_name);
	result.append(monad_m2string(monad_to_add));
	result.append(id_d2string(oid));
      }
    }
  } catch (EMdFNULLValueException e) {
    DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetObjectsHavingMonadsInMonadSet");

    // Release results
    pConn->ReleaseResults();

    // There was a NULL value exception, hence we return an error
    return false;
  }
    
  // Release results
  pConn->ReleaseResults();

  // If we got this far, the command was as success.
  return true;
}


#define MAX_MSES (200)

bool EMdFDB::GetObjectsHavingMonadsInMonadSet(const std::string object_type_name,
					      const CSet_of_monad_ms& monad_ms,
					      /* out */ CTable& result)
{
  if (pConn == 0)
    return false;
  else {

    // Declare set of objects already seen.  This is initially empty,
    // but is (possibly) added to every time we call the other
    // member function.
    std::set<id_d_t> objects_already_seen;

    // Get const_iterator from monad_ms
    SOMConstIterator ci = monad_ms.const_iterator();

    // Declare temporary set of monads which we will use for calling
    // the other function
    CSet_of_monad_ms temp_som;

    // Start with a count of 0
    int mse_count = 0;

    // Iterate though monad_ms.
    while (ci.hasNext()) {
      // Add current MSE to temp_som and advance iterator
      temp_som.AddMSE(ci.next());

      // Increment count of mses
      mse_count++;

      // If we have reached the maximum count of MSEs, 
      // or ci has now advanced past the end,
      // then run query.
      if ((mse_count == MAX_MSES) || (!ci.hasNext())) {
	// We have reached maximal mse count, 
	// or there are no more mses left in monad_ms.

	// Run query
	if (!GetObjectsHavingMonadsInMonadSet(object_type_name, temp_som, objects_already_seen, result)) {
	  // Something went wrong

	  // We have already reported this error,
	  // so no need to do it again.
	  // DEBUG_X_FAILED("EMdFDB::GetObjectsHavingMonadsInMonadSet", "");

	  // Clear table 
	  // (because we must return an empty table if an error occurred).
	  result.clear();

	  // Return error
	  return false;
	} else {
	  // Everything went fine

	  // Now we must prepare for another iteration

	  // Clear temp_som
	  temp_som.clear();

	  // Set mse_count to 0
	  mse_count = 0;
	}
      }
    }
    // If we got here, there were no errors.
    return true;
  }
}

bool EMdFDB::GetObjectsInUniverse(const std::string& object_type_name,
				  monad_m first, monad_m last,
				  /* out */ IDDList& Result)
{
  if (pConn == 0)
    return false;
  else {

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT object_id_d\n"
      << "FROM " << OTN << "_objects\n"
      << "WHERE " << first << " <= first_monad AND last_monad <= " << last;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetObjectsInUniverse", query_stream.str().c_str());
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	id_d_t oid;
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetObjectsInUniverse", oid))
	  return false;
	else
	  Result.AddOid(oid);
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetObjectsInUniverse");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }

    // Release results
    pConn->ReleaseResults();
    
    // If we got this far, the command was as success.
    return true;
  }
}


#define MAX_CONDITIONS (2*200)

bool EMdFDB::GetInst(const std::string& object_type_name,
		     const CSet_of_monad_ms& Su,
		     const CSet_of_monad_ms& all_m_1,
		     const std::string& pre_query_string,
		     const std::list<feature_info>& features_to_get,
		     /* out */ CInst& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Get Inst
    if (!GetInst(object_type_name,
		 Su.first(), Su.last(),
		 all_m_1,
		 pre_query_string,
		 features_to_get,
		 Result)) {
      return false;
    }
    
    // If we got this far, the command was as success.
    return true;
  }
}


bool EMdFDB::GetInst(const std::string& object_type_name,
		     monad_m first_monad, monad_m last_monad,
		     const CSet_of_monad_ms& all_m_1,
		     const std::string& pre_query_string,
		     const std::list<feature_info>& features_to_get,
		     /* out */ CInst& Result)
{
  if (pConn == 0)
    return false;
  else {
    // ****************************************************
    //
    // Preliminaries
    //
    // ****************************************************
    // Make string of features to get
    std::string str_features_to_get;
    std::vector<id_d_t> feature_types_vec;
    std::list<feature_info>::const_iterator ci(features_to_get.begin());
    std::list<feature_info>::const_iterator cend(features_to_get.end());
    while (ci != cend) {
      // Get lowercase name
      std::string locase_name;
      str_tolower(ci->get_name(), locase_name);

      // Is it self?
      // TODO: We should check for computed status when we add
      // ability to do computed features.
      if (locase_name == "self") {
	str_features_to_get += ", OS.object_id_d";
      } else {
	str_features_to_get += ", OS." + encode_feature_name(locase_name);
      }

      // Decode feature type
      long feature_length;
      id_d_t feature_type;
      decode_feature_type_id(ci->get_type(), feature_type, feature_length);

      // Add type to feature_types_vec
      feature_types_vec.push_back(feature_type);

      // Advance iterators
      ++ci;
    }

    // Straight join
    std::string straight_join;
#if USE_MYSQL
    straight_join = "STRAIGHT_JOIN ";
#else
    straight_join = "";
#endif

    // Make monad_constraint1
    std::ostringstream monad_constraint_stream1;
    monad_constraint_stream1  << "OS.first_monad >= " << first_monad
			      << " AND OS.last_monad <= " << last_monad;
    std::string str_monad_constraint1;
    str_monad_constraint1 = monad_constraint_stream1.str();

    // Get min_m and max_m
    monad_m min_m = all_m_1.first();
    monad_m max_m = all_m_1.last();

    // Should we make use of monad constraints?
    bool bUseMonadConstraints;
    if (first_monad == min_m
	&& last_monad == max_m) {
      bUseMonadConstraints = false;
    } else {
      bUseMonadConstraints = true;
    }

    // Should we make use of pre_query_string?
    bool bUsePreQueryString;
    if (pre_query_string.length() > 0) {
      bUsePreQueryString = true;
    } else {
      bUsePreQueryString = false;
    }
    
    // Make str_pre_query_string
    std::string str_pre_query_string;
    if (bUsePreQueryString) {
      if (bUseMonadConstraints) {
	str_pre_query_string = " (" + pre_query_string + ") AND ";
      } else {
	str_pre_query_string = " (" + pre_query_string + ")";
      }
    } 

    // Should we make use of str_features_to_get
    bool bUseFeaturesToGet;
    if (str_features_to_get.length() > 0) {
      bUseFeaturesToGet = true;
    } else {
      bUseFeaturesToGet = false;
    }
    
    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // ****************************************************
    //
    // Build and execute first query
    //
    // ****************************************************
    //LOG_WRITE_TIME("EMdFDB::GetInst", "Query starting.");
    std::ostringstream query_stream;
    if (bUsePreQueryString) {
      query_stream
	<< "SELECT " 
	<< straight_join
	<< "OS.object_id_d, MMS.mse_first, MMS.mse_last" << str_features_to_get << "\n"
	<< "FROM " << OTN << "_objects OS, " << OTN << "_monad_ms MMS\n"
	<< "WHERE OS.object_id_d = MMS.object_id_d\n"
	<< "      AND " << str_pre_query_string;
      if (bUseMonadConstraints) {
	query_stream << '\n' << str_monad_constraint1;
      }
      // It is slightly kamikaze not to do the following.
      // However, doing it slows us down quite a bit, and
      // in practice, it doesn't seem to be a problem.
      // query_stream << "\nORDER BY OS.object_id_d";
    } else { // pre_query_string is empty
      if (bUseMonadConstraints || bUseFeaturesToGet) {
	query_stream
	  << "SELECT "
	  << straight_join 
	  << "OS.object_id_d, MMS.mse_first, MMS.mse_last" << str_features_to_get << "\n"
	  << "FROM " << OTN << "_objects OS, " << OTN << "_monad_ms MMS\n"
	  << "WHERE OS.object_id_d = MMS.object_id_d";
	if (bUseMonadConstraints) {
	  query_stream << "\n       AND " << str_monad_constraint1;
	}
	// It is slightly kamikaze not to do the following.
	// However, doing it slows us down quite a bit, and
	// in practice, it doesn't seem to be a problem.
	// query_stream << "\nORDER BY OS.object_id_d";
      } else {
	query_stream
	  << "SELECT MMS.object_id_d, MMS.mse_first, MMS.mse_last\n"
	  << "FROM " << OTN << "_monad_ms MMS\n";
	// It is slightly kamikaze not to do the following
	// However, doing it slows us down quite a bit, and
	// in practice, it doesn't seem to be a problem.
	// query_stream << "\nORDER BY MMS.object_id_d";
      }
    }
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetInst", query_stream.str().c_str());
      return false;
    }
 
    //LOG_WRITE_TIME("EMdFDB::GetInst", "Query finished.");
    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    int index;
    int no_of_features_to_get = features_to_get.size();

    LOG_WRITE_TIME("EMdFDB::GetInst", "Retrieval starting.");
    id_d_t current_id_d = NIL;
    CInstObject *pObj = 0;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {	
	// Declare this pFeature_values array of dummy CEMdFValue objects.
	CEMdFValue *pFeature_values;
	if (no_of_features_to_get > 0) {
	  // This uses the () constructor, which gives us a kEVInt with 
	  // value 0.
	  pFeature_values = new CEMdFValue[no_of_features_to_get];
	} else {
	  pFeature_values = 0;
	}

	// Get object id_d
	id_d_t oid;
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetInst", oid))
	  return false;

	// Get first_monad
	monad_m first_monad;
	if (!pConn->AccessTuple(tuple, 1, "EMdFDB::GetInst", first_monad))
	  return false;

	// Get last_monad
	monad_m last_monad;
	if (!pConn->AccessTuple(tuple, 2, "EMdFDB::GetInst", last_monad))
	  return false;

	// Get object if we don't have it already
	if (current_id_d != oid) {
	  // Add previous object if it has been created (i.e., if this isn't
	  // the first object in the results.)
	  if (pObj != 0) {
	    // Insert object in inst
	    Result.add_object(pObj);
	  }

	  // Make sure that next time, we don't retrieve it again
	  // from Result.
	  current_id_d = oid;

	  // Get feature values
	  if (no_of_features_to_get > 0) {
	    std::string value_str;
	    long value_long;
	    for (index = 0; index < no_of_features_to_get; ++index) {
	      CEMdFValue *pValue = 0;
	      switch (feature_types_vec[index]) {
	      case FEATURE_TYPE_ASCII:
		if (!pConn->AccessTuple(tuple, 3 + index, "EMdFDB::GetInst", value_str))
		  return false;
		pValue = new CEMdFValue(value_str);
		break;
	      case FEATURE_TYPE_STRING:
		if (!pConn->AccessTuple(tuple, 3 + index, "EMdFDB::GetInst", value_str))
		  return false;
		pValue = new CEMdFValue(dbstring2string(value_str));
		break;
	      case FEATURE_TYPE_INTEGER:
		if (!pConn->AccessTuple(tuple, 3 + index, "EMdFDB::GetInst", value_long))
		  return false;
		pValue = new CEMdFValue(kEVInt, value_long);
		break;
	      case FEATURE_TYPE_ID_D:
		if (!pConn->AccessTuple(tuple, 3 + index, "EMdFDB::GetInst", value_long))
		  return false;
		pValue = new CEMdFValue(kEVID_D, value_long);
		break;
	      default: // Enumerations: same as ID_D
		if (!pConn->AccessTuple(tuple, 3 + index, "EMdFDB::GetInst", value_long))
		  return false;
		pValue = new CEMdFValue(kEVEnum, value_long);
		break;
	      };

	      // Just checking...
	      assert(pValue != 0);

	      // Set pFeature_values vector
	      pFeature_values[index] = *pValue;

	      // Discard value
	      delete pValue;
	    }
	  }

	  // Make new object
	  pObj = new CInstObject(oid, first_monad, last_monad, pFeature_values);

	} else { 
	  // if we get here, then current_id_d == oid

	  // Add mse to object
	  pObj->add_mse(first_monad, last_monad);
	}
      } // end of iteration over query-results
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetInst");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }

    // Add last object if it has been created (i.e., 
    // if the results were non-empty).
    if (pObj != 0) {
      // Insert object in inst
      Result.add_object(pObj);
    }

    // Write "retrieval finished" to log
    LOG_WRITE_TIME("EMdFDB::GetInst", "Retrieval finished.");

    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command was as success.
    return true;
  }
}



bool EMdFDB::GetObjectsStartingAtSm(const std::string& object_type_name,
				    monad_m Sm,
				    /* out */ IDDList& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT object_id_d\n"
      << "FROM " << OTN << "_monad_ms\n"
      << "WHERE mse_first = " << Sm << " AND is_first = 'Y'";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetObjectsStartingAtSm", query_stream.str().c_str());
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	id_d_t oid;
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetObjectsStartingAtSm", oid))
	  return false;
	else
	  Result.AddOid(oid);
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetObjectsStartingAtSm");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }
    
    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::GetObjectsAtM(const std::string& object_type_name,
			   monad_m at_monad,
			   /* out */ IDDList& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT object_id_d\n"
      << "FROM " << OTN << "_objects\n"
      << "WHERE first_monad <= " << at_monad << " AND last_monad >= " << at_monad << "\n";
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetObjectsAtM", query_stream.str().c_str());
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	id_d_t oid;
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetObjectsAtM", oid))
	  return false;
	else
	  Result.AddOid(oid);
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetObjectsAtM");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }

    
    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command was as success.
    return true;
  }
}

//
// Gets the set of monads making up the object 
// with id_d object_id_d
//
bool EMdFDB::GetMonadsFromID_D(id_d_t object_id_d, 
			       id_d_t object_type_id,
			       /* out */ CSet_of_monad_ms& monads)
{
  // Get object type name
  std::string object_type_name;
  bool bObjectTypeExists;
  if (!ObjectTypeExists(object_type_id, bObjectTypeExists, object_type_name)) {
    DEBUG_X_FAILED("GetMonadsFromID_D", "Could not get object type name for object id_d " + id_d2string(object_id_d) + ".\n");
    return false;
  }
  if (!bObjectTypeExists) {
    DEBUG_X_FAILED("GetMonadsFromID_D", "Object type for object with object id_d " + id_d2string(object_id_d) + " does not exist.\n");
    return true;
  }

  // Get set of monads
  if (!GetSOMForObject(object_type_name, object_id_d, monads)) {
    DEBUG_X_FAILED("GetMonadsFromID_D", "Could not GetSOMForObject for object with object id_d " + id_d2string(object_id_d) + ".\n");
    return false;
  }

  // If we got this far, there were no errors
  return true;
}

bool EMdFDB::GetMonadsFromID_Ds(const std::list<id_d_t>& id_ds, 
				const std::string &object_type_name,
				CTable *pOutTable)
{
  if (pConn == 0)
    return false;
  else {
    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Build first part of query
    std::ostringstream query_stream_select_from;
    query_stream_select_from
      << "SELECT object_id_d, mse_first, mse_last\n"
      << "FROM " << OTN << "_monad_ms\n";

    std::list<id_d_t>::const_iterator ci(id_ds.begin());
    std::list<id_d_t>::const_iterator cend(id_ds.end());
    while (ci != cend) {
#define WHERE_MAX_ID_DS (16)
      // Build where-part
      std::ostringstream query_stream_where;
      query_stream_where << "WHERE";
      for (int i = 0; (i < WHERE_MAX_ID_DS) && (ci != cend); ++i) {
	query_stream_where << " object_id_d = " << *ci;
	++ci;
	if (!(ci == cend || i == (WHERE_MAX_ID_DS-1))) {
	  query_stream_where << " OR\n";
	}
      }
#undef WHERE_MAX_ID_DS

      // Build query-string
      std::string query(query_stream_select_from.str());
      query += query_stream_where.str();

      // Execute query
      if (!pConn->Exec(query)) {
	// Release results
	pConn->ReleaseResults();

	DEBUG_QUERY_FAILED("EMdFDB::GetMonadsFromID_Ds", query.c_str());
	return false;
      }

      // Build table
      int NoOfTuples = pConn->NoOfResultTuples();
      int tuple;
      id_d_t oid;
      monad_m mse_first, mse_last;
      try {
	for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	  if (pConn->AccessTuple(tuple, 0, "EMdFDB::GetMonadsFromID_Ds", oid)
	      && pConn->AccessTuple(tuple, 1, "EMdFDB::GetMonadsFromID_Ds", mse_first)
	      && pConn->AccessTuple(tuple, 2, "EMdFDB::GetMonadsFromID_Ds", mse_last)) {
	    pOutTable->start_new_row();
	    pOutTable->append(id_d2string(oid));
	    pOutTable->append(monad_m2string(mse_first));
	    pOutTable->append(monad_m2string(mse_last));
	  } else {
	    // Release results
	    pConn->ReleaseResults();

	    // Return failure
	    return false;
	  }
	}
      } catch (EMdFNULLValueException e) {
	DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetMonadsFromID_Ds");

	// Release results
	pConn->ReleaseResults();

	// There was a NULL value exception, hence we return an error
	return false;
      }

    }
    
    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command was as success.
    return true;
  }

}


//
// Get set of monads from object_id_d belonging to object type
// with name object_type_name
//
bool EMdFDB::GetSOMForObject(const std::string& object_type_name,
			     id_d_t object_id_d,
			     /* out */ CSet_of_monad_ms& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT mse_first, mse_last\n"
      << "FROM " << OTN << "_monad_ms\n"
      << "WHERE object_id_d = " << object_id_d;
    if (!pConn->Exec(query_stream.str())) {
      // Possibly emit debug message
      DEBUG_QUERY_FAILED("EMdFDB::GetSOMForObject", query_stream.str().c_str());
      // Release results
      pConn->ReleaseResults();

      // Return failure
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    monad_m mse_first, mse_last;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	if (pConn->AccessTuple(tuple, 0, "EMdFDB::GetSOMForObject", mse_first)
	    && pConn->AccessTuple(tuple, 1, "EMdFDB::GetSOMForObject", mse_last)) {
	  Result.add(mse_first, mse_last);
	} else {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failure
	  return false;
	}
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetSOMForObject");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }
     
    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::GetObjectTypes(/* out */ std::list<std::string>& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Declare szQuery
    char szQuery[512];

    // Build and execute query
    sprintf(szQuery, 
	    "SELECT object_type_name\n"
	    "FROM object_types"
	    );
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::GetObjectTypes", szQuery);
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	std::string str;
	if (pConn->AccessTuple(tuple, 0, "EMdFDB::GetObjectTypes", str))
	  Result.push_back(str);
	else {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failure
	  return false;
	}
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetObjectTypes");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }
    
    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command was as success.
    return true;
  }
}

//
// This function is part of an overall invariant which says
// that if there is one feature for a given object type id 
// in the feature cache, then they are all there.
// This must be maintained by calling this function before
// adding a feature to the feature cache.
bool EMdFDB::GetFeaturesForObjectType(id_d_t object_type_id,
				      /* out */ std::list<feature_info>& Result)
{
  if (pConn == 0)
    return false;
  else {
    //
    // First try the cache
    //

    // Make object_type_id to string for later comparison
    std::string str_oid = id_d2string(object_type_id);

    // Assume we didn't find any.
    bool bReturnSuccess = false;

    // Go through feature cache
    TableIterator i = m_feature_cache->iterator();
    while (i.hasNext()) {
      // Does this feature in the feature cache
      // have the same object type id as the one
      // we are looking for?
      if (str_oid == m_feature_cache->get_column(i, 1)) {
	// Get info
	std::string feature_name = m_feature_cache->get_column(i, 2);
	id_d_t feature_type_id = string2id_d(m_feature_cache->get_column(i, 3));
	std::string default_value = m_feature_cache->get_column(i, 4);
	bool is_computed = string2bool(m_feature_cache->get_column(i, 5));
	
	// Make feature_info
	feature_info f_info(feature_name, feature_type_id, 
			    default_value, is_computed);

	// Include it in result
	Result.push_back(f_info);

	// At least we found one that belonged to the object type.
	// Se let's assume we have found them all.
	bReturnSuccess = true;
      }

      // Advance iterator
      i.next();
    }

    // If we found at least one, return success.
    if (bReturnSuccess) 
      return true;

    //
    // Otherwise, look in database
    //

    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT feature_name, feature_type_id, default_value, computed\n"
      << "FROM features\n"
      << "WHERE object_type_id = " << object_type_id;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetFeaturesForObjectType", query_stream.str().c_str());
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	std::string name, default_value;
	id_d_t feature_type;
	bool is_computed;

	/* Get name */
	std::string temp_name;
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetFeaturesForObjectType", temp_name)) {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failure
	  return false;
	}
	name = decode_feature_name(temp_name);

	/* Get type */
	if (!pConn->AccessTuple(tuple, 1, "EMdFDB::GetFeaturesForObjectType", feature_type)) {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failure
	  return false;
	}

	/* Get default value */
	std::string defval;
	if (!pConn->AccessTuple(tuple, 2, "EMdFDB::GetFeaturesForObjectType", defval)) {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failure
	  return false;
	}

	// Convert default value
	default_value = dbstring2string(defval);

	/* Get is_computed */
	if (!pConn->AccessTuple(tuple, 3, "EMdFDB::GetFeaturesForObjectType", is_computed)) {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failure
	  return false;
	}

	/* Construct feature_info and push on the back. */
	feature_info f_info(name, feature_type, default_value, is_computed);
	Result.push_back(f_info);

	// Add it to the feature cache if not already there.
	add_feature_to_cache_if_not_already_there(object_type_id, 
						  name, 
						  feature_type, 
						  default_value, 
						  is_computed);
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetFeaturesForObjectType");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }
    
    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::GetEnumerations(/* out */ std::list<std::string>& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Declare szQuery
    char szQuery[512];

    // Build and execute query
    sprintf(szQuery, 
	    "SELECT enum_name\n"
	    "FROM enumerations"
	    );
    if (!pConn->Exec(szQuery)) {
      // Possibly emit debug message
      DEBUG_QUERY_FAILED("EMdFDB::GetEnumerations", szQuery);

      // Release results
      pConn->ReleaseResults();
      
      // Return failure
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	std::string enum_name;

	/* Get enumeration name. */
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetEnumerations", enum_name)) {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failure
	  return false;
	}

	/* Put it at the back of the list. */
	Result.push_back(enum_name);
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetEnumerations");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }
    
    // Release results
    pConn->ReleaseResults();
    
    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::GetEnumConstants(id_d_t enum_id,
			      /* out */ std::list<enum_const_info>& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT enum_value_name, value, is_default\n"
      << "FROM enumeration_constants\n"
      << "WHERE enum_id = " << enum_id;
    if (!pConn->Exec(query_stream.str().c_str())) {
      // Possibly emit debug message
      DEBUG_QUERY_FAILED("EMdFDB::GetEnumConstants", query_stream.str().c_str());
      // Release results
      pConn->ReleaseResults();
      
      // Return failue
      return false;
    }
    
    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	std::string name;
	long value;
	bool is_default;
      
	/* Get name. */
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetEnumConstants", name)) {
	  // Release results
	  pConn->ReleaseResults();
	  
	  // Return failue
	  return false;
	}

	/* Get name. */
	if (!pConn->AccessTuple(tuple, 1, "EMdFDB::GetEnumConstants", value)) {
	  // Release results
	  pConn->ReleaseResults();
	  
	  // Return failue
	  return false;
	}

	/* Get name. */
	if (!pConn->AccessTuple(tuple, 2, "EMdFDB::GetEnumConstants", is_default)) {
	  // Release results
	  pConn->ReleaseResults();
	  
	  // Return failue
	  return false;
	}

	/* Put it at the back of the list. */
	enum_const_info ec_info(name, value, is_default);
	Result.push_back(ec_info);
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetEnumConstants");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }
    
    // Release results
    pConn->ReleaseResults();
	  
    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::GetObjectTypesUsingEnumeration(id_d_t enum_id,
					    /* out */ std::list<std::string>& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Declare query_stream
    std::ostringstream query_stream;

    // Build and execute query
    query_stream
      << "SELECT OT.object_type_name\n"
      << "FROM object_types OT, features F\n"
      << "WHERE OT.object_type_id = F.object_type_id \n"
      << "      AND F.feature_type_id = " << enum_id;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetObjectTypesUsingEnumeration", query_stream.str().c_str());
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	std::string name;

	/* Get name. */
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetObjectTypesUsingEnumeration", name)) {
	  // Release results
	  pConn->ReleaseResults();
	  
	  // Return failue
	  return false;
	}

	/* Push it at the back. */
	Result.push_back(name);
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetObjectTypesUsingEnumeration");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }
    
    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::GetSegments(/* out */ std::list<segment_info>& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Declare szQuery
    char szQuery[512];

    // Build and execute query
    sprintf(szQuery, 
	    "SELECT segment_name, first_monad, last_monad\n"
	    "FROM segments\n"
	    "ORDER BY first_monad"
	    );
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::GetSegments", szQuery);
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	std::string name;
	monad_m first, last;

	/* Get name. */
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetSegments", name)) {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failue
	  return false;
	}

	/* Get name. */
	if (!pConn->AccessTuple(tuple, 1, "EMdFDB::GetSegments", first)) {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failue
	  return false;
	}

	/* Get name. */
	if (!pConn->AccessTuple(tuple, 2, "EMdFDB::GetSegments", last)) {
	  // Release results
	  pConn->ReleaseResults();

	  // Return failue
	  return false;
	}

	/* Push it at the back. */
	segment_info seg_info(name, first, last);
	Result.push_back(seg_info);
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetSegments");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }

    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::GetObjectsWithinMonads(const CSet_of_monad_ms& monads, 
				    const std::string& object_type_name,
				    /* out */ std::list<id_d_t>& result)
{
  if (pConn == 0)
    return false;
  else {
    // Make sure we are pristine
    result.clear();
    
    // Return if monads are empty.
    if (monads.is_empty())
      return true;

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Make monad constraints
    std::string monad_constraints = MakeMonadConstraintsFromSOM(monads);

    // Make query for selecting object ids 
    std::string query =
      std::string(" SELECT object_id_d\n") 
      + "FROM " + OTN + "_objects\n"
      + "WHERE (" + monad_constraints + ")";
    if (!pConn->Exec(query.c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::GetObjectsWithinMonads", query);
      return false;
    }

    int NoOfTuples = pConn->NoOfResultTuples();
    int tuple;
    try {
      for (tuple = 0; tuple < NoOfTuples; ++tuple) {
	id_d_t oid;
	if (!pConn->AccessTuple(tuple, 0, "EMdFDB::GetObjectsWithinMonads", oid)) {
	  // Make sure we don't return any results
	  result.clear();

	  // Release results
	  pConn->ReleaseResults();

	  // Return failure
	  return false;
	}
	result.push_back(oid);
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::GetObjectsWithinMonads");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }

    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::ObjectID_DExistsInType(id_d_t object_id_d, 
				    const std::string& object_type_name,
				    /* out */ bool& Result)
{
  if (pConn == 0)
    return false;
  else {
    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Build and execute query
    std::ostringstream query_stream;
    query_stream
      << "SELECT object_id_d\n" 
      << "FROM " << OTN << "_objects \n" //     
      << "WHERE object_id_d = " << object_id_d;
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::ObjectID_DExistsInType", query_stream.str().c_str());
      return false;
    }
    
    /* If there was a non-zero number of tuples, the object
     * exists. */
    Result = pConn->NoOfResultTuples() != 0;

    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::ObjectID_DExists(id_d_t object_id_d, 
			      /* out */ bool& Result,
			      /* out */ id_d_t& object_type_id)
{
  if (pConn == 0)
    return false;
  else {
    // Assume we didn't find it
    Result = false;

    // Get object type names
    std::list<std::string> object_type_names;
    GetObjectTypes(object_type_names);

    // Iterate through object type names
    std::list<std::string>::const_iterator ci(object_type_names.begin());
    std::list<std::string>::const_iterator cend(object_type_names.end());
    while (!Result && ci != cend) {
      // See if it exists in this type
      bool bObjectExistsInType;
      if (!ObjectID_DExistsInType(object_id_d, ci->c_str(),
				  bObjectExistsInType))
	return false;

      // We found it
      if (bObjectExistsInType) {
	// Tell caller we found it
	Result = true;

	// Find object type id
	bool bObjectTypeExists;
	if (!ObjectTypeExists(ci->c_str(), 
			      bObjectTypeExists, object_type_id))
	  return false;
      }

      // Advance iterator
      ++ci;
    }
    // If we got this far, the command was as success.
    return true;
  }
}

bool EMdFDB::CreateObject(id_d_t object_id_d,
			  const std::string& object_type_name,
			  const CSet_of_monad_ms& monads,
			  const std::list<feature_info>& features)
{
  if (pConn == 0)
    return false;
  else {
    // Get object type id 
    id_d_t object_type_id;
    bool TypeExists;
    if (!ObjectTypeExists(object_type_name, TypeExists, object_type_id)
	|| !TypeExists) {
      DEBUG_X_FAILED("EMdFDB::CreateObject", "Object type did not exist.");
      return false;
    }

    // Begin transaction
    bool bDoCommit;
    bDoCommit = pConn->BeginTransaction();

    // Generate a new monad_ms id
    id_d_t monad_ms_id;
    if (!GetNextOtherID(monad_ms_id)) {
      DEBUG_X_FAILED("EMdFDB::CreateObject", "Could not create monad_ms id.");
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Insert in 'OT_monad_ms'
    SOMConstIterator msei = monads.const_iterator();
    bool bIsFirst = true;
    while (msei.hasNext()) {
      std::ostringstream query_stream;
      Monad_Set_Element mse = msei.current();
      query_stream
	<< "INSERT INTO " << OTN << "_monad_ms "
	<< "(monad_ms_id, object_id_d, mse_first, mse_last, is_first)\n"
	<< "VALUES ( " << monad_ms_id
	<< ", " << object_id_d 
	<< ", " << mse.first() 
	<< ", " << mse.last() 
	<< ", '" << bool2char(bIsFirst) << "'"
	<< " )";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::CreateObject", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }

      // Next time around, we are not first
      bIsFirst = false;
      
      // Advance iterator
      msei.next();
    }

    //
    // Insert in 'OT_objects

    // Build query
    std::ostringstream strQuery;
    strQuery << "INSERT INTO "
	     << OTN << "_objects "
	     << "(object_id_d, monad_ms_id, first_monad, last_monad";
    
    // Add names of features to be added
    std::list<feature_info>::const_iterator fi(features.begin());
    std::list<feature_info>::const_iterator fiend(features.end());
    while (fi != fiend) {
      if (!fi->get_is_computed()) {
	strQuery << ", " << encode_feature_name(fi->get_name());
      }
      ++fi;
    }
    strQuery << ")\n";

    // Add values
    strQuery << "VALUES ( ";
    strQuery << object_id_d << ", ";
    strQuery << monad_ms_id << ", ";
    strQuery << monads.first() << ", ";
    strQuery << monads.last();
    fi = features.begin();
    while (fi != features.end()) {
      if (!fi->get_is_computed()) {
	strQuery << ", " << feature_info2SQLvalue(*fi);
      }
      ++fi;
    }
    strQuery << ")\n";

    // Execute query
    if (!pConn->Exec(strQuery.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateObject", strQuery.str().c_str());
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }
 
    // Update min_m to monads.first() if necessary
    monad_m min_m_candidate = monads.first();
    if (!Set_min_m(min_m_candidate)) {
      DEBUG_X_FAILED("EMdFDB::CreateObject", "Setting min_m");
      return false;
    }

    // Update max_m to monads.last() if necessary
    monad_m max_m_candidate = monads.last();
    if (!Set_max_m(max_m_candidate)) {
      DEBUG_X_FAILED("EMdFDB::CreateObject", "Setting max_m");
      return false;
    }

    // Commit transaction if necessary
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::CreateObject");
	return false;
      }
    }

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::UpdateObject(id_d_t object_id_d,
			  const std::string& object_type_name,
			  const std::list<feature_info>& features)
{
  if (pConn == 0)
    return false;
  else {
    //
    // Create query
    //
    std::ostringstream query_stream;

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Add UPDATE
    query_stream 
      << "UPDATE " << OTN << "_objects\n";

    // Add features
    query_stream << "SET ";
    std::list<feature_info>::const_iterator ci(features.begin());
    std::list<feature_info>::const_iterator cend(features.end());
    while (ci != cend) {
      // Add feature
      query_stream << encode_feature_name(ci->get_name()) << " = " << feature_info2SQLvalue(*ci);

      // Advance iterator
      ++ci;

      // Add comma or newline
      if (ci != cend)
	query_stream << ",\n";
      else 
	query_stream << "\n";
    }

    // Add WHERE
    query_stream << "WHERE object_id_d = " << object_id_d;

    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::UpdateObject", query_stream.str().c_str());
      return false;
    }

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::DropObject(id_d_t object_id_d, id_d_t object_type_id, const std::string& object_type_name)
{
  if (pConn == 0)
    return false;
  else {
    // Begin transaction
    bool bDoCommit;
    bDoCommit = pConn->BeginTransaction();

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Get monads (we need them later)
    CSet_of_monad_ms monads;
    GetMonadsFromID_D(object_id_d, object_type_id, monads);

    // Delete object from OT_monad_ms
    {
      std::ostringstream query_stream;
      query_stream 
	<< "DELETE\n"
	<< "FROM " << OTN << "_monad_ms\n"
	<< "WHERE object_id_d = " << object_id_d;
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropObject", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Delete object from OT_objects
    {
      std::ostringstream query_stream;
      query_stream 
	<< "DELETE\n"
	<< "FROM " << OTN << "_objects\n"
	<< "WHERE object_id_d = " << object_id_d;
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropObject", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    //
    // Check to see whether we should update min_m/max_m
    //

    // Get current min_m and max_m
    monad_m current_min_m;
    if (!Get_min_m(current_min_m)) {
      DEBUG_X_FAILED("EMdFDB::DropObject", "Getting current min_m");
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }
 
    monad_m current_max_m;
    if (!Get_max_m(current_max_m)) {
      DEBUG_X_FAILED("EMdFDB::DropObject", "Getting current max_m");
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Should we update min_m and/or max_m?
    // That is, is there a danger that this object was the one
    // causing the min_m?
    // That is, is current_min_m == the last monad of the object?
    if (current_min_m == monads.first()
	|| current_max_m == monads.last()) {
      // Yes, we should update
      if (!Set_min_max_m_from_object_types(false)) {
	DEBUG_X_FAILED("EMdFDB::DropObject", "Setting min_m/max_m from all object types");
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    // Commit transaction if necessary
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::DropObject");
	return false;
      }
    }

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::DropObjectsInMonads(const CSet_of_monad_ms& monads, const std::string& object_type_name, id_d_t object_type_id, /* out */ std::list<id_d_t>& result)
{
  if (pConn == 0)
    return false;
  else {
    // See if monads are empty.  Return with success if they are
    if (monads.is_empty()) {
      return true;
    }

    // Get results
    if (!GetObjectsWithinMonads(monads, object_type_name, result)) {
      result.clear();
      DEBUG_X_FAILED("EMdFDB::DropObjectsInMonads", "GetObjectsWithinMonads");
      return false;
    }

    // If there are no objects to delete, return at this point with success
    if (result.empty()) {
      return true;
    }

    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Make monad constraints
    std::string monad_constraints_first_last_monad = MakeMonadConstraintsFromSOM(monads);


    // Begin transaction
    bool bDoCommit;
    bDoCommit = pConn->BeginTransaction();

    std::cerr << "Doing " << OTN << ": " << std::flush;
    std::cerr << "OT_monad_ms, " << std::flush;

    // Do the monad_ms 200 at a time
    // This is because we can't just do it with the mse_first/mse_last 
    // pair: We might get half an object's monads deleted
    // if the object happens to have mses inside and outside
    // a monad range to be deleted.
    std::list<id_d_t>::const_iterator ci = result.begin();
    std::list<id_d_t>::const_iterator cend = result.end();
    int count = 0;
    std::string object_id_ds = "";
    for (; ci != cend;) {
      bool bDoIt = false;
      object_id_ds += id_d2string(*ci);
      ++ci;
      ++count;
      if ((count % 200) == 199
	  || ci == cend) {
	bDoIt = true;
      }

      if (bDoIt) {
	// Delete object from OT_monad_ms
	std::ostringstream query_stream;
	query_stream 
	  << "DELETE\n"
	  << "FROM " << OTN << "_monad_ms\n"
	  << "WHERE object_id_d IN (" 
	  << object_id_ds
	  << ")";
	if (!pConn->Exec(query_stream.str().c_str())) {
	  DEBUG_QUERY_FAILED("EMdFDB::DropObjectsInMonads", query_stream.str().c_str());
	  if (bDoCommit)
	    pConn->AbortTransaction();
	  result.clear();
	  return false;
	}
	object_id_ds = "";
	count = 0;
      } else { // !bDoIt
	object_id_ds += ",";
      }
    }
      
    std::cerr << "OT_objects..." << std::flush;
    // Delete object from OT_objects
    {
      std::ostringstream query_stream;
      query_stream 
	<< "DELETE\n"
	<< "FROM " << OTN << "_objects\n"
	<< "WHERE " << monad_constraints_first_last_monad;
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::DropObjectsInMonads", query_stream.str().c_str());
	if (bDoCommit)
	  pConn->AbortTransaction();
	result.clear();
	return false;
      }
    }

    // Update min_m/max_m if necessary
    if (!Set_min_max_m_from_object_types(true)) {
      DEBUG_X_FAILED("EMdFDB::DropObjectsInMonads", "Setting min_m/max_m from all object types");
      if (bDoCommit)
	pConn->AbortTransaction();
      result.clear();
      return false;
    }
    

    // Commit transaction if necessary
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::DropObjectsInMonads");
	result.clear();
	return false;
      }
    }
    std::cerr << "!" << std::endl << std::flush;

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::GetCurrentDatabaseVersion(const std::string& database_name, 
				       std::string& /* out */ strCurrentVersion,
				       int& /* out */ nCurrentVersion,
				       bool& /* out */ bNeedsUpgrading,
				       bool& /* out */ bCanUpgrade)
{
  if (pConn == 0)
    return false;
  else {
    // Assume we can upgrade
    bCanUpgrade = true;

    if (!UseDatabase(database_name)) {
      std::cerr << "FAILURE: Could not USE DATABASE " << database_name << "!" << std::endl;
      bCanUpgrade = false;
      return false;
    }

    // Assume we need upgrading
    bNeedsUpgrading = true;

    // Check if sequence_0 table exists
    bool bSequence_0TableExists;
    if (!TableExists("sequence_0", bSequence_0TableExists)) {
      bCanUpgrade = false;
      return false;
    }

    bool bSequence_118_0TableExists;
    if (!TableExists("sequence_118_0", bSequence_118_0TableExists)) {
      bCanUpgrade = false;
      return false;
    }

    // Did sequence_118_0 exist?
    if (bSequence_118_0TableExists) {
      strCurrentVersion = "1.1.8 or later";
      nCurrentVersion = EMDROS_VERSION_1_1_8;

      // If we are here, we needn't upgrade
      bNeedsUpgrading = false;
      bCanUpgrade = false;

      return true;
    }

    // If not, and if sequence_0 didn't exist either, 
    // then we are pre-1.1.0
    if (!bSequence_0TableExists) {
      strCurrentVersion = "pre-1.1.0";
      nCurrentVersion = EMDROS_VERSION_PRE_1_1_0;

      return true;
    }

    bool bTableExists;
    // Check to see whether max_m exists
    if (!TableExists("max_m", bTableExists)) {
      return false;
    }

    // If not, then we are pre-1.1.5
    if (!bTableExists) {
      strCurrentVersion = "1.1.0 to 1.1.4";
      nCurrentVersion = EMDROS_VERSION_1_1_0_TO_1_1_4;

      return true;
    } else {
      // Current version is 1.1.5 to 1.1.7
      strCurrentVersion = "1.1.5 to 1.1.7";
      nCurrentVersion = EMDROS_VERSION_1_1_5_TO_1_1_7;

      return true;
    }

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::UpgradeDatabase(const std::string& database_name, 
			     int nCurrentVersion)
{
  if (pConn == 0)
    return false;
  else {
    if (!UseDatabase(database_name)) {
      std::cerr << "FAILURE: Could not USE DATABASE " << database_name << "!" << std::endl;
      return false;
    }

    // If version is pre-1.1.0, then upgrade to 1.1.0
    if (nCurrentVersion <= EMDROS_VERSION_PRE_1_1_0) {
      if (!UpgradeDatabaseTo110(database_name))
	return false;
    }

  
    // If version is pre-1.1.5, then upgrade to 1.1.5
    if (nCurrentVersion <= EMDROS_VERSION_1_1_0_TO_1_1_4) {
      UpgradeDatabaseTo115(database_name);
    } 

    // If version is pre-1.1.8, then upgrade to 1.1.8
    if (nCurrentVersion <= EMDROS_VERSION_1_1_5_TO_1_1_7) {
      UpgradeDatabaseTo118(database_name);
    } 


    // If we got this far, there were no errors
    return true;
  }
}

//
// Upgrading to 1.1.0 from pre-1.1.0
bool EMdFDB::UpgradeDatabaseTo110(const std::string& database_name)
{
  if (pConn == 0)
    return false;
  else {
    std::list<std::string> object_type_names;
    if (!GetObjectTypes(object_type_names)) {
      std::cerr << "FAILURE: Could not get list of object types in database " << database_name << "!" << std::endl;
      return false;
    }

    std::list<std::string>::const_iterator ci;
    ci = object_type_names.begin();
    while (ci != object_type_names.end()) {
      if (!UpgradeObjectType(*ci)) {
	return false;
      }
      ++ci;
    }

    std::cerr << "DROPping 'objects' table... " << std::flush;
    // DROP objects
    {
      std::ostringstream query_stream;
      query_stream 
	<< "DROP TABLE objects\n";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::UpgradeDatabase110", query_stream.str().c_str());
	std::cerr << "Failed!" << std::endl;
	return false;
      }
    }
    std::cerr << "Done!" << std::endl;

    // Creaing sequence-tables
    std::cerr << "Creating sequence_X tables... " << std::flush;
    if (!CreateSequenceTables110(false)) {
      std::cerr << "Failed!" << std::endl;
      return false;
    }
    std::cerr << "Done!" << std::endl;

    // Moving sequence-values
    std::cerr << "Moving sequence_X values... " << std::flush;
    if (!MoveSequenceValue110(SEQUENCE_OBJECT_ID_DS)) {
      DEBUG_X_FAILED("EMdFDB::UpgradeDatabase110", "Moving sequence OBJECT_ID_DS");
      std::cerr << "Failed!" << std::endl;
      return false;
    }
    if (!MoveSequenceValue110(SEQUENCE_OTHER_IDS)) {
      DEBUG_X_FAILED("EMdFDB::UpgradeDatabase110", "Moving sequence OTHER_ID");
      std::cerr << "Failed!" << std::endl;
      return false;
    }
    std::cerr << "Done!" << std::endl;

    // Dropping 'sequences' table
    std::cerr << "DROPping 'sequences' table... " << std::flush;
    {
      std::ostringstream query_stream;
      query_stream 
	<< "DROP TABLE sequences\n";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::UpgradeDatabase110", query_stream.str().c_str());
	std::cerr << "Failed!" << std::endl;
	return false;
      }
    }
    std::cerr << "Done!" << std::endl;

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::UpgradeObjectType(const std::string& object_type_name)
{
  if (pConn == 0)
    return false;
  else {
    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);

    // Begin transaction
    bool bDoCommit = pConn->BeginTransaction();

    std::cerr << "Adding column is_first to table " << OTN << "_monad_ms..." << std::flush;
    // ALTER TABLE
    {
      std::ostringstream query_stream;
      query_stream 
	<< "ALTER TABLE " << OTN << "_monad_ms\n"
	<< "ADD COLUMN is_first CHAR(1)\n";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::UpgradeObjectType", query_stream.str().c_str());
	std::cerr << "Failed!" << std::endl;
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }
    // SET DEFAULT
    {
      std::ostringstream query_stream;
      query_stream 
	<< "ALTER TABLE " << OTN << "_monad_ms\n"
	<< "ALTER COLUMN is_first SET DEFAULT 'N'\n";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::UpgradeObjectType", query_stream.str().c_str());
	std::cerr << "Failed!" << std::endl;
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }
    std::cerr << " Done!" << std::endl;

    std::cerr << "Adding indices to " << OTN << "_objects and to " << OTN << "_monad_ms... " << std::flush;
    // Create indices on OT_objects    
    if (!CreateIndicesOnOTObjects(OTN)) {
      std::cerr << "Failed!" << std::endl;
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create indices on OT_monad_ms
    if (!CreateIndicesOnOTMonadMs(OTN)) {
      std::cerr << "Failed!" << std::endl;
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }
    std::cerr << "Done!" << std::endl;

#if USE_POSTGRESQL
    std::cerr << "Setting column is_first in table " << OTN << "_monad_ms to 'N'..." << std::flush;
    // SET all to 'N'
    {
      std::ostringstream query_stream;
      query_stream 
	<< "UPDATE " << OTN << "_monad_ms\n"
	<< "SET is_first = 'N'";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::UpgradeObjectType", query_stream.str().c_str());
	std::cerr << "Failed!" << std::endl;
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }
    std::cerr << " Done!" << std::endl;
#endif


    std::cerr << "Getting which to update in table " << OTN << "_monad_ms..." << std::flush;
    // SELECT
    {
      std::ostringstream query_stream;
      query_stream 
	<< "SELECT monad_m, object_id_d\n"
	<< "FROM " << OTN << "_monad_ds\n"
	<< "WHERE starts_here = 'Y'\n";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::UpgradeObjectType", query_stream.str().c_str());
	std::cerr << "Failed!" << std::endl;
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }

    std::cerr << "!" << std::endl;

    std::list<std::pair<id_d_t, monad_m> > my_list;
    long no_of_tuples;
    try {
      if ((no_of_tuples = pConn->NoOfResultTuples()) > 0) {
	for (long i = 0; i < no_of_tuples; ++i) {
	  if ((i % 500) == 499) {
	    std::cerr << "." << std::flush;
	  }
	  // Access monad_m
	  monad_m monad;
	  if (!pConn->AccessTuple(i, 0, "EMdFDB::UpgradeObjectType", monad)) {
	    // DEBUG information has already been printed if it is on.
	    std::cerr << "Failed!" << std::endl;

	    // Abort transaction if necessary
	    if (bDoCommit)
	      pConn->AbortTransaction();

	    // Release results
	    pConn->ReleaseResults();
	  
	    // Return failure
	    return false;
	  }

	  // Access object_id_d
	  id_d_t object_id_d;
	  if (!pConn->AccessTuple(i, 1, "EMdFDB::UpgradeObjectType", object_id_d)) {
	    // DEBUG information has already been printed if it is on.
	    std::cerr << "Failed!" << std::endl;

	    // Abort transaction if necessary
	    if (bDoCommit)
	      pConn->AbortTransaction();

	    // Release results
	    pConn->ReleaseResults();
	  
	    // Return failure
	    return false;
	  }

	  my_list.push_back(std::pair<id_d_t, monad_m>(object_id_d, monad));

	}
      }     
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::UpgradeObjectType");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }
    std::cerr << " Done!" << std::endl;

    std::cerr << "Updating the ones..." << std::endl;
    std::list<std::pair<id_d_t, monad_m> >::const_iterator i;
    i = my_list.begin();
    long count = 0;
    while (i != my_list.end()) {
      if ((count % 500) == 499) {
	std::cerr << "." << std::flush;
      }

      if ((count % 10000) == 9999) {
	// Commit transaction
	if (bDoCommit) {
	  if (!pConn->CommitTransaction()) {
	    DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::UpgradeObjectType");
	    return false;
	  }
	}

	// Begin transaction
	bDoCommit = pConn->BeginTransaction();

      }
      // SET to 'Y' as needed
      {
	std::ostringstream query_stream;
	
	query_stream 
	  << "UPDATE " << OTN << "_monad_ms\n"
	  << "SET is_first = 'Y'\n"
	  << "WHERE object_id_d = " << i->first << "\n"
	  << "      AND mse_first = " << i->second << "\n";
	if (!pConn->Exec(query_stream.str().c_str())) {
	  DEBUG_QUERY_FAILED("EMdFDB::UpgradeObjectType", query_stream.str().c_str());
	  std::cerr << "Failed!" << std::endl;
	  if (bDoCommit)
	    pConn->AbortTransaction();
	  return false;
	}
      }
      ++count;
      ++i;
    }
    std::cerr << " Done!" << std::endl;

    std::cerr << "DROPping table " << OTN << "_monad_ds..." << std::flush;
    // DROP OT_monad_ds
    {
      std::ostringstream query_stream;
      query_stream 
	<< "DROP TABLE " << OTN << "_monad_ds\n";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::UpgradeObjectType", query_stream.str().c_str());
	std::cerr << "Failed!" << std::endl;
	if (bDoCommit)
	  pConn->AbortTransaction();
	return false;
      }
    }
    std::cerr << " Done!" << std::endl;

    // Commit transaction
    if (bDoCommit) {
      if (!pConn->CommitTransaction()) {
	DEBUG_COMMITTRANSACTION_FAILED("EMdFDB::UpgradeObjectType");
	return false;
      }
    }
    
    // Release results
    pConn->ReleaseResults();
	  
    // If we got this far, there were no errors
    return true;
  }
}

//
// Upgrading to 1.1.5 from pre-1.1.5
bool EMdFDB::UpgradeDatabaseTo115(const std::string& database_name)
{
  if (pConn == 0)
    return false;
  else {
    //
    // Adding min_m/max_m table
    //
    std::cerr << "Adding min_m/max_m tables..." << std::flush;
    if (!Create_min_max_m_table()) {
      std::cerr << "Failed!" << std::endl;
      return false;
    }

    //
    // Setting min_m/max_m
    //
    Set_min_max_m_from_object_types(true);

    // If we got this far, there were no errors
    return true;
  }
}

//
// Upgrading to 1.1.8 from pre-1.1.8
bool EMdFDB::UpgradeDatabaseTo118(const std::string& database_name)
{
  if (pConn == 0)
    return false;
  else {
    // Create sequece_118_X tables
    // Creaing sequence-tables
    std::cerr << "Creating sequence_118_X tables/sequences... " << std::flush;
    if (!CreateSequenceTables118()) {
      std::cerr << "Failed!" << std::endl;
      return false;
    }
    std::cerr << "Done!" << std::endl;

    // Moving sequence-values
    std::cerr << "Moving sequence_X values... " << std::flush;
    if (!MoveSequenceValue118(SEQUENCE_OBJECT_ID_DS)) {
      DEBUG_X_FAILED("EMdFDB::UpgradeDatabase118", "Moving sequence OBJECT_ID_DS");
      std::cerr << "Failed!" << std::endl;
      return false;
    }
    if (!MoveSequenceValue118(SEQUENCE_OTHER_IDS)) {
      DEBUG_X_FAILED("EMdFDB::UpgradeDatabase118", "Moving sequence OTHER_ID");
      std::cerr << "Failed!" << std::endl;
      return false;
    }
    std::cerr << "Done!" << std::endl;

    // Declare szQuery
    char szQuery[512];

    // Drop sequence
    std::cerr << "Dropping old sequence_0... " << std::flush;
    sprintf(szQuery, 
	    "DROP TABLE sequence_%d",
	    SEQUENCE_OBJECT_ID_DS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_X_FAILED("EMdFDB::UpgradeDatabase118", "Dropping old sequence OBJECT_ID_DS");
      std::cerr << "Failed!" << std::endl;
      return false;
    }
    std::cerr << "Done!" << std::endl;

    // Drop sequence
    std::cerr << "Dropping old sequence_1... " << std::flush;
    sprintf(szQuery, 
	    "DROP TABLE sequence_%d",
	    SEQUENCE_OTHER_IDS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_X_FAILED("EMdFDB::UpgradeDatabase118", "Dropping old sequence OTHER_IDS");
      std::cerr << "Failed!" << std::endl;
      return false;
    }
    std::cerr << "Done!" << std::endl;


    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::TableExists(std::string table_name, bool&  /* out */ bTableExists)
{
  if (pConn == 0)
    return false;
  else {
    // Test that connection is OK
    if (!pConn->ConnectionOk()) {
      DEBUG_X_FAILED("EMdFDB::TableExists", "ConnectionOk()");
      return false;
    }

    // SELECT
    {
      std::ostringstream query_stream;
      query_stream 
	<< "SELECT *" << '\n'
	<< "FROM " << table_name << '\n'
	<< "LIMIT 1";
      if (!pConn->Exec(query_stream.str().c_str())) {
	bTableExists = false;
      } else {
	bTableExists = true;
      }
    }

    // If we got this far, there were no errors
    return true;
  }
}


//
// min_m/max_m
// 

// Create tables
bool EMdFDB::Create_min_max_m_table(void)
{
  if (pConn == 0)
    return false;
  else {
    // CREATE TABLE
    {
      std::ostringstream query_stream;
      query_stream 
	<< "CREATE TABLE max_m (\n" 
	<< "    dummy_id INT NOT NULL,\n"
	<< "    max_m    INT NOT NULL,\n"
	<< "    PRIMARY KEY (dummy_id)\n"
	<< ")";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::Create_min_max_m_table", query_stream.str().c_str());
	return false;
      } 
    }

    // Fill table with initial value
    {
      std::ostringstream query_stream;
      query_stream 
	<< "INSERT INTO max_m (dummy_id, max_m)\n"
	<< "VALUES ( 0 , 0 )";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::Create_min_max_m_table", query_stream.str().c_str());
	return false;
      } 
    }

    // CREATE TABLE
    {
      std::ostringstream query_stream;
      query_stream 
	<< "CREATE TABLE min_m (\n" 
	<< "    dummy_id INT NOT NULL,\n"
	<< "    min_m    INT NOT NULL,\n"
	<< "    PRIMARY KEY (dummy_id)\n"
	<< ")";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::Create_min_max_m_table", query_stream.str().c_str());
	return false;
      } 
    }

    // Fill table with initial value
    {
      std::ostringstream query_stream;
      query_stream 
	<< "INSERT INTO min_m (dummy_id, min_m)\n"
	<< "VALUES ( 0 , " << MAX_MONAD << " )";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::Create_min_max_m_table", query_stream.str().c_str());
	return false;
      } 
    }

    // If we got this far, there were no errors
    return true;
  }
}

// Iterate over all object types, calling Set_min_max_m_from_object_type
// for each object type.
bool EMdFDB::Set_min_max_m_from_object_types(bool bBeVerbose)
{
  if (pConn == 0)
    return false;
  else {
    // Get list of object types
    std::list<std::string> object_type_names;
    if (!GetObjectTypes(object_type_names)) {
      std::cerr << "FAILURE: Could not get list of object types in database!" << std::endl;
      return false;
    }

    // Reset min_m 
    if (!Set_min_m(MAX_MONAD, true)) {
      std::cerr << "FAILURE: Could not set min_m unconditionally!" << std::endl;
      return false;
    }

    // Reset max_m 
    if (!Set_max_m(0, true)) {
      std::cerr << "FAILURE: Could not set max_m unconditionally!" << std::endl;
      return false;
    }

    // Output message if being verbose
    if (bBeVerbose) {
      std::cerr << std::endl << "Setting min_m/max_m from object types..." << std::endl << std::flush;
    }

    // Iterator over list
    std::list<std::string>::const_iterator ci;
    ci = object_type_names.begin();
    while (ci != object_type_names.end()) {
      // Output object type name if being verbose
      if (bBeVerbose) {
	std::cerr << *ci;
      }

      // Set min_m/max_m from object type
      if (!Set_min_max_m_from_object_type(*ci, bBeVerbose)) {
	return false;
      }

      // Advance iterator
      ++ci;

      // Output ", " if not end and being verbose
      if (ci != object_type_names.end()) {
	if (bBeVerbose) {
	  std::cerr << ", ";
	}
      }
    }

    // Output message if being verbose
    if (bBeVerbose) {
      std::cerr << "... Done!" << std::endl << std::flush;
    }

    // If we got this far, there were no errors
    return true;
  }
}

// Set min_m and max_m from a certain object type (if necessary)
bool EMdFDB::Set_min_max_m_from_object_type(const std::string& object_type_name, bool bBeVerbose)
{
  if (pConn == 0)
    return false;
  else {
    // Normalize object type name
    std::string OTN = NormalizeOTName(object_type_name);


    // Check the object type has rows
    bool bObjectTypeHasObjects = false;
    {
      std::ostringstream query_stream;
      query_stream 
	<< "SELECT object_id_d" << '\n'
	<< "FROM " << OTN << "_monad_ms" << '\n'
	<< "LIMIT 1";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::Set_min_max_m_from_object_type", query_stream.str().c_str());
	return false;
      }

      // Is there at least one tuple? (There should be either 1 or none)
      if (pConn->NoOfResultTuples() >= 1) {
	// There are objects in the table
	bObjectTypeHasObjects = true;
      } else {
	// There are no objects in the table
	bObjectTypeHasObjects = false;
      }

      // Release results
      pConn->ReleaseResults();
    }

    // Does the object type not have objects?
    if (!bObjectTypeHasObjects) {
      // If the object type has no objects,
      // there is no point in going further.
      // Return with success.
      // Also, it is an error to proceed, because
      // MIN(mse_first) and MAX(mse_last) will
      // return 0 and 0 respectively, thus
      // setting min_m to 0, at least on PostgreSQL.
      return true;
    }



    //
    // Get min_m/max_m candidate
    //
    monad_m min_m_candidate;
    monad_m max_m_candidate;

    // SELECT
    {
      std::ostringstream query_stream;
      query_stream 
	<< "SELECT MIN(mse_first), MAX(mse_last)\n"
	<< "FROM " << OTN << "_monad_ms";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::Set_min_max_m_from_object_type", query_stream.str().c_str());
	return false;
      } 
    }

    // Access result
    bool bSetMinMax = true;
    try {
      if (!pConn->AccessTuple(0, 0, "EMdFDB::Set_min_max_m_from_object_type", min_m_candidate)) {
	// DEBUG information has already been printed if it is on.
	// Release results
	pConn->ReleaseResults();
	
	// Return failure  
	return false;
      }

      if (!pConn->AccessTuple(0, 1, "EMdFDB::Set_min_max_m_from_object_type", max_m_candidate)) {
	// DEBUG information has already been printed if it is on.
	// Release results
	pConn->ReleaseResults();
	
	// Return failure  
	return false;
      }
    } catch (EMdFNULLValueException e) {
      // There was a NULL Value exception, so the object type
      // was probably empty.
      // Do not set min/max
      bSetMinMax = false;
    }

    // Should we set min and max?
    if (bSetMinMax) {
      // Set min_m (if smaller)
      if (!Set_min_m(min_m_candidate)) {
	DEBUG_X_FAILED("EMdFDB::Set_min_max_m_from_object_type", "Setting min_m");
	// Release results
	pConn->ReleaseResults();
	
	// Return failure  
	return false;
      }

      // Set max_m (if larger)
      if (!Set_max_m(max_m_candidate)) {
	DEBUG_X_FAILED("EMdFDB::Set_min_max_m_from_object_type", "Setting max_m");
	// Release results
	pConn->ReleaseResults();
	
	// Return failure  
	return false;
      }
    }

    // Release results
    pConn->ReleaseResults();
	
    // If we got this far, there were no errors
    return true;
  }
}

// Get max_m
bool EMdFDB::Get_max_m(monad_m& /* out */ max_m)
{
  if (pConn == 0)
    return false;
  else {
    //
    // Get max_m 
    //

    // SELECT
    {
      std::ostringstream query_stream;
      query_stream 
	<< "SELECT max_m\n"
	<< "FROM max_m\n"
	<< "WHERE dummy_id = 0";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::Get_max_m", query_stream.str().c_str());
	return false;
      } 
    }

    // Access result
    try {
      if (!pConn->AccessTuple(0, 0, "EMdFDB::Get_max_m", max_m)) {
	// DEBUG information has already been printed if it is on.
	// Release results
	pConn->ReleaseResults();
	
	// Return failure  
	return false;
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::Get_max_m");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }


    // Release results
    pConn->ReleaseResults();
	
    // If we got this far, there were no errors
    return true;
  }
}

// Set max_m if necessary
bool EMdFDB::Set_max_m(monad_m max_m_candidate, bool bSetUnconditionally)
{
  if (pConn == 0)
    return false;
  else {
    // Get current max_m for comparison
    monad_m current_max_m;
    if (!Get_max_m(current_max_m)) {
      DEBUG_X_FAILED("EMdFDB::Set_max_m", "Getting current max_m");
      return false;
    }

    // Check if we should update
    if (bSetUnconditionally || (max_m_candidate > current_max_m)) {
      // Yes, so UPDATE
      {
	std::ostringstream query_stream;
	query_stream 
	  << "UPDATE max_m\n"
	  << "SET max_m = " << max_m_candidate << '\n'
	  << "WHERE dummy_id = 0";
	if (!pConn->Exec(query_stream.str().c_str())) {
	  DEBUG_QUERY_FAILED("EMdFDB::Set_max_m", query_stream.str().c_str());
	  return false;
	} 
      }
    }

    // If we got this far, there were no errors
    return true;
  }
}

// Get min_m
bool EMdFDB::Get_min_m(monad_m& /* out */ min_m)
{
  if (pConn == 0)
    return false;
  else {
    //
    // Get min_m 
    //

    // SELECT
    {
      std::ostringstream query_stream;
      query_stream 
	<< "SELECT min_m\n"
	<< "FROM min_m\n"
	<< "WHERE dummy_id = 0";
      if (!pConn->Exec(query_stream.str().c_str())) {
	DEBUG_QUERY_FAILED("EMdFDB::Get_min_m", query_stream.str().c_str());
	return false;
      } 
    }

    // Access result
    try {
      if (!pConn->AccessTuple(0, 0, "EMdFDB::Get_min_m", min_m)) {
	// DEBUG information has already been printed if it is on.
	// Release results
	pConn->ReleaseResults();
	
	// Return failure  
	return false;
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::Get_min_m");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }

    // Release results
    pConn->ReleaseResults();
	
    // If we got this far, there were no errors
    return true;
  }
}

// Set min_m if necessary
bool EMdFDB::Set_min_m(monad_m min_m_candidate, bool bSetUnconditionally) 
{
  if (pConn == 0)
    return false;
  else {
    // Get current min_m for comparison
    monad_m current_min_m;
    if (!Get_min_m(current_min_m)) {
      DEBUG_X_FAILED("EMdFDB::Set_min_m", "Getting current min_m");
      return false;
    }

    // Check if we should update
    if (bSetUnconditionally || (min_m_candidate < current_min_m)) {
      // Yes, so UPDATE
      {
	std::ostringstream query_stream;
	query_stream 
	  << "UPDATE min_m\n"
	  << "SET min_m = " << min_m_candidate << '\n'
	  << "WHERE dummy_id = 0";
	if (!pConn->Exec(query_stream.str().c_str())) {
	  DEBUG_QUERY_FAILED("EMdFDB::Set_min_m", query_stream.str().c_str());
	  return false;
	} 
      }
    }

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::Get_all_m_1(CSet_of_monad_ms& /* out */ all_m_1)
{
  if (pConn == 0)
    return false;
  else {
    // Get min_m
    monad_m min_m;
    if (!Get_min_m(min_m)) {
      DEBUG_X_FAILED("EMdFDB::Get_all_m_1", "Getting min_m");
      return false;
    }

    // Get max_m
    monad_m max_m;
    if (!Get_max_m(max_m)) {
      DEBUG_X_FAILED("EMdFDB::Get_all_m_1", "Getting max_m");
      return false;
    }

    // Create new set of monads
    CSet_of_monad_ms empty;
    all_m_1 = empty;
    all_m_1.add(min_m, max_m);

    // If we got this far, there were no errors
    return true;
  }
}


//
// Transactions
// 

// Begin transaction
bool EMdFDB::BeginTransaction(void)
{
  if (pConn == 0)
    return false;
  else {
    return pConn->BeginTransaction();
  }
}

// Commit transaction
bool EMdFDB::CommitTransaction(void)
{
  if (pConn == 0)
    return false;
  else {
    return pConn->CommitTransaction();
  }
}

// Abort transaction
bool EMdFDB::AbortTransaction(void)
{
  if (pConn == 0)
    return false;
  else {
    return pConn->AbortTransaction();
  }
}

std::string EMdFDB::encode_feature_name(const std::string feature_name)
{
  // Get lowercase name
  std::string locase_name;
  str_tolower(feature_name, locase_name);

  // Get encoded name
  return "mdf_" + locase_name;
}

std::string EMdFDB::encode_feature_name_for_prequery_string(const std::string feature_name)
{
  std::string strResult;
  if (strcmp_nocase(feature_name, "self") == 0) {
    strResult = "OS.object_id_d";
  } else {
    strResult = std::string("OS.") + encode_feature_name(feature_name);
  }
  return strResult;
}


std::string EMdFDB::decode_feature_name(const std::string coded_feature_name)
{
  return coded_feature_name.substr(4);
}



bool EMdFDB::Type_id_to_string(id_d_t type_id, std::string& result)
{
  bool bResult;
  bool bExists = false;

  // Get feature_type
  long length;
  id_d_t ft;
  decode_feature_type_id(type_id, ft, length);

  switch(ft) {
  case FEATURE_TYPE_INTEGER:
    result = "integer";
    bResult = true;
    break;
  case FEATURE_TYPE_ID_D:
    result = "id_d";
    bResult = true;
    break;
  case FEATURE_TYPE_STRING:
    result = "string(" + long2string(length/2) + ")";
    bResult = true;
    break;
  case FEATURE_TYPE_ASCII:
    result = "ascii(" + long2string(length) + ")";
    bResult = true;
    break;
  default: // Enumerations: same as ID_D
    if (!EnumExists(type_id, bExists, result)) {
      bResult = false;
    } else {
      if (!bExists) {
	bResult = false;
      } else {
	bResult = true;
      }
    }
  }

  return true;
}

bool EMdFDB::Type_id_to_TableColumnType(id_d_t type_id, TableColumnType& tc_type, std::string& enum_type)
{
  bool bResult;

  // Get feature_type
  long length;
  id_d_t ft;
  decode_feature_type_id(type_id, ft, length);

  bool bExists = false;
  switch(ft) {
  case FEATURE_TYPE_INTEGER:
    tc_type = kTCInteger;
    bResult = true;
    break;
  case FEATURE_TYPE_ID_D:
    tc_type = kTCID_D;
    bResult = true;
    break;
  case FEATURE_TYPE_STRING:
    tc_type = kTCString;
    bResult = true;
    break;
  case FEATURE_TYPE_ASCII:
    tc_type = kTCString;
    bResult = true;
    break;
  default: // Enumerations: same as ID_D
    tc_type = kTCEnum;
    if (!EnumExists(type_id, bExists, enum_type)) {
      bResult = false;
    } else {
      if (!bExists) {
	bResult = false;
      } else {
	bResult = true;
      }
    }
  }

  return true;
}

bool EMdFDB::CreateIndex(const std::string& index_name, const std::string& table_name, const std::list<std::string>& columns)
{
  if (pConn == 0)
    return false;
  else {
    //
    // Create query
    //
    std::ostringstream query_stream;
    query_stream 
      << "CREATE INDEX " << index_name << '\n'
      << "ON " << table_name << '\n'
      << "(";
    std::list<std::string>::const_iterator ci;
    ci = columns.begin();
    while (ci != columns.end()) {
      query_stream << *ci;
      ++ci;
      if (ci != columns.end()) {
	query_stream << ", ";
      }
    }
    query_stream << ")\n";


    //
    // Execute query
    //
    if (!pConn->Exec(query_stream.str().c_str())) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateIndex", query_stream.str().c_str());
      return false;
    }


    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::CreateIndicesOnOTObjects(const std::string& object_type_name) 
{
  // Normalize object type name
  std::string OTN = NormalizeOTName(object_type_name);
  
  // Create index on first_monad on OT_objects
  std::list<std::string> column_list;
  column_list.push_back("first_monad");
  if (!CreateIndex(OTN + "_o_fm_i", OTN + "_objects", column_list)) {
    return false;
  }
  column_list.pop_back();
  
  // Create index on last_monad on OT_objects
  column_list.push_back("last_monad");
  if (!CreateIndex(OTN + "_o_lm_i", OTN + "_objects", column_list)) {
    return false;
  }
  column_list.pop_back();
  
  // Create index on monad_ms_id on OT_objects
  column_list.push_back("monad_ms_id");
  if (!CreateIndex(OTN + "_o_mmid_i", OTN + "_objects", column_list)) {
    return false;
  }

  // If we got this far, there were no errors
  return true;
}

bool EMdFDB::CreateIndicesOnOTMonadMs(const std::string& object_type_name)
{
  // Normalize object type name
  std::string OTN = NormalizeOTName(object_type_name);
  
  // Create index on object_id_d on OT_monad_ms
  std::list<std::string> column_list;
  column_list.push_back("object_id_d");
  if (!CreateIndex(OTN + "_mm_o_i", OTN + "_monad_ms", column_list)) {
    return false;
  }
  column_list.pop_back();

  // Create index on mse_first on OT_monad_ms
  column_list.push_back("mse_first");
  if (!CreateIndex(OTN + "_mm_mf_i", OTN + "_monad_ms", column_list)) {
    return false;
  }
  column_list.pop_back();

  // Create index on mse_first on OT_monad_ms
  column_list.push_back("mse_last");
  if (!CreateIndex(OTN + "_mm_ml_i", OTN + "_monad_ms", column_list)) {
    return false;
  }
  column_list.pop_back();

  // Create index on is_first on OT_monad_ms
  column_list.push_back("is_first");
  if (!CreateIndex(OTN + "_mm_if_i", OTN + "_monad_ms", column_list)) {
    return false;
  }
  column_list.pop_back();

  // Create index on monad_ms_id on OT_monad_ms
  column_list.push_back("monad_ms_id");
  if (!CreateIndex(OTN + "_mm_mmid_i", OTN + "_monad_ms", column_list)) {
    return false;
  }
  
  // If we got this far, there were no errors
  return true;
}

bool EMdFDB::CreateIndicesOnDatabase(const std::string& database_name)
{
  if (pConn == 0)
    return false;
  else {
    if (!UseDatabase(database_name)) {
      std::cerr << "FAILURE: Could not USE DATABASE " << database_name << "!" << std::endl;
      return false;
    }

    std::list<std::string> object_type_names;
    if (!GetObjectTypes(object_type_names)) {
      std::cerr << "FAILURE: Could not get list of object types in database " << database_name << "!" << std::endl;
      return false;
    }

    std::list<std::string>::const_iterator ci;
    ci = object_type_names.begin();
    while (ci != object_type_names.end()) {
      if (!CreateIndicesOnObjectType(*ci)) {
	return false;
      }
      ++ci;
    }

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::DropIndicesOnDatabase(const std::string& database_name)
{
  if (pConn == 0)
    return false;
  else {
    if (!UseDatabase(database_name)) {
      std::cerr << "FAILURE: Could not USE DATABASE " << database_name << "!" << std::endl;
      return false;
    }

    std::list<std::string> object_type_names;
    if (!GetObjectTypes(object_type_names)) {
      std::cerr << "FAILURE: Could not get list of object types in database " << database_name << "!" << std::endl;
      return false;
    }

    std::list<std::string>::const_iterator ci;
    ci = object_type_names.begin();
    while (ci != object_type_names.end()) {
      if (!DropIndicesOnObjectType(*ci)) {
	return false;
      }
      ++ci;
    }

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::CreateIndicesOnObjectType(const std::string& database_name, const std::string& object_type_name)
{
  if (pConn == 0)
    return false;
  else {
    if (!UseDatabase(database_name)) {
      std::cerr << "FAILURE: Could not USE DATABASE " << database_name << "!" << std::endl;
      return false;
    }

    if (!CreateIndicesOnObjectType(object_type_name))
      return false;

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::DropIndicesOnObjectType(const std::string& database_name, const std::string& object_type_name)
{
  if (pConn == 0)
    return false;
  else {
    if (!UseDatabase(database_name)) {
      std::cerr << "FAILURE: Could not USE DATABASE " << database_name << "!" << std::endl;
      return false;
    }

    if (!DropIndicesOnObjectType(object_type_name))
      return false;

    // If we got this far, there were no errors
    return true;
  }
}

bool EMdFDB::CreateIndicesOnObjectType(const std::string& object_type_name)
{
  // Create indices on OT_objects
  std::cerr << "Creating indices on " << object_type_name << "_objects..." << std::flush;
  CreateIndicesOnOTObjects(object_type_name);
  std::cerr << "!" << std::endl;

  // Create indices on OT_objects
  std::cerr << "Creating indices on " << object_type_name << "_monad_ms..." << std::flush;
  CreateIndicesOnOTMonadMs(object_type_name);
  std::cerr << "!" << std::endl;

  return true;
}

bool EMdFDB::DropIndicesOnObjectType(const std::string& object_type_name)
{
  // Create indices on OT_objects
  std::cerr << "Dropping indices on " << object_type_name << "_objects..." << std::flush;
  DropIndicesOnOTObjects(object_type_name);
  std::cerr << "!" << std::endl;

  // Create indices on OT_objects
  std::cerr << "Dropping indices on " << object_type_name << "_monad_ms..." << std::flush;
  DropIndicesOnOTMonadMs(object_type_name);
  std::cerr << "!" << std::endl;

  return true;
}

void EMdFDB::DropIndicesOnOTObjects(const std::string& object_type_name)
{
  // Normalize object type name
  std::string OTN = NormalizeOTName(object_type_name);

  // Drop indices  
  DropIndex(OTN + "_o_fm_i", OTN + "_objects");
  DropIndex(OTN + "_o_lm_i", OTN + "_objects");
}

void EMdFDB::DropIndicesOnOTMonadMs(const std::string& object_type_name)
{
  // Normalize object type name
  std::string OTN = NormalizeOTName(object_type_name);
  
  // Drop indices
  DropIndex(OTN + "_mm_o_i", OTN + "_monad_ms");
  DropIndex(OTN + "_mm_mf_i", OTN + "_monad_ms");
  DropIndex(OTN + "_mm_ml_i", OTN + "_monad_ms");
  DropIndex(OTN + "_mm_if_i", OTN + "_monad_ms");
}

void EMdFDB::DropIndex(const std::string& index_name, const std::string& table_name)
{
  if (pConn == 0)
    return;
  else {
    //
    // Create query
    //
    std::ostringstream query_stream;
    query_stream 
      << "DROP INDEX " << index_name 
#if USE_MYSQL
      << " ON " << table_name
#endif 
      << '\n';

    //
    // Execute query
    // Don't bother checking result
    //
    pConn->Exec(query_stream.str().c_str());


    // If we got this far, there were no errors
    return;
  }
}

bool EMdFDB::CreateSequenceTables110(bool bSeedTables)
{
  if (pConn == 0)
    return false;
  else {
    // Declare szQuery
    char szQuery[512];

    // Create table 'sequence_0'
    sprintf(szQuery, "CREATE TABLE sequence_%d (\n"
	    "   sequence_id INT NOT NULL,\n"
	    "   sequence_value INT NOT NULL,\n"
	    "   PRIMARY KEY (sequence_id)\n"
	    ")", SEQUENCE_OBJECT_ID_DS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables110", szQuery);
      return false;
    }

    // Create table 'sequence_1'
    sprintf(szQuery, "CREATE TABLE sequence_%d (\n"
	    "   sequence_id INT NOT NULL,\n"
	    "   sequence_value INT NOT NULL,\n"
	    "   PRIMARY KEY (sequence_id)\n"
	    ")", SEQUENCE_OTHER_IDS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables110", szQuery);
      return false;
    }

    if (bSeedTables) {
      // Seed sequence table with start value for SEQUENCE_OBJECT_ID_DS
      sprintf(szQuery,
	      "INSERT INTO sequence_%d (sequence_id, sequence_value)\n"
	      "VALUES ( %d, %d )", 
	      SEQUENCE_OBJECT_ID_DS, 
	      SEQUENCE_OBJECT_ID_DS, 
	      SEQUENCE_START);
      if (!pConn->Exec(szQuery)) {
	DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables110", szQuery);
	return false;
      }

      // Seed sequence table with start value for SEQUENCE_OTHER_ID_DS
      sprintf(szQuery,
	      "INSERT INTO sequence_%d (sequence_id, sequence_value)\n"
	      "VALUES ( %d, %d )", 
	      SEQUENCE_OTHER_IDS, 
	      SEQUENCE_OTHER_IDS, 
	      SEQUENCE_START);
      if (!pConn->Exec(szQuery)) {
	DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables110", szQuery);
	return false;
      }
    }

    return true;
  }
}

bool EMdFDB::CreateSequenceTables118()
{
  if (pConn == 0)
    return false;
  else {
    // Declare szQuery
    char szQuery[512];

    // Begin transaction
    bool bDoCommit = pConn->BeginTransaction();

#if USE_POSTGRESQL
    // Create table/sequence 'sequence_118_0'
    sprintf(szQuery,
	    "CREATE SEQUENCE sequence_118_%d\n"
	    "START %d;", SEQUENCE_OBJECT_ID_DS, SEQUENCE_START);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables118", szQuery);
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

    // Create table/sequence 'sequence_118_1'
    sprintf(szQuery,
	    "CREATE SEQUENCE sequence_118_%d\n"
	    "START %d;", SEQUENCE_OTHER_IDS, SEQUENCE_START);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables118", szQuery);
      if (bDoCommit)
	pConn->AbortTransaction();
      return false;
    }

#elif USE_MYSQL
    // Create table/sequence 'sequence_118_0'
    sprintf(szQuery, "CREATE TABLE sequence_118_%d (\n"
	    "   sequence_id INT NOT NULL,\n"
	    "   sequence_value INT NOT NULL,\n"
	    "   PRIMARY KEY (sequence_id)\n"
	    ")", SEQUENCE_OBJECT_ID_DS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables118", szQuery);
      return false;
    }

    // Create table/sequence 'sequence_118_1'
    sprintf(szQuery, "CREATE TABLE sequence_118_%d (\n"
	    "   sequence_id INT NOT NULL,\n"
	    "   sequence_value INT NOT NULL,\n"
	    "   PRIMARY KEY (sequence_id)\n"
	    ")", SEQUENCE_OTHER_IDS);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables118", szQuery);
      return false;
    }
#endif


#if USE_MYSQL
    // Seed sequence table with start value for SEQUENCE_OBJECT_ID_DS
    sprintf(szQuery,
	    "INSERT INTO sequence_118_%d (sequence_id, sequence_value)\n"
	    "VALUES ( %d, %d )", 
	    SEQUENCE_OBJECT_ID_DS, 
	    SEQUENCE_OBJECT_ID_DS, 
	    SEQUENCE_START-1);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables118", szQuery);
      return false;
    }

    // Seed sequence table with start value for SEQUENCE_OTHER_ID_DS
    sprintf(szQuery,
	    "INSERT INTO sequence_118_%d (sequence_id, sequence_value)\n"
	    "VALUES ( %d, %d )", 
	    SEQUENCE_OTHER_IDS, 
	    SEQUENCE_OTHER_IDS, 
	    SEQUENCE_START-1);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::CreateSequenceTables", szQuery);
      return false;
    }
#endif

    // Commit transaction if necessary
    if (bDoCommit)
      pConn->CommitTransaction();

    return true;
  }
}

bool EMdFDB::MoveSequenceValue110(int sequence)
{
  if (pConn == 0)
    return false;
  else {
    // Declare szQuery
    char szQuery[512];

    // Ask what the current value is
    sprintf(szQuery, "SELECT sequence_value\n"
	    "FROM sequences\n"
	    "WHERE sequence_id = %d\n", sequence);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::MoveSequenceValue110", szQuery);
      return false;
    }

    // Access result
    long sequence_value;
    try {
      if (!pConn->AccessTuple(0, 0, "EMdFDB::MoveSequenceValue110", sequence_value)) {
	// DEBUG information has already been printed if it is on.
	// Release results
	pConn->ReleaseResults();

	// Return failure
	return false;
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::MoveSequenceValue110");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }

    // Update sequence table with next id.
    sprintf(szQuery, 
	    "INSERT INTO sequence_%d (sequence_id, sequence_value)\n"
	    "VALUES ( %d, %ld )\n", sequence, sequence, sequence_value);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::MoveSequenceValue110", szQuery);
      return false;
    }

    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::MoveSequenceValue118(int sequence)
{
  if (pConn == 0)
    return false;
  else {
    // Declare szQuery
    char szQuery[512];

    // Ask what the current value is
    sprintf(szQuery, "SELECT sequence_value\n"
	    "FROM sequence_%d\n"
	    "WHERE sequence_id = %d\n", sequence, sequence);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::MoveSequenceValue118", szQuery);
      // Release results
      pConn->ReleaseResults();

      // Return failure
      return false;
    }

    // Access result
    long sequence_value;
    try {
      if (!pConn->AccessTuple(0, 0, "EMdFDB::MoveSequenceValue118", sequence_value)) {
	// DEBUG information has already been printed if it is on.
	// Release results
	pConn->ReleaseResults();

	// Return failure
	return false;
      }
    } catch (EMdFNULLValueException e) {
      DEBUG_NULL_VALUE_EXCEPTION("EMdFDB::MoveSequenceValue118");

      // Release results
      pConn->ReleaseResults();

      // There was a NULL value exception, hence we return an error
      return false;
    }

#if USE_POSTGRESQL
    // Update sequence table with next id.
    sprintf(szQuery, 
	    "SELECT setval('sequence_118_%d', %ld);", 
	    sequence, sequence_value);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::MoveSequenceValue118", szQuery);
      // Release results
      pConn->ReleaseResults();

      // Return failure
      return false;
    }
#elif USE_MYSQL
    // Update sequence table with next id.
    sprintf(szQuery, 
	    "UPDATE sequence_118_%d\n"
	    "SET sequence_value = %ld",
	    sequence, sequence_value);
    if (!pConn->Exec(szQuery)) {
      DEBUG_QUERY_FAILED("EMdFDB::MoveSequenceValue118", szQuery);
      // Release results
      pConn->ReleaseResults();

      // Return failure
      return false;
    }
#endif

    // Release results
    pConn->ReleaseResults();

    // If we got this far, the command succeeded.
    return true;
  }
}

bool EMdFDB::Vacuum(bool bAnalyze)
{
  if (pConn == 0)
    return false;
  else {
    return pConn->Vacuum(bAnalyze);
  }
}


std::string EMdFDB::BackendName(void)
{
#if USE_POSTGRESQL
  return "PostgreSQL";
#elif USE_MYSQL
  return "MySQL";
#else
  assert(false);
  return "Unknown backend.";
#endif
}

//
// Encodes a feature type and a length into a feature_type_id:
//
//  - If it is ASCII or STRING, then
//    - feature_type + 10*length
//  - Otherwise,
//    - feature_type
//
id_d_t encode_feature_type_id(id_d_t feature_type, long length)
{
  id_d_t result;
  if (feature_type == FEATURE_TYPE_ASCII || feature_type == FEATURE_TYPE_STRING) {
    result = feature_type + length*10;
  } else {
    result = feature_type;
  }
  return result;
}

//
// Decodes feature_type_id
//
void decode_feature_type_id(id_d_t feature_type_id, id_d_t& feature_type, long& length)
{
  if (feature_type_id >= MIN_ENUM_TYPE_ID) {
    // ENUM
    feature_type = feature_type_id;
    length = 11;
  } else {
    id_d_t ft = feature_type_id % 10;
    if (ft == FEATURE_TYPE_ASCII || ft == FEATURE_TYPE_STRING) {
      // ASCII or STRING
      feature_type = ft;
      length = (feature_type_id - ft) / 10;
    } else { // INTEGER or ID_D
      feature_type = feature_type_id;
      length = 11;
    }
  }
}

bool feature_type_id_is_ASCII(id_d_t feature_type_id) 
{
  id_d_t ft = feature_type_id % 10;
  return (feature_type_id < SEQUENCE_START) && (ft == FEATURE_TYPE_ASCII);
}

bool feature_type_id_is_STRING(id_d_t feature_type_id) 
{
  id_d_t ft = feature_type_id % 10;
  return (feature_type_id < SEQUENCE_START) && (ft == FEATURE_TYPE_STRING);
}

std::string feature_info2SQLvalue(const feature_info& fi)
{
  // Declare result
  std::string result;

  // Get feature type
  id_d_t ft;
  long dummy_length;
  decode_feature_type_id(fi.get_type(), ft, dummy_length);

  // Make string
  if (ft == FEATURE_TYPE_STRING) {
    result = "'" + string2dbstring(fi.get_default_value()) + "'";
  } else if (ft == FEATURE_TYPE_ASCII) {
    result = "'" + encodeASCIIstring(fi.get_default_value()) + "'";
  } else
    result = fi.get_default_value();

  // Return result
  return result;
}

std::string NormalizeOTName(const std::string& OTN)
{
  std::string result;
  str_tolower(OTN, result);
  return result;
}

